// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 07-09-1996
// -> double* knots in plaats van knots[100]

#ifndef __BSPLINE_H
#define __BSPLINE_H

#include <iostream>
#include <fstream>
#include <string>

#include <bbox.h>
#include <poly.h>

#include <double3.h>
#include <listtmpl.h>
#include <numtyp.h>
#include <sample.h>
#include <geotyp.h>

const int NUMCTLPTS = 500;

//-----------------------------------------------------------------------//
//                          BSplineSegment                               //
//-----------------------------------------------------------------------//
// The struct BSplineSegment represents the parameter interval [v,w], which
// is located in segment k.

struct BSplineSegment {
  double v;
  double w;
  int k;
  BSplineSegment(double v0, double w0, int k0): v(v0), w(w0), k(k0) {}
};

//-----------------------------------------------------------------------//
//                          BSplineSegmentList                           //
//-----------------------------------------------------------------------//

typedef List<BSplineSegment> BSplineSegmentList;

//-----------------------------------------------------------------------//
//                          BSpline                                      //
//-----------------------------------------------------------------------//
// 3 uniform B-spline curve of degree 3 with knot vector
// [0 1 2 3 ... ]
//
//             L+2
//      s(x) = Sum N[i,3](x) P[i]
//             i=0
//
// with P[i] the controlpoints and N[i,n] the i-th basisfunction of degree n
//
// See [Farin, p...].
//

class BSpline
{
  protected:
    List<double3> V;            // control points
    double knots[NUMCTLPTS];    // knots[0] ... knots[L+4]

  public:
    // constructors
    BSpline();
    BSpline(int n);            // n is number of controlpoints
    BSpline(std::string FileName);

    int NumControlPoints() const { return V.Count; }
    // returns the number of control points of the curve

    int NumKnots() const { return V.Count+2; }
    // returns the number of knots of the curve

    int NumSegments() const { return V.Count-3; }
    // returns the number of segments of the curve

    double MinPar() const { return knots[2]; }
    // returns the minimum absolute parameter value of the curve

    double MaxPar() const { return knots[NumKnots()-3]; }
    // returns the maximum absolute parameter value of the curve

    int Index(double u) const;
    // returns the index i of u (i.e. knots[i] <= u < knots[i+1])

    int Segment(double u) const { return Index(u)-2; }
    // returns the number of the segment that contains u
    // the segments are numbered from 0 .. NumSegments()-1

    double& Knot(int i) { return knots[i]; }
    // returns the i-th knot of the curve (0 <= i < NumKnots())

    double Knot(int i) const { return knots[i]; }
    // returns the i-th knot of the curve (0 <= i < NumKnots())

    double3& ControlPoint(int i) { return V[i]; }
    // returns the i-th ControlPoint of the curve (0 <= i < NumControlPoints())

    const double3& ControlPoint(int i) const { return V[i]; }
    // returns the i-th ControlPoint of the curve (0 <= i < NumControlPoints())

    DoubleVector GetControlVector() const;
    // computes the concatenation vector of the control points

    void SetControlVector(DoubleVector& x);
    // initializes the control points with the concatenation vector x

    void SetLine(int NumPoints, double Length);
    // arranges the control points of the curve such that they lie
    // on the line:
    //
    //     f(u) = (u, 0, 0)
    //
    // The curve is an (exact) approximation of the line on the domain
    // 0 <= u <= Length

    void SetCircle(int NumPoints, double Radius, int Angle);
    // arranges the control points of the curve such that they lie
    // on the circle:
    //
    //     f(u) = (R cos(u), R sin(u), 0)
    //
    // The curve is an approximation of the circle on the domain
    // 0 <= u <= M_PI*Angle/180

    double* KnotVector() { return knots; }
    // returns a pointer to the knot vector

    void DeleteControlPoint(int i) { V.AtDelete(i); }
    // removes the i-th control point

    void DeleteControlPoint(double3* x) { V.Delete(x); }
    // removes control point x

    void InsertControlPoint(int i, double3* x) { V.AtInsert(i, x); }
    // inserts a new control point between the (i-1)-th and i-th control point

    void InsertControlPoint(double3* x) { V.Insert(x); }
    // inserts a new control point at the end

    void Flush() { V.DeleteAll(); }
    // removes all control points

    double3 GetPoint(double u) const;
    // returns the point of the curve with absolute parameter u

    double3 GetDeriv(double u, int k) const;
    // returns the k-th derivative of the curve in the point with
    // absolute parameter u

    double3 GetTangent(double u) const { return GetDeriv(u,1); }
    // returns the tangent vector of the curve in the point with
    // absolute parameter u

    CurvSample GetSample(double u) const;
    // returns the values x(u), x'(u) and x''(u)

    double3 operator()(double t) const { return GetPoint(t); }
    // returns the point of the curve with relative parameter t

    double Curvature(double u) const;
    // returns the curvature in the point with
    // absolute parameter u

    double3 LaplaceBeltrami(double t, const BSpline& c);
    //
    //                          x''       (c',c'')x'
    //  computes    /\ (x)  = -------  -  ---------- at the point with relative parameter t
    //                c       (c',c')     (c',c')^2

    double ComputeBendEnergy(int degree = 9) const;
    // computes the integral
    //
    //      MaxPar()
    //         - 
    //        |      kappa(t)^2 ||x'(t)|| dt
    //       -   
    //     MinPar()
    //
    // using a degree 'degree' numerical integration routine for each segment
    // N.B. valid degrees are 1, 3, 5, 7 and 9!

    double ComputeSimpleBendEnergy(int degree = 9) const;
    // computes the integral
    //
    //      MaxPar()
    //         -
    //        |      ||x''(t)||^2 dt
    //       -
    //     MinPar()
    //
    // using a degree 'degree' numerical integration routine for each segment
    // N.B. valid degrees are 1, 3, 5, 7 and 9!

    double ComputeStretchEnergy(int degree = 9) const; 
    // computes the integral
    // 
    //      MaxPar() 
    //         -  
    //        |      ||x'(t)|| dt
    //       -    
    //     MinPar() 
    //
    // using a degree 'degree' numerical integration routine for each segment 
    // N.B. valid degrees are 1, 3, 5, 7 and 9!

    double ComputeSimpleStretchEnergy(int degree = 9) const;  
    // computes the integral 
    //   
    //      MaxPar()  
    //         -    
    //        |      ||x'(t)||^2 dt 
    //       -     
    //     MinPar() 
    // 
    // using a degree 'degree' numerical integration routine for each segment  
    // N.B. valid degrees are 1, 3, 5, 7 and 9!
 
    BoundingBox3D BBox(int i, int n) const;
    // returns the bounding box of the controlpoints P[i] .. P[i+n-1]

    Polynomial XPol(int i) const;
    // polynomial of the x-value of the curve on interval [knots[i+2], knots[i+3]]

    Polynomial YPol(int i) const;
    // polynomial of the y-value of the curve on interval [knots[i+2], knots[i+3]]

    Polynomial ZPol(int i) const;
    // polynomial of the z-value of the curve on interval [knots[i+2], knots[i+3]]
    double FindParameter(const Line3D& line);
    // estimates the relative parameter in [0,1] of the point on the curve with
    // minimal distance to the Line3D line
    // This parameter is derived from the point on the control polygon with
    // minimal distance to the line.

    int FindParameter(double u0, int index, const double3& p, double& result) const;
    // this method searches in the i-th segment of the curve for a point
    // with minimum distance to the point p; N.B. A Newton method is used,
    // so convergence is not guaranteed

    int FindParameter(const double3& p, double& result) const;
    // this method searches for a point on the curve with minimum distance
    // to the point p; again convergence is not guaranteed

    double ControlPolygonLength() const;
    // returns the sum of the distances of successive control points

    void ShowControlPoints() const;
    // print controlpoints

    void Show() const;
    // print controlpoints, knot vector

    int LoadFile(const char*);
    // loads control points from a file

    int SaveFile(std::string) const;
    // saves control points to a file

    friend std::istream& operator>>(std::istream& s, BSpline&);
    friend std::ostream& operator<<(std::ostream& s, const BSpline&);

  protected:

    void CalcSegments(double v, double w, BSplineSegmentList& sl) const;
    // This function splits the interval [v,w] into a number of subintervals
    // that lie entirely in one segment.   
};

#endif

