// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 06-12-1996

#ifndef __BSPLINEVALUES_H
#define __BSPLINEVALUES_H 1

#include <poly.h>
#include <sample.h>
#include <geotyp.h>
#include <dvector.h>
#include <dmatrix.h>
#include <trapezd.h>

// Notations
// N[i,3] = i-th B-spline basisfunction N of degree 3
// with knotvector [0 1 2 3 4 5]

class TBSplineEvaluator
{
  public:
    TBSplineEvaluator();
    // constructor (reads values from files N.dat, ...)
    
    void GetNValues(double t, double V[4]);
    // the vector V is computed such that
    //
    //       V[i] = N[i,3](t)
    
    void GetNxValues(double t, double V[4]);
    // the vector V is computed such that
    //
    //       V[i] = N'[i,3](t)
    //

    void GetNxxValues(double t, double V[4]);
    // the vector V is computed such that
    //
    //       V[i] = N''[i,3](t)
    
    void GetNNValues(double t, double M[4][4]);
    // the matrix M is computed such that
    //
    //       M[i][j] = N[i,3](t) N[j,3](t)

    void GetNxNxValues(double t, double M[4][4]);
    // the matrix M is computed such that
    //
    //       M[i][j] = N'[i,3](t) N'[j,3](t)
    
    void GetNxNxxValues(double t, double M[4][4]);
    // the matrix M is computed such that
    //
    //       M[i][j] = N'[i,3](t) N''[j,3](t)
    
    void GetNxxNxxValues(double t, double M[4][4]);
    // the matrix M is computed such that
    //
    //       M[i][j] = N''[i,3](t) N''[j,3](t)

    void GetNNValues(double u, double v, double M[4][4]);
    // the matrix M is computed such that
    //
    //       M[i][j] = N[i,3](u) N[j,3](v)

    void GetNNuValues(double u, double v, double M[4][4]);
    // the matrix M is computed such that
    //
    //       M[i][j] = N'[i,3](u) N[j,3](v)

    void GetNNvValues(double u, double v, double M[4][4]);
    // the matrix M is computed such that
    //
    //       M[i][j] = N[i,3](u) N'[j,3](v)

    void GetNNSamples(double u, double v, FuncSample2D M[4][4]);
    // returns samples of the functions
    //
    //      f(u,v) = N[i,3](u) N[j,3](v),   0 <= i,j <= 3
    //

    void GetNNIntegrals(double v, double w, double M[4][4]);
    // the matrix M is computed such that
    //
    //                    w
    //                    -
    //       M[i][j] =   |   N[i,3](x) N[j,3](x) dx
    //                  -
    //                  v
    //
    
    void GetNNxxIntegrals(double v, double w, double M[4][4]);
    // the matrix M is computed such that
    //
    //                    w
    //                    -
    //       M[i][j] =   |   N[i,3](x) N''[j,3](x) dx
    //                  -
    //                  v
    
    void GetNxNxIntegrals(double v, double w, double M[4][4]);
    // the matrix M is computed such that
    //
    //                    w
    //                    -
    //       M[i][j] =   |   N'[i,3](x) N'[j,3](x) dx
    //                  -
    //                  v
    
    void GetNxxNxxIntegrals(double v, double w, double M[4][4]);
    // the matrix M is computed such that
    //
    //                    w
    //                    -
    //       M[i][j] =   |   N''[i,3](x) N''[j,3](x) dx
    //                  -
    //                  v
    
    void GetNIntegrals(double v, double w, double V[4]);
    // the vector V is computed such that
    //
    //                    w
    //                    -
    //       V[i]    =   |   N[i,3](x) dx
    //                  -
    //                  v
    
    void GetNxIntegrals(double v, double w, double V[4]);
    // the vector V is computed such that
    //
    //                    w
    //                    -
    //       V[i]    =   |   N'[i,3](x) dx
    //                  -
    //                  v

    void GetSimpleStretchEnergyMatrix(DoubleMatrix& A);
    // The matrix A is computed with coefficients A(row,col) = a_ijkl such that
    //
    //             3  3
    //             -  -
    // a_ijkl  =  |  |  N_ij  N_kl   +  N_ij   N_kl   du dv
    //           -  -       u     u         v      v
    //           2  2
    
    void GetSimpleBendEnergyMatrix(DoubleMatrix& A);
    // The matrix A is computed with coefficients A(row,col) = a_ijkl such that
    //
    //             3  3
    //             -  -
    // a_ijkl  =  |  |  N_ij  N_kl   +  2 N_ij   N_kl   +  N_ij   N_kl   du dv
    //           -  -       uu    uu          uv     uv        vv     vv
    //           2  2
    //
    
    void GetNNNNValues(double u, double v, DoubleMatrix& A);
    // The matrix A is computed with coefficients A(row,col) = a_ijkl such that
    //           
    // a_ijkl  =  N (u) N (v) N (u) N (v)         0 <= i,j,k,l <= 3
    //             i     j     k     l
    //
    // where 2 <= u,v <= 3

    void GetNuNuNvNvValues(double u, double v, DoubleMatrix& A);
    // The matrix A is computed with coefficients A(row,col) = a_ijkl such that
    //
    //
    // a_ijkl  =  N'(u) N (v) N'(u) N (v) + N (u) N'(v) N (u) N'(v)       0 <= i,j,k,l <= 3
    //             i     j     k     l       i     j     k     l   
    //
    // where 2 <= u,v <= 3
    
    void GetNNValues(double u, double v, DoubleVector& B);
    // The vector B is computed with coefficients B(row) = b_ij such that
    //           
    // b_ij  =  N (u) N (v)         0 <= i,j <= 3
    //           i     j   
    //
    // where 2 <= u,v <= 3
    
    void GetNNIntegrals(const VerticalTrapezoid2D& gamma, DoubleVector& B);
    // The vector B is computed with coefficients B(row) = b_ij such that
    //
    //           -   3     3
    // b_ij  =  |   N (u) N (v) du dv
    //         -     i     j
    //        gamma
    
    void GetNNNNIntegrals(const VerticalTrapezoid2D& gamma, DoubleMatrix& A);
    // The matrix A is computed with coefficients A(row,col) = a_ijkl such that
    //
    //             -   3     3     3     3
    // a_ijkl  =  |   N (u) N (v) N (u) N (v) du dv
    //           -     i     j     k     l
    //          gamma

    void GetNuNuNvNvIntegrals(const VerticalTrapezoid2D& gamma, DoubleMatrix& A);
    // The matrix A is computed with coefficients A(row,col) = a_ijkl such that
    //
    //             -
    // a_ijkl  =  |   N'(u) N (v) N'(u) N (v) + N (u) N'(v) N (u) N'(v) du dv
    //           -     i     j     k     l       i     j     k     l    
    //          gamma
    
  private:
    double S[4][4];
    double B[4][4];
    Polynomial M[4];
    Polynomial Mx[4];
    Polynomial Mxx[4];
    Polynomial MM[4][4];
    Polynomial MxMx[4][4];
    Polynomial N[4];
    Polynomial Nx[4];
    Polynomial Nxx[4];
    Polynomial xN[4];
    Polynomial xxN[4];
    Polynomial NN[4][4];
    Polynomial NNxx[4][4];
    Polynomial xNN[4][4];
    Polynomial NxNx[4][4];
    Polynomial xNxNx[4][4];
    Polynomial NxxNxx[4][4];
    Polynomial xNxxNxx[4][4];

    void ReadMatrices();
    void ShowMatrices();
    void ReadDoubleMatrix(const char *FileName, double M[4][4]);
    void ReadPolynomialMatrix(const char *FileName, Polynomial M[4][4]);
    void ReadPolynomialVector(const char *FileName, Polynomial V[4]);
    void ShowDoubleMatrix(double M[4][4], const char* s);
};

extern TBSplineEvaluator BSplineEvaluator;

#endif
