// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

//

#ifndef __BOUNDINGBOX_H
#define __BOUNDINGBOX_H

#include <double3.h>
#include <iostream>

//-----------------------------------------------------------------------//
//                          BoundingBox3D                                //
//-----------------------------------------------------------------------//
// This class represents the 3-dimensional bounding box
//
//  { (x,y,z) | (xmin <= x <= xmax) && (ymin <= y <= ymax) && (zmin <= z <= zmax) }

class BoundingBox3D
{
  protected:
    double xmin, ymin, zmin, xmax, ymax, zmax;

  public:
    BoundingBox3D(double BIG=1e100);
    BoundingBox3D(const double3& min, const double3& max);
    BoundingBox3D(int x0, int y0, int z0, int x1, int y1, int z1);
    BoundingBox3D(const BoundingBox3D& box);

    void Insert(const double3& p);
    void Grow(double d);

    int Inside(const double3& p) const;
    double Size() const;
    void Show() const;
};

inline BoundingBox3D::BoundingBox3D(double BIG)
{ xmin=ymin=zmin=BIG; xmax=ymax=zmax=-BIG; }

inline BoundingBox3D::BoundingBox3D(int x0, int y0, int z0, int x1, int y1, int z1):
  xmin(x0), ymin(y0), zmin(z0), xmax(x1), ymax(y1), zmax(z1) {}

inline BoundingBox3D::BoundingBox3D(const double3& min, const double3& max)
{
  xmin=min.x(); ymin=min.y(); zmin=min.z(); xmax=max.x(); ymax=max.y(); zmax=max.z();
}

inline BoundingBox3D::BoundingBox3D(const BoundingBox3D& box)
{
  xmin = box.xmin;
  ymin = box.ymin;
  zmin = box.zmin;
  xmax = box.xmax;
  ymax = box.ymax;
  zmax = box.zmax;
}

inline void BoundingBox3D::Insert(const double3& p)
{
  if (p.x()<xmin) xmin=p.x(); if (p.x()>xmax) xmax=p.x();
  if (p.y()<ymin) ymin=p.y(); if (p.y()>ymax) ymax=p.y();
  if (p.z()<zmin) zmin=p.z(); if (p.z()>zmax) zmax=p.z();
}

inline void BoundingBox3D::Grow(double d)
{
  xmin -= d; ymin -= d; zmin -= d;
  xmax += d; ymax += d; zmax += d;
}

inline int BoundingBox3D::Inside(const double3& p) const
{
  return (xmin<=p.x()) && (ymin<=p.y()) && (zmin<=p.z()) &&
         (p.x()<=xmax) && (p.y()<=ymax) && (p.z()<=zmax);
}

inline void BoundingBox3D::Show() const
{
  std::cout << "BoundingBox3D: "<< xmin << " " << xmax << " " << ymin
       << " " << ymax << " " << zmin << " " << zmax << std::endl;
}

inline double BoundingBox3D::Size() const
{
  return (double3(xmax,ymax,zmax) - double3(xmin,ymin,zmin)).Length();
}

#endif

