// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Mon 06-24-1996

#ifndef __DMATRIX_H
#define __DMATRIX_H

#include <iostream>
#include <cmath>
#include <dvector.h>
#include <double3x3.h>
extern "C" {
#include <matrix.h>
#include <matrix2.h>
}

//-----------------------------------------------------------------------//
//                          DoubleMatrix                                 //
//-----------------------------------------------------------------------//
// The components of the matrix can be retrieved in two ways:
//   numbered from (0..m-1, 0..n-1), with the method operator(int,int)
//   numbered from (1..m, 1..n) with the method Elem(int)

class DoubleMatrix
{
  private:
    MAT* A;

  protected:
    double& elem(int i, int j) { return A->me[i][j]; }
    double elem(int i, int j) const { return A->me[i][j]; }
    MAT* mat() const { return A; }

  public:
    /////////////////////////////////////////////////////////////////
    //              Constructors and destructors                   //
    /////////////////////////////////////////////////////////////////

    DoubleMatrix() { A = m_get(0,0); }
    // constructor

    DoubleMatrix(int m, int n) { A = m_get(m,n); }
    // constructor (creates an uninitialized matrix of size m x n)

    DoubleMatrix(int m, int n, double a) { A = m_get(m,n); Fill(a); }
    // constructor (creates the matrix [{a,a,...,a}, {a,a,...,a} ...])

    DoubleMatrix(const DoubleMatrix& B) { A = m_copy(B.mat(), MNULL); }
    // copy constructor

    ~DoubleMatrix() { M_FREE(A); }
    // destructor

    /////////////////////////////////////////////////////////////////
    //              Methods for retrieving data                    //
    /////////////////////////////////////////////////////////////////

    int NumRows() const { return A->m; }
    // returns the number of rows
    
    int NumCols() const { return A->n; }
    // returns the number of columns

    double& operator()(int i, int j) { return elem(i,j); }
    // (i = 0..m-1; j = 0..n-1)

    double operator()(int i, int j) const { return elem(i,j); }
    // (i = 0..m-1; j = 0..n-1)

    double* operator[](int i) { return A->me[i]; }
    // makes it possible to use [][]

    double* operator[](int i) const { return A->me[i]; }
    // makes it possible to use [][]

    double& Elem(int i, int j) { return elem(i-1,j-1); }
    // returns the (i,j)-th component; (i = 1..m; j = 1..n)

    double Elem(int i, int j) const { return elem(i-1,j-1); }
    // returns the (i,j)-th component; (i = 1..m; j = 1..n)

    DoubleVector Row(int i) const;
    // returns the i-th row of the matrix (i = 0..m-1)

    DoubleVector Col(int i) const;
    // returns the i-th column of the matrix (i = 0 .. n-1)

    DoubleMatrix GetSubMatrix(int row1, int row2, int col1, int col2) const;
    // returns the submatrix with rows row1 .. row2 and columns col1..col2,
    // (row? = 0..m-1; col? = 0..n-1)

    DoubleVector EigenValues() const;
    // computes the eigenvalues of the matrix; the matrix must be
    // symmetric!!!!

    DoubleMatrix EigenVectors() const;
    // imaginary parts of eigenvectors are lost
    // result need not be exact

    DoubleMatrix EigenVectors(DoubleMatrix &result_im) const;
    // imaginary parts of eigenvectors are returned in parameter.
    // result need not be exact

    /////////////////////////////////////////////////////////////////
    //              Methods for changing data                      //
    /////////////////////////////////////////////////////////////////

    void SetSize(int m, int n) { m_resize(A, m, n); }
    // sets the size of the matrix to m x n
    // N.B. the contents of the matrix are lost!!

    void Fill(double fill);
    // makes all elements of the matrix equal to fill

    void FillDiag(double fill);
    // makes all elements (i,i) of the matrix equal to fill

    void FillRandom(double min, double max);
    // fills the matrix with random elements in the interval [min,max]

    void MakeZero() { m_zero(A); }
    // makes all elements of the matrix equal to 0

    void MakeUnit() { m_ident(A); }
    // makes all (i,i) elements equal to 1 and the rest equal to 0

    void MakeSymmetric();
    // the matrix A is replaced with 1/2(A + A^T)

    void CopyUpperToLower();
    // if the matrix is square every element A(i,j) with i<j is replaced with A(j,i)

    void SwapCols(int i, int j);
    // swaps the i-th and j-th column

    void SwapRows(int i, int j);
    // swaps the i-th and j-th row

    void SetColumn(int col, const DoubleVector& a);
    // overwrites the i-th column with the vector a
    
    void SetRow(int row, const DoubleVector& a);
    // overwrites the i-th row with the vector a
   
    void MultColumn(double t, int col);
    // multiplies the i-th column of the matrix by t

    void MultRow(double t, int row);
    // multiplies the Row(i) of the matrix by t

    void MultAddColumn(double t, int col1, int col2);
    // adds t*Col(col1) to Col(col2)

    void MultAddRow(double t, int row1, int row2);
    // adds t*Row(row1) to Row(row2)

    void DeleteColumns(const IntVector& cols);
    // deletes the rows of the matrix with indices in cols

    void DeleteRows(const IntVector& rows);
    // deletes the rows of the matrix with indices in rows

    /////////////////////////////////////////////////////////////////
    //              Miscellaneous methods                          //
    /////////////////////////////////////////////////////////////////

    void AddDouble3x3(int row, int col, const double3x3& A);
    // Adds A to the submatrix [row..row+2, col..col+2]

    DoubleVector LeftMultiplyXYZ(const DoubleVector& X) const;
    // returns the vector X^T A, where X is a concatenation
    // vector of 3D points
    
    DoubleVector RightMultiplyXYZ(const DoubleVector& X) const;
    // returns the vector AX, where X is a concatenation
    // vector of 3D points
    
    void GetColumnPermutation(DoubleMatrix& B, const IntVector& sigma) const;
    // returns a DoubleMatrix B with the property that the i-th column of B
    // equals the sigma[i]-th column of this DoubleMatrix

    void GetInverseColumnPermutation(DoubleMatrix& B, const IntVector& sigma) const;
    // returns a DoubleMatrix B with the property that the sigma[i]-th column
    // of B equals the i-th column of this DoubleMatrix

    void GetRowPermutation(DoubleMatrix& B, const IntVector& sigma) const;
    // returns a DoubleMatrix B with the property that the i-th row of B
    // equals the sigma[i]-th row of this DoubleMatrix

    void GetInverseRowPermutation(DoubleMatrix& B, const IntVector& sigma) const;
    // returns a DoubleMatrix B with the property that the sigma[i]-th row of B
    // equals the i-th row of this DoubleMatrix

    void RowHouse(const DoubleVector& v);
    // implementation of algorithm 5.1.2 of [Golub, p. 197]

    void ColHouse(const DoubleVector& v);
    // implementation of algorithm 5.1.3 of [Golub, p. 197]
    
    void RowRot(int col1, int col2, int startrow, double c, double s);
    // implementation of algorithm 5.1.6 of [Golub, p. 203]
    
    void ColRot(int row1, int row2, int startcol, double c, double s);
    // implementation of algorithm 5.1.7 of [Golub, p. 203]
    
    void RowFastRot(int col, int row, double alpha, double beta, int type);
    // implementation of algorithm 5.1.9 of [Golub, p. 208]
    // (N.B. slightly adapted!)

    DoubleMatrix Transpose() const;
    // returns the transpose of the matrix
   
    double FrobeniusNorm() const;
    // returns the Frobenius norm of the matrix
    // (this norm is defined as
    //
    //                  m   n           2
    //          Sqrt[  Sum Sum  |A(i,j)|  ]
    //                 i=1 j=1

    DoubleVector SolveBKP(const DoubleVector& b);
    // solves the system Ax=b
               
    DoubleVector SolveCH(const DoubleVector& b);
    // solves the system Ax=b
               
    DoubleVector SolveLDL(const DoubleVector& b);
    // solves the system Ax=b
               
    DoubleVector SolveLU(const DoubleVector& b);
    // solves the system Ax=b
               
    DoubleVector SolveQR(const DoubleVector& b);
    // solves the system Ax=b
               
    void operator  = (const DoubleMatrix& B) { A = m_copy(B.mat(), A); }
    void operator += (const DoubleMatrix& B) { m_add(A, B.mat(), A); }
    void operator -= (const DoubleMatrix& B) { m_sub(A, B.mat(), A); }
    void operator *= (double t)        { sm_mlt(t, A, A); }
    void operator /= (double t)        { sm_mlt(1/t, A, A); }

    friend DoubleMatrix operator - (const DoubleMatrix&);
    friend DoubleMatrix operator + (const DoubleMatrix& A, const DoubleMatrix& B);
    friend DoubleMatrix operator - (const DoubleMatrix& A, const DoubleMatrix& B);
    friend DoubleMatrix operator * (double t, const DoubleMatrix& A);
    friend DoubleMatrix operator * (const DoubleMatrix& A, double t);
    friend DoubleMatrix operator / (const DoubleMatrix& A, double t);
    friend DoubleMatrix operator * (const DoubleMatrix& A, const DoubleMatrix& B);

    friend DoubleVector operator * (const DoubleMatrix& A, const DoubleVector& X);
    friend DoubleVector operator * (const DoubleVector& X, const DoubleMatrix& A);

    friend std::ostream& operator<<(std::ostream& s, const DoubleMatrix& A);
    friend std::istream& operator>>(std::istream& s, DoubleMatrix& A);

    friend class SparseMatrix;
    friend class DoubleBKPFact;
    friend class DoubleCHFact;
    friend class DoubleLDLFact;
    friend class DoubleLUFact;
    friend class DoubleQRFact;
};

DoubleMatrix DiagonalMatrix(const DoubleVector& x);
// returns the diagonal matrix with x(i) as (i,i)-th element

DoubleMatrix DyadeMatrix(const DoubleVector& x, const DoubleVector& y);
// returns the matrix        T
//                       x  y

//-----------------------------------------------------------------------//
//                          DoubleBKPFact                                //
//-----------------------------------------------------------------------//
// the class DoubleBKPFact can be used to compute a
// Bunch-Kaufman-Parlett decomposition of a matrix
// for symmetric indefinite matrices (!)

class DoubleBKPFact {
  private:
    MAT* BKP;
    PERM* pivot;
    PERM* blocks;

  public:
    DoubleBKPFact(const DoubleMatrix& A);
    // constructor

    ~DoubleBKPFact();
    // destructor

    DoubleVector solve(const DoubleVector& b);
    // returns the solution x to the equation Ax = b
};

//-----------------------------------------------------------------------//
//                          DoubleCHFact                                 //
//-----------------------------------------------------------------------//
// the class DoubleCHFact can be used to compute a
// Cholesky decomposition of a matrix
// for symmetric and positive definite matrices (!)

class DoubleCHFact {
  private:
    MAT* CH;

  public:
    DoubleCHFact(const DoubleMatrix& A);
    // constructor

    ~DoubleCHFact();
    // destructor

    DoubleVector solve(const DoubleVector& b);
    // returns the solution x to the equation Ax = b
};

//-----------------------------------------------------------------------//
//                          DoubleLDLFact                                //
//-----------------------------------------------------------------------//
// the class DoubleLDLFact can be used to compute a
// LDL^T factorization of a matrix

class DoubleLDLFact {
  private:
    MAT* LDL;

  public:
    DoubleLDLFact(const DoubleMatrix& A);
    // constructor

    ~DoubleLDLFact();
    // destructor

    DoubleVector solve(const DoubleVector& b);
    // returns the solution x to the equation Ax = b
};

//-----------------------------------------------------------------------//
//                          DoubleLUFact                                 //
//-----------------------------------------------------------------------//
// the class DoubleLUFact can be used to compute a
// LU factorization of a matrix

class DoubleLUFact {
  private:
    PERM* pivot;
    MAT* LU;

  public:
    DoubleLUFact(const DoubleMatrix& A);
    // constructor

    ~DoubleLUFact();
    // destructor

    DoubleVector solve(const DoubleVector& b);
    // returns the solution x to the equation Ax = b
};

//-----------------------------------------------------------------------//
//                          DoubleQRFact                                 //
//-----------------------------------------------------------------------//
// the class DoubleQRFact can be used to compute a
// QR factorization of a matrix

// need dim(diag) >= # rows of A !!??

class DoubleQRFact {
  private:
    MAT* QR;
    VEC* diag;

  public:
    DoubleQRFact(const DoubleMatrix& A);
    // constructor

    ~DoubleQRFact();
    // destructor

    DoubleVector solve(const DoubleVector& b);
    // returns the solution x to the equation Ax = b
};

#endif

