// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 05-29-1996

#ifndef __DOUBLE3_H
#define __DOUBLE3_H

#include <cmath>
#include <iostream>
#include <int3.h>
#include <mymanip.h>

#ifndef EPSILON
#define EPSILON 1e-10
#endif

//-----------------------------------------------------------------------//
//                          double3                                      //
//-----------------------------------------------------------------------//
// this struct represents a vector containing 3 doubles

struct double3
{
  protected:
    double X[3];

  public:
    // constructors
    double3();
    double3(double fill) { X[0]=X[1]=X[2]=fill; }
    double3(double, double, double);
    double3(const double3&);
    double3(const int3&);

    int Dim() { return 3; }

    double Elem(int i) const { return X[i-1]; }
    // returns the i-th component (i = 1..3)

    double& Elem(int i) { return X[i-1]; }
    // returns a reference to the i-th component (i = 1..3)

    double operator[](int i) const { return X[i]; }
    // returns the i-th component (i = 0..2)

    double& operator[](int i) { return X[i]; }
    // returns a reference to the i-th component (i = 0..2)

    double& x() { return X[0]; }
    double& y() { return X[1]; }
    double& z() { return X[2]; }
    double x() const { return X[0]; }
    double y() const { return X[1]; }
    double z() const { return X[2]; }

    double Length2() const;
    // returns the squared length of the vector

    double Length() const;
    // returns the length of the vector

    double getNorm() const;
    // returns the length of the vector

    double Dist(const double3& p) const;
    // returns the Euclidian distance from the point X to p

    double3 Unit() const;
    // returns the vector X / Length(X)

    double3 getNormalized() const;
    // returns the vector X / Length(X)

    void MakeUnit();
    // makes the vector equal to (X) / Length(X)

    double normalize();
    // makes the vector equal to (X) / Length(X) and returns Length(X)

    void MakeZero();
    // makes the vector equal to (0,0,0)

    bool NearZero(double eps = 1e-7) const;
    // tests if the length is <= eps

    void Negate();
    // replaces the vector with -X

    void PerpendicularVectors(double3& v0, double3& v1) const;
    // returns two unit vectors v0 and v1, such that v0, v1 and X are orthogonal

    void operator = (const double3&);
    void operator = (double);
    double3  operator - () const;
    double3& operator += (const double3&);
    double3& operator -= (const double3&);
    double3& operator *= (double);
    double3& operator /= (double);
    int operator== (const double3&) const;

    friend double3 operator + (const double3&, const double3&);
    friend double3 operator - (const double3&, const double3&);
    friend double3 operator * (const double3&, double);
    friend double3 operator * (double, const double3&);
    friend double3 operator / (const double3&, double);
    friend double operator * (const double3&, const double3&); // inner product
    friend double operator , (const double3&, const double3&); // inner product
    friend double operator | (const double3&, const double3&); // inner product
    friend double3 Cross (const double3&, const double3&);     // cross product
    friend double3 operator^(const double3&, const double3&);  // cross product
    friend double Dist(const double3&, const double3&);
    friend std::ostream& operator<<(std::ostream& s, const double3& p);
    friend std::istream& operator>>(std::istream& s, double3& p);
};

inline double3::double3 ()
{ X[0]=X[1]=X[2]=0; }

inline double3::double3 (double x0, double x1, double x2)
{ X[0]=x0; X[1]=x1; X[2]=x2; }

inline double3::double3 (const double3& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; }

inline double3::double3 (const int3& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; }

inline void double3::operator = (const double3& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; }

inline void double3::operator = (double fill)
{ X[0]=fill; X[1]=fill; X[2]=fill; }

inline double3 double3::operator - () const
{ return double3(-X[0], -X[1], -X[2]); }

inline double3& double3::operator += (const double3& p)
{ X[0]+=p[0]; X[1]+=p[1]; X[2]+=p[2]; return *this; }

inline double3& double3::operator -= (const double3& p)
{ X[0]-=p[0]; X[1]-=p[1]; X[2]-=p[2]; return *this; }

inline double3& double3::operator *= (double t)
{ X[0]*=t; X[1]*=t; X[2]*=t; return *this; }

inline double3& double3::operator /= (double t)
{ X[0]/=t; X[1]/=t; X[2]/=t; return *this; }

inline int double3::operator== (const double3& p) const
{ return (X[0]==p[0] && X[1]==p[1] && X[2]==p[2]); }

inline double double3::Length2() const
{ return X[0]*X[0] + X[1]*X[1] + X[2]*X[2]; }

inline double double3::Length () const
{ return sqrt( (*this).Length2() ); }

inline double double3::getNorm() const
{ return Length(); }

inline double3 operator + (const double3& a, const double3& b)
{ return double3(a[0]+b[0], a[1]+b[1], a[2]+b[2]); }

inline double3 operator - (const double3& a, const double3& b)
{ return double3(a[0]-b[0], a[1]-b[1], a[2]-b[2]); }

inline double double3::Dist(const double3& p) const
{ return (*this - p).Length(); }

inline double3 operator * (const double3& a, double t)
{ return double3(t*a[0], t*a[1], t*a[2]); }

inline double3 operator * (double t, const double3& p)
{ return p * t; }

inline double3 operator / (const double3& a, double t)
{ return double3(a[0]/t, a[1]/t, a[2]/t); }

inline double operator * (const double3& a, const double3& b)
{ return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]; }

inline double operator , (const double3& a, const double3& b)
{ return a*b; }

inline double operator | (const double3& a, const double3& b)
{ return a*b; }

inline double3 Cross(const double3& a, const double3& b)
{
  return double3(a[1]*b[2] - a[2]*b[1],
                a[2]*b[0] - a[0]*b[2],
                a[0]*b[1] - a[1]*b[0] );
}

inline double3 operator^(const double3& a, const double3& b)
{ return Cross(a,b); }

inline double Dist(const double3& a, const double3& b)
{ return a.Dist(b); }

inline double3 double3::Unit() const
{ double l = Length();
  if (l > 0.0) return (*this/l); else return double3(0,0,0);
}

inline double3 double3::getNormalized() const
{ return Unit(); }

inline void double3::MakeUnit()
{ double l = Length(); if (l > 0.0) (*this)/=l; }

inline double double3::normalize()
{ double l = Length(); if (l > 0.0) (*this)/=l; return l; }

inline void double3::MakeZero()
{ X[0] = X[1] = X[2] = 0; }

inline bool double3::NearZero (double eps) const
{ return Length() <= eps; }

inline void double3::Negate ()
{ X[0] = -X[0]; X[1] = -X[1]; X[2] = -X[2]; }

inline void double3::PerpendicularVectors(double3& v0, double3& v1) const
{
  v0 = double3(X[1]-X[2], X[2]-X[0], X[0]-X[1]).Unit(); 
  v1 = ((*this)^v0).Unit();
}

inline std::ostream& operator<<(std::ostream& s, const double3& a)
{ return s << '(' << a[0] << ',' << a[1] << ',' << a[2] << ')'; }

inline std::istream& operator>>(std::istream& s, double3& a)
{
  return s >> std::ws >> skipbracket >> a[0]
           >> skipcomma >> a[1]
           >> skipcomma >> a[2]
           >> skipbracket;
}

#endif

