// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#ifndef __DOUBLE4_H
#define __DOUBLE4_H

#include <cmath>
#include <iostream>
#include <int4.h>
#include <mymanip.h>

//-----------------------------------------------------------------------//
//                          double4                                       //
//-----------------------------------------------------------------------//
// this struct represents a vector containing 4 doubles

struct double4
{
  protected:
    double X[4];

  public:
    // constructors
    double4();
    double4(double, double, double, double);
    double4(const double4&);
    double4(const int4&);

    double Elem(int i) const { return X[i-1]; }
    // returns the i-th component (i = 1..4)

    double& Elem(int i) { return X[i-1]; }
    // returns a reference to the i-th component (i = 1..4)

    double operator[](int i) const { return X[i]; }
    // returns the i-th component (i = 0..3)

    double& operator[](int i) { return X[i]; }
    // returns a reference to the i-th component (i = 0..3)

    double& x() { return X[0]; }
    double& y() { return X[1]; }
    double& z() { return X[2]; }
    double& w() { return X[3]; }
    double x() const { return X[0]; }
    double y() const { return X[1]; }
    double z() const { return X[2]; }
    double w() const { return X[3]; }

    double Length2() const;
    // returns the squared length of the vector

    double Length() const;
    // returns the length of the vector

    double Dist(const double4& p) const;
    // returns the Euclidian distance from the point X to p

    double4 Unit() const;
    // returns the vector X / Length(X)

    void MakeUnit();
    // makes the vector equal to (X) / Length(X)

    void MakeZero();
    // makes the vector equal to (0,0,0,0)

    bool NearZero(double eps = 1e-7) const;
    // tests if length <= eps

    void Negate();
    // replaces the vector with -X

    void    operator = (const double4&);
    double4  operator - () const;
    double4& operator += (const double4&);
    double4& operator -= (const double4&);
    double4& operator *= (double);
    double4& operator /= (double);

    friend double4 operator + (const double4&, const double4&);
    friend double4 operator - (const double4&, const double4&);
    friend double4 operator * (const double4&, double);
    friend double4 operator * (double, const double4&);
    friend double4 operator / (const double4&, double);
    friend double operator * (const double4&, const double4&); // inner product
    friend double operator , (const double4&, const double4&); // inner product
    friend double Dist(const double4&, const double4&);
    friend std::ostream& operator<<(std::ostream& s, const double4& p);
    friend std::istream& operator>>(std::istream& s, double4& p);
};

inline double4::double4 ()
{ X[0]=X[1]=X[2]=X[3]=0; }

inline double4::double4 (double x0, double x1, double x2, double x3)
{ X[0]=x0; X[1]=x1; X[2]=x2; X[3]=x3; }

inline double4::double4 (const double4& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; X[3]=p[3]; }

inline double4::double4 (const int4& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; X[3]=p[3]; }

inline void double4::operator = (const double4& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; X[3]=p[3]; }

inline double4 double4::operator - () const
{ return double4(-X[0], -X[1], -X[2], -X[3]); }

inline double4& double4::operator += (const double4& p)
{ X[0]+=p[0]; X[1]+=p[1]; X[2]+=p[2]; X[3]+=p[3]; return *this; }

inline double4& double4::operator -= (const double4& p)
{ X[0]-=p[0]; X[1]-=p[1]; X[2]-=p[2]; X[3]-=p[3]; return *this; }

inline double4& double4::operator *= (double t)
{ X[0]*=t; X[1]*=t; X[2]*=t; X[3]*=t; return *this; }

inline double4& double4::operator /= (double t)
{ X[0]/=t; X[1]/=t; X[2]/=t; X[3]/=t; return *this; }

inline double double4::Length2() const
{ return X[0]*X[0] + X[1]*X[1] + X[2]*X[2] + X[3]*X[3]; }

inline double double4::Length () const
{ return sqrt( (*this).Length2() ); }

inline double4 operator + (const double4& a, const double4& b)
{ return double4(a[0]+b[0], a[1]+b[1], a[2]+b[2], a[3]+b[3]); }

inline double4 operator - (const double4& a, const double4& b)
{ return double4(a[0]-b[0], a[1]-b[1], a[2]-b[2], a[3]-b[3]); }

inline double double4::Dist(const double4& p) const
{ return (*this - p).Length(); }

inline double4 operator * (const double4& a, double t)
{ return double4(t*a[0], t*a[1], t*a[2], t*a[3]); }

inline double4 operator * (double t, const double4& p)
{ return p * t; }

inline double4 operator / (const double4& a, double t)
{ return double4(a[0]/t, a[1]/t, a[2]/t, a[3]/t); }

inline double operator * (const double4& a, const double4& b)
{ return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3]; }

inline double operator , (const double4& a, const double4& b)
{ return a*b; }

inline double Dist(const double4& a, const double4& b)
{ return a.Dist(b); }

inline double4 double4::Unit() const
{ double l = Length();
  if (l > 0.0) return (*this/l); else return double4(0,0,0,0);
}

inline void double4::MakeUnit()
{ double l = Length(); if (l > 0.0) (*this)/=l; }

inline void double4::MakeZero()
{ X[0] = X[1] = X[2] = X[3] = 0; }

inline bool double4::NearZero(double eps) const
{ return Length() <= eps; }

inline void double4::Negate ()
{ X[0] = -X[0]; X[1] = -X[1]; X[2] = -X[2]; X[3] = -X[3]; }

inline std::ostream& operator<<(std::ostream& s, const double4& a)
{ return s << '(' << a[0] << ',' << a[1] << ',' << a[2] << ',' << a[3] << ')'; }

inline std::istream& operator>>(std::istream& s, double4& a)
{
  return s >> std::ws >> skipbracket >> a[0]
           >> skipcomma >> a[1]
           >> skipcomma >> a[2]
           >> skipcomma >> a[3]
           >> skipbracket;
}

# endif
