// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#ifndef __DOUBLE4X4_H
#define __DOUBLE4X4_H

#include <double3.h>
#include <double4.h>

//-----------------------------------------------------------------------//
//                          double4x4                                     //
//-----------------------------------------------------------------------//
class double4x4
{
  protected:
    double A[4][4]; 

  public:
    double4x4() {}
    // constructor

    double4x4(const double4x4& B) { *this = B;}
    // copy constructor

    double4x4(const double4& row0, const double4& row1, const double4& row2, const double4& row3); 
    // constructor (initializes rows)
    
    ~double4x4() {}
    // destructor

    double& Elem(int i, int j) { return A[i][j]; }
    // returns the (i,j)-th component; 1 <= i <= 4     1 <= j <= 4

    double Elem(int i, int j) const { return A[i][j]; }
    // returns the (i,j)-th component; 1 <= i <= 4     1 <= j <= 4

    double4 Row(int i) const;
    // returns the i-th row of the matrix (i = 0..3)

    double4 Col(int i) const;
    // returns the i-th column of the matrix (i = 0..3)

    const double* operator[](int i) const { return A[i]; }
    // returns a pointer to the i-th row of A
    // this makes it possible to access a variable B of type double4x4 via B[i][j]

    double* operator[](int i) { return A[i]; }
    // returns a pointer to the i-th row of A
    // this makes it possible to access a variable B of type double4x4 via B[i][j]

    double4x4 Transpose() const;
    // returns the transpose of the matrix
   
    double FrobeniusNorm() const;
    // returns the Frobenius norm of the matrix
    // (this norm is defined as
    //
    //                  3   3            2
    //          Sqrt[  Sum Sum  |A[i][j]|   ]
    //                 i=0 j=0
               
    double Det() const;
    // returns the determinant of the matrix

    void Fill(double fill);
    // makes all elements of the matrix equal to fill

    void FillDiag(double fill) { A[0][0] = A[1][1] = A[2][2] = fill; }
    // makes all elements (i,i) of the matrix equal to fill

    void MakeZero() { Fill(0); }
    // makes all elements of the matrix equal to 0

    void MakeUnit() { Fill(0); FillDiag(1); }
    // makes all (i,i) elements equal to 1 and the rest equal to 0

    void MakeSymmetric();
    // if the matrix is square every element A(i,j) is replaced with (A(i,j) + A(j,i))/2

    void SwapCols(int i, int j);
    // swaps the i-th and j-th column (i,j = 0..3)

    void SwapRows(int i, int j);
    // swaps the i-th and j-th row (i,j = 0..3)

    void MultCol(int i, double t);
    // multiplies the i-th column of the matrix by t  (i,j = 0..3)

    void MultRow(int i, double t);
    // multiplies the i-th row of the matrix by t (i,j = 0..3)

    // translation matrix over vector v
    double4x4& Transl(const double3&);

    // axial rotation matrix
    double4x4& AxRot(const double3&, double);

    // rotation matrix over angle phi round X|Y|Z-axis
    double4x4& Xrot(double);
    double4x4& Yrot(double);
    double4x4& Zrot(double);

    void operator  = (const double4x4&);
    void operator += (const double4x4&);
    void operator -= (const double4x4&);
    void operator *= (double t);
    void operator /= (double t);

    friend double4x4 operator - (const double4x4&);
    friend double4x4 operator + (const double4x4& A, const double4x4& B);
    friend double4x4 operator - (const double4x4& A, const double4x4& B);
    friend double4x4 operator * (double t, const double4x4& A);
    friend double4x4 operator * (const double4x4& A, double t);
    friend double4x4 operator / (const double4x4& A, double t);
    friend double4x4 operator * (const double4x4& A, const double4x4& B);
    friend double4 operator * (const double4x4& A, const double4& x);
    friend double4 operator * (const double4& x, const double4x4& A);
    friend double3 operator * (const double3& x, const double4x4& A);
    friend double3 operator * (const double4x4& A, const double3& x);

    friend std::ostream& operator<<(std::ostream& s, const double4x4& A);
    friend std::istream& operator>>(std::istream& s, double4x4& A);
};

double4x4 DiagonalMatrix(const double4& x);
// returns the diagonal matrix with x[i] as (i,i)-th element

double4x4 DyadeMatrix(const double4& x, const double4& y);
// returns the matrix        T
//                       x  y

//--- INLINE FUNCTIONS ------------------------------------------------//

inline double4 double4x4::Row(int i) const
{ return double4(A[i][0], A[i][1], A[i][2], A[i][3]); }

inline double4 double4x4::Col(int i) const
{ return double4(A[0][i], A[1][i], A[2][i], A[3][i]); }

inline double4x4 operator * (const double4x4& A, double t) { return t*A; }

inline double4 operator * (const double4x4& A, const double4& x)
{
  return double4( (A.Row(0),x), (A.Row(1),x), (A.Row(2),x), (A.Row(3),x) );
}

inline double4 operator * (const double4& x, const double4x4& A)
{
  return double4( (A.Col(0),x), (A.Col(1),x), (A.Col(2),x), (A.Col(3),x) );
}

inline double3 operator * (const double4x4& A, const double3& v)
{ double w = v[0]*A[3][0]+v[1]*A[3][1]+v[2]*A[3][2]+A[3][3];
  return double3((v[0]*A[0][0]+v[1]*A[0][1]+v[2]*A[0][2]+A[0][3])/w,
                (v[0]*A[1][0]+v[1]*A[1][1]+v[2]*A[1][2]+A[1][3])/w,
                (v[0]*A[2][0]+v[1]*A[2][1]+v[2]*A[2][2]+A[2][3])/w);
}

inline double3 operator * (const double3& x, const double4x4& A)
{ 
  return A*x;
}

inline double4x4& double4x4::Transl(const double3& v)
{
  A[0][0]=1.0; A[0][1]=0.0; A[0][2]=0.0; A[0][3]=v.x();
  A[1][0]=0.0; A[1][1]=1.0; A[1][2]=0.0; A[1][3]=v.y();
  A[2][0]=0.0; A[2][1]=0.0; A[2][2]=1.0; A[2][3]=v.z();
  A[3][0]=0.0; A[3][1]=0.0; A[3][2]=0.0; A[3][3]=1.0;

  return(*this);
}


inline double4x4& double4x4::Xrot(double phi)
{ double  cp=(double)cos((double)phi), sp=(double)sin((double)phi);

  A[0][0]= 1.0; A[0][1]= 0.0; A[0][2]= 0.0; A[0][3]=0.0;
  A[1][0]= 0.0; A[1][1]=  cp; A[1][2]=  sp; A[1][3]=0.0;
  A[2][0]= 0.0; A[2][1]= -sp; A[2][2]=  cp; A[2][3]=0.0;
  A[3][0]= 0.0; A[3][1]= 0.0; A[3][2]= 0.0; A[3][3]=1.0;

  return(*this);
}


inline double4x4& double4x4::Yrot(double phi)
{ double  cp=(double)cos((double)phi), sp=(double)sin((double)phi);

  A[0][0]=  cp; A[0][1]= 0.0; A[0][2]=  sp; A[0][3]=0.0;
  A[1][0]= 0.0; A[1][1]= 1.0; A[1][2]= 0.0; A[1][3]=0.0;
  A[2][0]= -sp; A[2][1]= 0.0; A[2][2]=  cp; A[2][3]=0.0;
  A[3][0]= 0.0; A[3][1]= 0.0; A[3][2]= 0.0; A[3][3]=1.0;

  return(*this);
}


inline double4x4& double4x4::Zrot(double phi)
{ double cp=(double)cos((double)phi), sp=(double)sin((double)phi);

  A[0][0]=  cp; A[0][1]=  sp; A[0][2]= 0.0; A[0][3]=0.0;
  A[1][0]= -sp; A[1][1]=  cp; A[1][2]= 0.0; A[1][3]=0.0;
  A[2][0]= 0.0; A[2][1]= 0.0; A[2][2]= 1.0; A[2][3]=0.0;
  A[3][0]= 0.0; A[3][1]= 0.0; A[3][2]= 0.0; A[3][3]=1.0;

  return(*this);
}

#endif
