// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 06-19-1996

#ifndef __DVECTOR_H
#define __DVECTOR_H

#include <iostream>
#include <cmath>
#include <string>

#include <ivector.h>
#include <double3.h>
extern "C" {
#include <matrix.h>
}

class DoubleMatrix;

//-----------------------------------------------------------------------//
//                          DoubleVector                                 //
//-----------------------------------------------------------------------//
// This class represents an m-dimensional vector x of doubles.

class DoubleVector
{
  private:
    VEC* x;

  protected:
    double& elem(int i) { return x->ve[i]; }
    double elem(int i) const { return x->ve[i]; }
    int size() const { return x->dim; }
    VEC* vec() const { return x; }

  public:
    DoubleVector() { x = v_get(0); }
    // constructor

    DoubleVector(int m) { x = v_get(m); }
    // constructor (creates an uninitialized vector of size m)

    DoubleVector(int m, double a) { x = v_get(m); Fill(a); }
    // constructor (creates the vector [a,a,...,a]

    DoubleVector(const DoubleVector& y) { x = v_copy(y.vec(), VNULL); }
    // copy constructor

    DoubleVector(VEC* x0): x(x0) {}
    // constructor

    DoubleVector(const DoubleVector& X, const DoubleVector& Y);
    // constructor (creates the concatenation vector X Y)

    DoubleVector(const DoubleVector& X, const DoubleVector& Y, const DoubleVector& Z);
    // constructor (creates a concatenation vector xyzxyz...)

    ~DoubleVector() { V_FREE(x); }
    // destructor

    int Size() const { return size(); }
    // returns the size of the vector

    double& operator()(int i) { return x->ve[i]; }
    // returns the i-th component; 0 <= i <= m-1

    double operator()(int i) const { return x->ve[i]; }
    // returns the i-th component; 0 <= i <= m-1

    double Min() const;
    // returns the minimum value of the components

    double Max() const;
    // returns the maximum value of the components

    double Sum() const;
    // returns the sum of the components

    double Median() const;
    // returns the median of the components

    double Norm1() const { return v_norm1(x); }
    // returns ||x||_1

    double Norm2() const { return v_norm2(x); }
    // returns ||x||_2 (the Euclidian norm)

    double Length2() const;
    // returns the square of the Euclidian norm of the vector

    double Length() const { return sqrt(Length2()); }
    // returns the Euclidian norm of the vector

    void Sort() { x = v_sort(x, PNULL); }
    // sorts the components of the vector

    DoubleVector House();
    // implementation of algorithm 5.1.1 of [Golub, p. 196]

    DoubleVector Permute(const IntVector& Sigma) const;
    // returns a permutation y of this vector x, that satisfies
    // y.Elem(i) = x.Elem(Sigma.Elem(i));

    DoubleVector PermuteInverse(const IntVector& Sigma) const;
    // returns a permutation y of this vector x, that satisfies
    // y.Elem(Sigma.Elem(i)) = x.Elem(i)

    void Split(int k, DoubleVector& y, DoubleVector& z) const;
    // returns two vectors y and z, where y contains the first k and z
    // contains the last m-k components of this vector

    void SplitXYZ(DoubleVector& x, DoubleVector& y, DoubleVector& z) const;
    // splits the concatenation vector into x, y and z-components

    DoubleVector SubVector(int i0, int i1) const;
    // returns the subvector with components i0 .. i1, where the indexing
    // of i0 and i1 starts at 0!!

    DoubleVector Concat(const DoubleVector&) const;
    // returns the concatenation of the 2 vectors

    void SetSize(int m) { x = v_resize(x,m); }
    // sets the size of the vector to m
    // N.B. the contents of the vector are lost??

    void Fill(double fill);
    // makes all elements of the vector equal to fill

    void FillRandom(double min, double max);
    // fills the vector with random elements in the interval [min,max]

    void Scale(double min, double max);
    // scales the components to the interval [min,max] using an affine transformation

    void Swap(int i, int j)
      { double temp = elem(i); elem(i) = elem(j); elem(j) = temp; }
    // swaps the i-th and j-th component (i = 0..m-1, j = 0..m-1)

    void MakeZero() { v_zero(x); }
    // makes all elements of the vector equal to 0

    void WriteToFile(std::string FileName) const;
    // writes: #components x[0] x[1] ... each on a new line
    // if the file already exists, nothing is done

    void operator  = (const DoubleVector& y) { x = v_copy(y.vec(), x); }
    void operator += (const DoubleVector& y) { v_add(x, y.vec(), x); }
    void operator -= (const DoubleVector& y) { v_sub(x, y.vec(), x); }
    void operator *= (double t)        { sv_mlt(t, x, x); }
    void operator /= (double t)        { sv_mlt(1/t, x, x); }

    friend DoubleVector operator - (const DoubleVector&);
    friend DoubleVector operator + (const DoubleVector&, const DoubleVector&);
    friend DoubleVector operator - (const DoubleVector&, const DoubleVector&);
    friend DoubleVector operator * (double t, const DoubleVector& x);
    friend DoubleVector operator * (const DoubleVector& x, double t);
    friend DoubleVector operator / (const DoubleVector&, double);
    friend double operator * (const DoubleVector&, const DoubleVector&);
    friend double operator , (const DoubleVector&, const DoubleVector&);
    // returns the inner product of x and y

    friend DoubleVector operator * (const DoubleMatrix& A, const DoubleVector& X);
    friend DoubleVector operator * (const DoubleVector& X, const DoubleMatrix& A);

    friend std::ostream& operator<<(std::ostream& s, const DoubleVector&);
    friend std::istream& operator>>(std::istream& s, DoubleVector&);

    friend class DoubleMatrix;
    friend class SparseMatrix;
    friend class DoubleBKPFact;
    friend class DoubleCHFact;
    friend class DoubleLDLFact;
    friend class DoubleLUFact;
    friend class DoubleQRFact;

    double& operator[](int i) { return x->ve[i]; }
    // returns the i-th component; 0 <= i <= m-1

    double operator[](int i) const { return x->ve[i]; }
    // returns the i-th component; 0 <= i <= m-1

    double& Elem(int i) { return x->ve[i-1]; }
    // returns the i-th component; 1 <= i <= m

    double Elem(int i) const { return x->ve[i-1]; }
    // returns the i-th component; 1 <= i <= m

    int NumRows() const { return size(); }
    // returns the size of the vector

    void AddDouble3(int row, const double3& X);
    // Adds X to the subvector [row..row+2]
};

#endif

