// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 06-04-1996

#ifndef __LINEAREQUATION_H
#define __LINEAREQUATION_H

// This module contains datatypes for (systems of) linear equations.

#include <iostream>
#include <cmath>
#include <ivector.h>
#include <dvector.h>
#include <dmatrix.h>
#include <listtmpl.h>
#include <geotyp.h>
#include <int4.h>
#include <numtyp.h>

//-----------------------------------------------------------------------//
//                          LinearEquation                               //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//                        Cx = d
//
// with C an (mxn)-matrix and d an m-vector, where m = NumRows() and n = NumCols().

class LinearEquation {
  public:
    virtual int NumRows() const = 0;
    // returns the number of rows of the matrix C

    virtual int NumCols() const = 0;
    // returns the number of columns of the matrix C

    virtual void AddToMatrix(DoubleMatrix& A, int row) const = 0;
    // adds the matrix C to the matrix A[row..row+m-1, 0..n-1]
    
    virtual void AddToVector(DoubleVector& b, int row) const = 0;
    // adds the vector d to the vector b[row..row+m-1]

    virtual void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const = 0;
    // adds the vector y = Cx to the vector result[row..row+m-1]
    
    virtual void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const = 0;
    // adds the vector y = x[row..row+m-1] C to the vector result

    void GetMatrix(DoubleMatrix&) const;
    // returns the matrix C

    void GetVector(DoubleVector&) const;
    // returns the vector d

    void Check(const DoubleVector& x) const;
    // checks if the equations are satisfied for x and reports this
    
    virtual void Show() const;
    // writes the contents to standard out
};

//-----------------------------------------------------------------------//
//                          PointConstraintEquation                      //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//             sum       a(i) P(i) = P,
//          i in index
//
// where P(i), 0 <= i < n is a set of 3D points.

class PointConstraintEquation: public LinearEquation {
  protected:
    int n;
    IntVector index;
    DoubleVector a;
    double3 P;
    
  public:
    PointConstraintEquation(int n0, const IntVector& i0, const DoubleVector& a0, const double3& P0):
      n(n0), index(i0), a(a0), P(P0) {}

    int NumRows() const;
    int NumCols() const;
    void AddToMatrix(DoubleMatrix& A, int row) const;
    void AddToVector(DoubleVector& b, int row) const;
    void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const;
    void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const;
};

//-----------------------------------------------------------------------//
//                          LineConstraintEquation                       //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//             sum       a(i) (P(i), n0) = d0,
//          i in index
//
//             sum       a(i) (P(i), n1) = d1,
//          i in index
//
// where P(i), 0 <= i < n is a set of 3D points
// and V0 = {(n0,x)=d0} and V1 = {(n1,x)=d1}

class LineConstraintEquation: public LinearEquation {
  protected:
    int n;
    IntVector index;
    DoubleVector a;
    Plane3D V0, V1;
    
  public:
    LineConstraintEquation(int N0, const IntVector& i0, const DoubleVector& a0, const Line3D& line):
      n(N0), index(i0), a(a0) { line.GetPlanes(V0, V1); }
      
    int NumRows() const;
    int NumCols() const;
    void AddToMatrix(DoubleMatrix& A, int row) const;
    void AddToVector(DoubleVector& b, int row) const;
    void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const;
    void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const;
};

//-----------------------------------------------------------------------//
//                          TangentConstraintEquation                    //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//             sum       a(i) (P(i)^P) = 0,
//          i in index
//
// where P(i), 0 <= i < n is a set of 3D points.
// and v0^v1 parallel to P

class TangentConstraintEquation: public LinearEquation {
  protected:
    int n;
    IntVector index;
    DoubleVector a;
    double3 v0,v1;

  public:
    TangentConstraintEquation(int n0, const IntVector& i0, const DoubleVector& a0, const double3& P0):
      n(n0), index(i0), a(a0) { P0.PerpendicularVectors(v0,v1); }

    int NumRows() const;
    int NumCols() const;
    void AddToMatrix(DoubleMatrix& A, int row) const;
    void AddToVector(DoubleVector& b, int row) const;
    void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const;
    void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const;
};

//-----------------------------------------------------------------------//
//                          NormalConstraintEquation                     //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//             sum       a(i) (P(i), normal) = 0,
//          i in index
//
//             sum       b(i) (P(i), normal) = 0,
//          i in index
//
// where P(i), 0 <= i < n is a set of 3D points
 
class NormalConstraintEquation: public LinearEquation {
  protected:
    int n;
    IntVector index;
    DoubleVector a, b;
    double3 normal;

  public:
    NormalConstraintEquation(int N0, const IntVector& i0, const DoubleVector& a0, const DoubleVector& b0, const double3& n0):
      n(N0), index(i0), a(a0), b(b0), normal(n0) {}
 
    int NumRows() const;
    int NumCols() const;
    void AddToMatrix(DoubleMatrix& A, int row) const;
    void AddToVector(DoubleVector& b, int row) const;
    void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const;
    void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const;
};

//-----------------------------------------------------------------------//
//                          PlaneConstraintEquation                      //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//             sum       a(i) (P(i), normal) = d
//          i in index
//
// where P(i), 0 <= i < n is a set of 3D points and plane is (normal,x)=d

class PlaneConstraintEquation: public LinearEquation {
  protected:
    int n;
    IntVector index;
    DoubleVector a;
    Plane3D plane;
    
  public:
    PlaneConstraintEquation(int n0, const IntVector& i0, const DoubleVector& a0, const Plane3D& p0):
      n(n0), index(i0), a(a0), plane(p0) {}
      
    int NumRows() const;
    int NumCols() const;
    void AddToMatrix(DoubleMatrix& A, int row) const;
    void AddToVector(DoubleVector& b, int row) const;
    void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const;
    void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const;
};

//-----------------------------------------------------------------------//
//                          ControlPointConstraintEquation               //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//                            P(index(i)) = points[i],   i in index
//
// where P(i), 0 <= i < n is a set of 3D points.
 
class ControlPointConstraintEquation: public LinearEquation {
  protected:
    int n;
    IntVector index;
    List<double3> points;
 
  public:
    ControlPointConstraintEquation(int n0, IntVector i0, List<double3> p0):
      n(n0), index(i0), points(p0) {}
 
    int NumRows() const;
    int NumCols() const;
    void AddToMatrix(DoubleMatrix& A, int row) const;
    void AddToVector(DoubleVector& b, int row) const;
    void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const;
    void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const;
};

//-----------------------------------------------------------------------//
//                          CurveConstraintEquation                      //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//            a.Count
//             sum       a(j) P(index[i][j]) = Q(i),     0 <= i < index.Count
//             j=0
//
// where P(i), 0 <= i < n is a set of 3D points.

class CurveConstraintEquation: public LinearEquation {
  protected:
    int n;   
    DoubleVector a;
    List<int4> index;
    List<double3> Q;
    
  public:
    CurveConstraintEquation(int n0, const DoubleVector& a0): n(n0), a(a0) {}

    void Add(const int4& i0, const double3& Q0) { index += new int4(i0); Q += new double3(Q0); }

    int NumRows() const;
    int NumCols() const;
    void AddToMatrix(DoubleMatrix& A, int row) const;
    void AddToVector(DoubleVector& b, int row) const;
    void AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const;
    void AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const;

    void Show() const;
};

//-----------------------------------------------------------------------//
//                          LinearEquationSystem                         //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//                        Cx = d
//
// with C an (mxn)-matrix and d an m-vector, where m = NumRows() and n = NumCols().
// The system is composed of a list of LinearEquations.

class LinearEquationSystem: public SparseMatrix {
  protected:
    List<LinearEquation> eqn;

  public:
    LinearEquationSystem() {}
    // default constructor

    LinearEquationSystem(List<LinearEquation>& eqs) { eqn = eqs; }
    // constructor

    int NumRows() const;
    int NumCols() const;

    void operator+= (LinearEquation* eq) { eqn += eq; }
    void operator+= (List<LinearEquation>& eqs) { eqn += eqs; }

    virtual void AddToMatrix(DoubleMatrix&) const;
    // adds the matrix to a DoubleMatrix

    void GetVector(DoubleVector&) const;
    // returns the vector d

    DoubleVector GetMatrixVectorProduct(const DoubleVector& x) const;
    // returns the vector x^T C

    DoubleVector GetVectorMatrixProduct(const DoubleVector& x) const;
    // returns the vector Cx

    void GetEliminateForm(DoubleMatrix& F, DoubleVector& g, IntVector& index) const;
    // Computes a matrix F and a vector G, such that the original system
    //
    //                          Cx = d
    // is equivalent to
    //                           x = Fy + g
    //
    // The vector index contains the indices of the remaining columns,
    // so y = (x(index(0)), x(index(1)), ...);

    void TestEliminateForm(const DoubleVector& x) const;
    // checks if F and g satisfy the following relations:
    //
    //           CF = 0   &&   Cg = d

    void Check(const DoubleVector& x) const;
    // checks if the equations are satisfied for x and reports this
    
    void Show() const;
    // writes the contents to standard out
};

//-----------------------------------------------------------------------//
//                          LinearSystem                                 //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//                     A x = b

class LinearSystem {
  public:
    DoubleMatrix A;
    DoubleVector b;

  public:
    LinearSystem(): A(0,0), b(0) {}
    ~LinearSystem() {}

    int GetPivotElement(int& row, int& col, const IntVector& done);
    // Searches for the element in the matrix A with maximum absolute value
    // The vector done has the following interpretation:
    //   if (done[i] >= 0), the element a(i,done[i]) has already been used
    //   as a pivot element, so this row will be skipped
    // If a nonzero pivot element is found true is returned.
    // The vector done is not updated!

    void ClearColumn(int row, int col);
    // by elementary row operations the column col is transformed into the
    // unitvector e[row]
    //
    // precondition: |a(row,col)| > 0

    int Clear(IntVector& done);
    // By elementary row operations and permutations of columns, the matrix
    // is transformed into a matrix that contains the m columns of the
    // m-dimensional unit matrix.
    //
    // postcondition:
    // The vector done contains the positions of these columns.
    //
    // if something goes wrong false is returned

    void TestClearColumn(int row, int col, const DoubleVector& x);
    // by elementary row operations the column col is transformed into the 
    // unitvector e[row] 
    //
    // precondition: |A(row,col)| > 0 

    int TestClear(IntVector& done, const DoubleVector& x);
    // Performs the same actions as Clear and prints inbetween results

    void TestEliminateForm(const DoubleVector& x);
    // Performs the same actions as GetSimpleForm and does some testing.

    void GetEliminateForm(DoubleMatrix& F, DoubleVector& G, IntVector& index);
    // Computes a matrix F and a vector G, such that the original system
    //
    //                          Ax = b
    // is equivalent to
    //                           x = Fy + G
    //
    // The vector index contains the indices of the remaining columns,
    // so y = (x(index(0)), x(index(1)), ...);
    
    void Show() const;
    // writes the contents to standard out
};

#endif

