// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Thu 06-13-1996

#ifndef __FUNCTION_H
#define __FUNCTION_H

#include <numtyp.h>
#include <tuple.h>
#include <dvector.h>
#include <dmatrix.h>

//-----------------------------------------------------------------------//
//                          Function                                     //
//-----------------------------------------------------------------------//
// The class Function is an abstract class that represents a function
// f: R^n -> R
//
// The abstract method dimension() should return the value n.
//
// There are virtual methods val(), grad() and hess() that return the
// function value, the gradient and the Hessian. These methods do not
// necessarily have to be implemented all for a descendant. The default
// implementations return 0, the 0-vector and the 0-matrix respectively.
//
// For efficiency reasons, some extra virtual methods are defined to
// compute inner products with the gradient and the Hessian directly.
// By default these methods are defined in terms of the functions grad()
// and hess().
//
// Some numerical methods are added to minimize the function f.

class Function {
  public:
    Function() {}
    // default constructor

    ~Function() {}
    // destructor

    virtual double val(const DoubleVector& x) const;
    // returns f(x)

    virtual DoubleVector grad(const DoubleVector& x) const;
    // returns grad f(x)

    virtual DoubleMatrix hess(const DoubleVector& x) const;
    // returns hess f(x)

    virtual void operator*=(double t);
    // multiply the function with a factor t

    virtual int dimension() const = 0;
    // returns n

    virtual double gradinp(const DoubleVector& x, const DoubleVector& d) const
      { return (grad(x),d); }
    // returns the inner product (grad(x),d)

    virtual DoubleVector hessmult(const DoubleVector& x, const DoubleVector& d) const
      { return hess(x)*d; }
    // returns the product hess(x)*d

    virtual double hessleftright(const DoubleVector& x, const DoubleVector& d, const DoubleVector& e) const
      { return (d, hess(x)*e); }
    // returns the value (d, hess(x)*e)

    double operator()(const DoubleVector& x) const { return val(x); }
    // returns f(x)
    
    int MinimizeConjugateGradient(DoubleVector& x, int imax=50,
       double epsCG=0.001, int jmax=5, double epsNewton=0.001);
    // Given a starting value x, a maximum number of CG iterations imax,
    // a CG error tolerance epsCG < 1, a maximum number of Newton-Raphson
    // iterations jmax, and a Newton-Raphson error tolerance epsNewton,
    // this method searches for a minimum of the function.
    // See ShewChuk, p.52
    // The algorithm terminates when the maximum number of iterations has
    // been exceeded (return value 0), or when 'r(i) <= eps r(0)'.

    int TestConjugateGradient(DoubleVector& x, int imax=50,
       double epsCG=0.001, int jmax=5, double epsNewton=0.001);
};

//-----------------------------------------------------------------------//
//                          DivFunction                                  //
//-----------------------------------------------------------------------//
// The class DivFunction represents the function g(x) = 1/f(x)

class DivFunction: public Function {
  protected: 
    Function* f;
    bool isowner;
    
  public:
    DivFunction(Function& f0, int owner=false): f(&f0), isowner(owner) {}
    // constructor

    DivFunction(Function* f0, int owner=false): f(f0), isowner(owner) {}
    // constructor

    DivFunction(DivFunction& func): f(func.f), isowner(0) {}
    // copy constructor

    ~DivFunction() { if (isowner) delete f; }
    // destructor

    double val(const DoubleVector& x) const { return 1/(*f).val(x); }
    DoubleVector grad(const DoubleVector& x) const;
    DoubleMatrix hess(const DoubleVector& x) const;
    int dimension() const { return f->dimension(); }
    void operator*=(double t) { f->operator *= (1/t); }
};

//-----------------------------------------------------------------------//
//                          SumFunction                                  //
//-----------------------------------------------------------------------//
// The class SumFunction represents the function h(x) = f(x) + g(x)

class SumFunction: public Function {
  public:
    Function* f;
    Function* g;
    bool isowner;

  public:
    SumFunction(Function& f0, Function& g0, int owner=false): f(&f0), g(&g0), isowner(owner) {}
    // default constructor

    double val(const DoubleVector& x) const
      { return (*f).val(x) + (*g).val(x); }
    DoubleVector grad(const DoubleVector& x) const;
    DoubleMatrix hess(const DoubleVector& x) const;
    int dimension() const { return (*f).dimension(); }
    void operator*=(double t) { (*f) *= t; (*g) *= t; }
};

//-----------------------------------------------------------------------//
//                          CompositeFunction                            //
//-----------------------------------------------------------------------//
// This class represents the function g(x) = f(Dx+e),
//
// where g: R^n -> R,        D an (mxn)-matrix
//       f: R^m -> R         e an m-vector

class CompositeFunction: public Function {
  protected:
    Function* f;
    DoubleMatrix D;
    DoubleVector e;
    bool isowner;
    
  public:
    CompositeFunction(Function& f0, const DoubleMatrix& D0, const DoubleVector& e0, int owner=false):
       f(&f0), D(D0), e(e0), isowner(owner) {}
    // default constructor
    
    CompositeFunction(const CompositeFunction& g): f(g.f), D(g.D), e(g.e) {}
    // copy constructor

    ~CompositeFunction() {}
    // destructor

    double val(const DoubleVector& x) const
      { return (*f).val(D*x+e); }
    DoubleVector grad(const DoubleVector& x) const
      { return (*f).grad(D*x+e)*D; }
    DoubleMatrix hess(const DoubleVector& x) const
      { return D.Transpose()*(*f).hess(D*x+e)*D; }
    double hessleftright(const DoubleVector& x, const DoubleVector& d0, const DoubleVector& e0) const
      { return (*f).hessleftright(D*x+e, D*d0, D*e0); }

    int dimension() const { return D.NumRows(); }
    void operator*=(double t) { (*f) *= t; }

    friend std::ostream& operator<<(std::ostream&, const CompositeFunction&);
    friend std::istream& operator>>(std::istream&, CompositeFunction&);
};

//-----------------------------------------------------------------------//
//                          FunctionList                                 //
//-----------------------------------------------------------------------//
// The class FunctionList represents a function f that is the sum of a
// sequence of functions f[i]:   f(x) = f[0](x) + f[1](x) + ...

class FunctionList: public Function {
  protected:
    List<Function> f;

  public:
    FunctionList() {}
    // default constructor

    FunctionList(int ALimit, int ADelta, int owner=false): f(ALimit, ADelta, owner) {}
    // constructor

    FunctionList(const FunctionList&);
    // copy constructor

    ~FunctionList() {}
    // destructor

    double val(const DoubleVector& x) const;
    DoubleVector grad(const DoubleVector& x) const;
    DoubleMatrix hess(const DoubleVector& x) const;
    double hessleftright(const DoubleVector& x, const DoubleVector& d, const DoubleVector& e) const;
    int dimension() const;

    void operator+= (Function* g) { f.Insert(g); }
    void operator+= (FunctionList& g);
    void operator*= (double t);

    Function& operator[] (int i) { return f[i]; }
    int Size() { return f.Count; }
    void DeleteAll() { f.DeleteAll(); }

    friend std::ostream& operator<<(std::ostream&, const FunctionList&);
    friend std::istream& operator>>(std::istream&, FunctionList&);
};

//-----------------------------------------------------------------------//
//                          Function1D                                   //
//-----------------------------------------------------------------------//
// This class represents a function f:R->R

class Function1D {
  public:
    Function1D() {}
    // default constructor

    virtual ~Function1D() {}
    // destructor

    virtual double operator()(double t) const = 0;
    // returns f(t)
 
    virtual double Deriv(double t) const = 0;
    // returns f'(t)
  
    virtual double DDeriv(double t) const = 0;
    // returns f''(t)

    void TestTaylor(double t, double h=0.1, int n=5) const;
    // this method computes
    //
    //         1)   f(t+h) - f(t)                               =  O(h)
    //         2)   f(t+h) - f(t) - h f'(t)                     =  O(h^2)
    //         3)   f(t+h) - f(t) - h f'(t) - (h^2)/2 f''(t)    =  O(h^3)
    //
    // for a number of values h. This can be used to verify if the functions
    // operator(), Deriv() and DDeriv() are implemented correctly.
    
    int MinimizeNewton(double& t, double eps, int maxiter, double a, double b) const;
    // minimizes f with a Newton method;
    //   t       = start value
    //   eps     = stop criterion: |f'(t)| < eps
    //   maxiter = maximum number of iterations
    //   [a,b]   = interval in which the solution is sought
    //
    // returns true if the minimization converges; in that case, t is set to the minimum
};

//-----------------------------------------------------------------------//
//                          Function3D
//-----------------------------------------------------------------------//
// This class represents a function f:R->R^3

class Function3D {
  public:
    Function3D() {}
    // default constructor

    virtual ~Function3D() {}
    // destructor

    virtual double3 operator()(double t) const = 0;
    // returns f(t)
 
    virtual double3 Deriv(double t) const;
    // returns f'(t)
  
    virtual double3 DDeriv(double t) const;
    // returns f''(t)

    void TestTaylor(double t, double h=0.1, int n=5) const;
    // this method computes
    //
    //         1)   f(t+h) - f(t)                               =  O(h)
    //         2)   f(t+h) - f(t) - h f'(t)                     =  O(h^2)
    //         3)   f(t+h) - f(t) - h f'(t) - (h^2)/2 f''(t)    =  O(h^3)
    //
    // for a number of values h. This can be used to verify if the functions
    // operator(), Deriv() and DDeriv() are implemented correctly.   
};

//-----------------------------------------------------------------------//
//                          PiecewiseFunction3D                          //
//-----------------------------------------------------------------------//
// This class represents the function f:R->R^3 with
//
//      f(t) = a0 + a1 t    (t<=t0)
//             b0 + b1 t    (t>t0)
//
 
class PiecewiseFunction3D: public Function3D {
  protected:
    double3 a0, a1, b0, b1;
    double t0;

  public:
    PiecewiseFunction3D() {}
    // default constructor
 
    PiecewiseFunction3D(double tleft, const double3& left, double tmiddle,
                        const double3& middle, double tright, const double3& right);
    // constructor
 
    double3 operator()(double t) const;
    // returns f(t)
};

#endif

