// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Mon 06-17-1996

#ifndef __GEOTYP_H
#define __GEOTYP_H

#include <cmath>
#include <double2.h>
#include <double2x2.h>
#include <double3.h>
#include <double3x3.h>
#include <listtmpl.h>

class Triangle2D;

//-----------------------------------------------------------------------//
//                          HyperPlane                                   //
//-----------------------------------------------------------------------//
// The template HyperPlane<T> represents the HyperPlane with (T,x)=d

template <class T>
class HyperPlane {
  public:
    T       n;
    double  d;

    HyperPlane(const T& n0, double d0): n(n0), d(d0) {}
    // constructor

    HyperPlane(const HyperPlane<T>& hp): n(hp.n), d(hp.d) {}
    // copy constructor

    void operator=(const HyperPlane<T>& hp) { n=hp.n; d=hp.d; }
    // assignment operator

    double Value(const T& x) const { return (n,x)-d; }
      // return value of implicit hyperplane function.
      // if |n|=1 this equals SignedDistance

    double SignedDistance(const T& x) const { return ((n,x)-d)/sqrt((n,n)); }
    // returns the signed distance between x and V

    double Distance(const T& x) const { return fabs(SignedDistance(x)); }
    // returns the distance between x and V

    T Project(const T& x) const { return x + ((d - (n,x))/(n,n))*n; }
    // returns the projection of x in V
};

//-----------------------------------------------------------------------//
//                          Line2D                                       //
//-----------------------------------------------------------------------//
// The class Line2D represents the line { x in R^2 | x = r + t s }.

class Line2D : public HyperPlane<double2> {
  protected:
    double2 r;
    double2 s;

  public:
    Line2D(const double2& r0, const double2& s0)
      :HyperPlane<double2>(double2(s0[1],-s0[0]),s0[1]*r0[0]-s0[0]*r0[1])
      ,r(r0), s(s0) {}

    Line2D(const Line2D& line)
      :HyperPlane<double2>(line) { r = line.r; s = line.s; }
    // copy constructor

    Line2D(const double2& n0, const double d0)
      :HyperPlane<double2>(n0,d0)
    { r=HyperPlane<double2>::Project(double2(0,0));
      s=double2(n[1],-n[0]);
    }
    // calling this constructor means that no startpoint is preferred.
    // the startpoint choosen is the projection of (0,0) on the line.
    // as direction a vector, orthogonal to n is made, but the direction
    // is more or less arbitrary. The length of the direction vector is the
    // same as the length of the specified normal vector.

    double VerticalDistance(const double2& point) const;
    // returns the 'vertical' distance of point

    void operator= (const Line2D& line)
    { HyperPlane<double2>::n=line.HyperPlane<double2>::n;
      HyperPlane<double2>::d=line.HyperPlane<double2>::d;
      r = line.r; s = line.s; }
    // assignment operator

    double2 Evaluate(double t) const { return r + t*s; }
    // returns the point with parameter t

    int IntersectLine(const double2& A, const double2& B, double& t) const;
    // computes the intersection of this line with the line AB.
    // if the lines are (nearly) parallel, false is returned, else t
    // contains the parameter of the intersection point

    int IntersectLineSegment(const double2& A, const double2& B, double& t) const;
    // computes the intersection of this line with the line segment AB.
    // if the lines are (nearly) parallel, false is returned, else t
    // contains the parameter of the intersection point
};

//-----------------------------------------------------------------------//
//                          LineSegment2D                                //
//-----------------------------------------------------------------------//
// This class represents the line segment { x in R^2 | x = (1-t)*A + t B, 0 <= t <= 1 }.

class LineSegment2D {
  protected:
    double2 A;
    double2 B;

  public:
    LineSegment2D(const double2& a, const double2& b): A(a), B(b) {}

    double Distance(const double2& x) const;
    // returns the distance between this linesegment and the point x.
};

class Plane3D;

//-----------------------------------------------------------------------//
//                          Line3D                                       //
//-----------------------------------------------------------------------//
// The class Line3D represents the line l = { x in R^3 | x = r + t s }.

class Line3D {
  protected:
    double3 r;
    double3 s;

  public:
    Line3D(): r(0,0,0), s(1,0,0) {}

    Line3D(const double3& r0, const double3& s0): r(r0), s(s0) {}
    // constructor

    Line3D(const Line3D& line) { r = line.r; s = line.s; }
    // copy constructor

    void operator= (const Line3D& line) { r = line.r; s = line.s; }
    // assignment operator

    double3 Evaluate(double t) const { return r + t*s; }
    // returns the point with parameter t

    double3 operator()(double t) const { return r + t*s; }
    // returns the point with parameter t

    void GetPlanes(Plane3D& V0, Plane3D& V1) const;
    // returns two planes V0 and V1, such that this line is the intersection
    // of V0 and V1
    
    void GetDistanceMap(double3x3& A, double3& b) const;
    // computes a mapping x -> Ax + b such that ||Ax + b|| is the distance
    // between x and l

    double Distance(const double3& p, double& tmin) const;
    // returns the distance between this line and the point p.
    // tmin is the parameter of the point with minimal distance.

    double Distance(const double3& p) const
      { double t; return Distance(p,t); }
    // returns the distance between this line and the point p.

    double GetParameter(const double3& p) const
      { double t, dist; dist = Distance(p,t); return t; }
    // returns the parameter of the point on l with minimal distance to p

    double Distance(const Line3D& m, double& tmin) const;
    // returns the distance between the lines l and m.
    // tmin is the parameter of the point with minimal distance.
    // If the lines are (nearly) parallel, tmin is set to 0.

    double Distance(const Line3D& m) const
      { double t; return Distance(m,t); }
    // returns the distance between the lines l and m.

    double3 Project(const double3& p) const
    // returns the point on l with minimal distance to p
      { double t, dist; dist = Distance(p,t); return (*this)(t); }

    double3 Rotate(const double3& p, double angle);
    // returns the rotation of the point p along this line
    // over the angle 'angle', where angle is measured in radians

    double3 Direction() const { return s.Unit(); }
    const double3& R() const { return r; }
    const double3& S() const { return s; }

    friend class LineSegment3D;
    friend class Triangle3D;
    friend class Plane3D;

    friend std::ostream& operator<<(std::ostream& os, const Line3D& line);
    friend std::istream& operator>>(std::istream& is, Line3D& line);
};

inline std::ostream& operator<<(std::ostream& os, const Line3D& line)
{
  return os << line.r << " " << line.s << std::endl;
}

inline std::istream& operator>>(std::istream& is, Line3D& line)
{
  return is >> line.r >> line.s;
}

//-----------------------------------------------------------------------//
//                          LineSegment3D                                //
//-----------------------------------------------------------------------//
// This class represents the line segment { x in R^3 | x = (1-t)*A + t B, 0 <= t <= 1 }.

class LineSegment3D {
  public:
    double3 A;
    double3 B;

  public:
    LineSegment3D(const double3& a, const double3& b): A(a), B(b) {}

    double3 Evaluate(double t) const { return (1-t)*A + t*B; }
    // returns the point with parameter t

    double Distance(const Line3D& l, double& tmin) const;
    // returns the distance between this linesegment and the line l.
    // tmin is the parameter of the point with minimal distance.
    // If the lines are (nearly) parallel, tmin is set to 0.5.
};

//-----------------------------------------------------------------------//
//                          Plane3D                                      //
//-----------------------------------------------------------------------//
// The class Plane3D represents the plane V = { x in R^3 | n^x = d }.

class Plane3D: public HyperPlane<double3>
{
  public:
    Plane3D(): HyperPlane<double3>(double3(1,0,0), 0) {}

    Plane3D(const double3& normal, const double3& point):
      HyperPlane<double3>(normal, (normal,point)) {}
    // constructor

    Plane3D(const double3& normal, double d0): HyperPlane<double3>(normal,d0) {}
    // constructor

    Plane3D(const Plane3D& plane): HyperPlane<double3>(plane.n,plane.d) {}
    // copy constructor

    Plane3D(const double3& P0, const double3& P1, const double3& P2):
      HyperPlane<double3>((P2-P0)^(P1-P0), ((P2-P0)^(P1-P0),P0)) {}

    double3 ProjectParallel(const double3& x, const double3& dir)
      {   return x + ((d - (n,x))/(n,dir))*dir; }
    // returns the projection of the point x in this plane along the direction dir

    void GetParallelProjectionMap(const double3& dir, double3x3& A, double3& b);
    // The parallel projection along the direction dir is a linear map of the form
    //
    //                x -> Ax + b.
    //
    // This method returns the matrix A and the vector b.

    int IntersectLine(const Line3D& line, double& t) const;
    // computes the intersection of this plane with the Line3D line.
    // if the line and the plane are (nearly) parallel, false is returned,
    // else t contains the parameter of the intersection point

    void Set(const double3& n0, double d0) { n=n0; d=d0; }
    // set parameters

    friend std::ostream& operator<<(std::ostream& os, const Plane3D& plane);
    friend std::istream& operator>>(std::istream& is, Plane3D& plane);
};

inline std::ostream& operator<<(std::ostream& os, const Plane3D& plane)
{
  return os << plane.n << " " << plane.d << std::endl;
}

inline std::istream& operator>>(std::istream& is, Plane3D& plane)
{
  return is >> plane.n >> plane.d;
}

//-----------------------------------------------------------------------//
//                          Rectangle2D                                  //
//-----------------------------------------------------------------------//
// The class Rectangle2D represents a 2D rectangle.

class Rectangle2D {
  protected:
    double2 leftdown;
    double width, height;

  public:
    Rectangle2D(): leftdown(0,0), width(1), height(1) {}
    // default constructor
    
    Rectangle2D(const double2& ld, double w, double h): leftdown(ld), width(w), height(h) {}
    // constructor
    
    Rectangle2D(double left, double right, double bottom, double top):
      leftdown(left,bottom), width(right-left), height(top-bottom) {}
    // constructor

    int ContainsPoint(const double2& x, double EPS=1e-12) const { return 0; }
    // Returns true if the point x is contained in the rectangle or if the
    // distance to the boundary < EPS.

    double2 Center() const { return leftdown + 0.5*double2(width, height); }
    // Returns the 'zwaartepunt' of the rectangle.

    double BoundaryDistance(const double2& x) const { return -1; }
    // returns the distance of the point x to the boundary of this rectangle

    void MakeCube();
    // sets the coordinates to (-1,-1), (1,-1), (1,1) and (-1,1)

    void GetRectangleToCubeMap(double2x2& A, double2& b) const;
    // computes a linear mapping x -> A x + b such that 
    //                                                  
    //              A r0 + b = (-1,-1)                                     
    //              A r1 + b = ( 1,-1)                                     
    //              A r2 + b = ( 1, 1) 
    //              A r3 + b = (-1, 1)
    //
    // where r0 ... r3 are the corners of this rectangle

    void GetCubeToRectangleMap(double2x2& A, double2& b) const;
    // computes a linear mapping x -> A x + b such that 
    //                                                  
    //              A (-1,-1) + b = r0
    //              A ( 1,-1) + b = r1
    //              A ( 1, 1) + b = r2
    //              A (-1, 1) + b = r3
    //
    // where r0 ... r3 are the corners of this rectangle

    void GetSquareToRectangleMap(double2x2& A, double2& b) const;
    // computes a linear mapping x -> A x + b such that
    //
    //              A (0,0) + b = r0
    //              A (1,0) + b = r1
    //              A (1,1) + b = r2
    //              A (0,1) + b = r3
    //
    // where r0 ... r3 are the corners of this rectangle
};

//-----------------------------------------------------------------------//
//                          Triangle2D                                   //
//-----------------------------------------------------------------------//
// The class Triangle2D represents a 2D triangle. N.B. it is assumed that  //
// the triangle is non-degenerate, so if this is not the case, some of the //
// methods will give incorrect results.                                    //
//                                                                         //
//                        B                                                //
//                       / \                                               //
//                    0 /   \ 1                                            //
//                     /     \                                             //
//                    /       \                                            //
//                   A---------C                                           //
//                        2                                                //

class Triangle2D {
  protected:
    double2 A, B, C;

  public:
    Triangle2D(): A(0,0), B(1,0), C(0,1) {}

    Triangle2D(const double2& a, const double2& b, const double2& c): A(a), B(b), C(c) {}

    double2& operator[](int i);
    // Returns the i-th vertex: i=0 -> A; i=1 -> B; i=2 -> C

    double2 operator[](int i) const;
    // Returns the i-th vertex: i=0 -> A; i=1 -> B; i=2 -> C

    double Area() const;
    // returns the area of the triangle

    int ContainsPoint(const double2& x, double EPS=1e-12) const;
    // Returns true if the point x is contained in the triangle or if the
    // distance to the boundary < EPS.

    double2 Center() const { return (A + B + C)/3; }
    // Returns the 'zwaartepunt' of the triangle.

    double BoundaryDistance(const double2& x) const;
    // returns the distance of the point x to the boundary of this triangle

    void GetBarycentricCoords(const double2& x, double& u, double& v) const;
    // This method computes the values u and v such that
    //
    //    x = u A + v B + (1-u-v) C

    void GetBarycentricMap(double2x2& AA, double2& b) const;
    // computes a linear mapping x -> A x + b such that 
    //
    //    (u)   =   A (x) + b
    //    (v)         (y)
    //
    // are the first two barycentric coordinates of (x,y) with respect
    // to the triangle

    void GetTriangleToSimplexMap(double2x2& A, double2& b) const;
    // computes a linear mapping x -> A x + b such that 
    //                                                  
    //              A u + b = (0,0)                                     
    //              A v + b = (1,0)                                     
    //              A w + b = (0,1) 
    //
    // where u, v and w are the corners of this triangle (A, B and C)

    void GetSimplexToTriangleMap(double2x2& A, double2& b) const;
    // computes a linear mapping x -> A x + b such that 
    //                                                  
    //              A (0,0) + b = u
    //              A (1,0) + b = v
    //              A (0,1) + b = w
    //
    // where u, v and w are the corners of this triangle (A, B and C)

    void MakeEquiLateral();
    // sets the coordinates of A,B,C to (0,0), (1,0) and (1/2, 1/2 sqrt(3))

    void MakeSimplex();
    // sets the coordinates of A,B,C to (0,0), (1,0) and (0,1)

    int IntersectEdge(int edge, const Line2D& line, double& t) const;
    // This method computes the intersection between a Line2D line and
    // edge 'edge' (edge = 0..2). If an intersection point is found, true
    // is returned and t contains the parameter of the intersection value.

    List<double2> GetGridPoints(int n);
    // computes a regular grid of (n+1)(n+2)/2 points on the triangle

    List<int3> GetGridTriangles(int n);
    // computes indices for the triangles of the grid that is obtained
    // from the method GetGridPoints

    Triangle2D Shrink(double x) {
      double2 center = Center();
      return Triangle2D(A + x*(center-A), B + x*(center-B), C + x*(center-C));
    }
    // returns a triangle that is obtained by moving the corners towards
    // the center of the triangle
};

//-----------------------------------------------------------------------//
//                          Triangle3D                                   //
//-----------------------------------------------------------------------//
// The class Triangle3D represents a 3D triangle. N.B. it is assumed that   //
// the triangle is non-degenerate, so if this is not the case, some of the  //
// methods will give incorrect results.                                     //
//                                                                          //
//                        B                                                 //
//                       / \                                                //
//                    0 /   \ 1                                             //
//                     /     \                                              //
//                    /       \                                             //
//                   A---------C                                            //
//                        2                                                 //

class Triangle3D {
  protected:
    double3 A, B, C;

  public:
    Triangle3D(const double3& a, const double3& b, const double3& c): A(a), B(b), C(c) {}
    // constructor

    double BoundaryDistance(const Line3D& l) const;
    // returns the distance of the boundary of this triangle and the line l

    int IntersectLine(const Line3D& l) const;
    // returns true if this triangle and the line l intersect

    double Area() const;
    // returns the area of the triangle
};

#endif

