// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#ifndef __INT3_H
#define __INT3_H

#include <cmath>
#include <iostream>
#include <mymanip.h>

//-----------------------------------------------------------------------//
//                          int3                                         //
//-----------------------------------------------------------------------//
// this struct represents a vector containing 3 integers

struct int3
{
  protected:
    int X[3];

  public:
    // constructors
    int3();
    int3(int, int, int);
    int3(const int3&);

    int Elem(int i) const { return X[i-1]; }
    // returns the i-th component (i = 1..3)

    int& Elem(int i) { return X[i-1]; }
    // returns a reference to the i-th component (i = 1..3)

    int operator[](int i) const { return X[i]; }
    // returns the i-th component (i = 0..2)

    int& operator[](int i) { return X[i]; }
    // returns a reference to the i-th component (i = 0..2)

    int& x() { return X[0]; }
    int& y() { return X[1]; }
    int& z() { return X[2]; }
    int x() const { return X[0]; }
    int y() const { return X[1]; }
    int z() const { return X[2]; }

    int Length2() const;
    // returns the squared length of the vector

    double Length() const;
    // returns the length of the vector

    double Dist(const int3& p) const;
    // returns the Euclidian distance from the point X to p

    int3 Unit() const;
    // returns the vector X / Length(X)

    void MakeZero();
    // makes the vector equal to (0,0,0)

    bool IsZero() const;
    // tests if the vector is equal to (0,0,0)

    void Negate();
    // replaces the vector with -X

    void  operator = (const int3&);
    int3  operator - () const;
    int3& operator += (const int3&);
    int3& operator -= (const int3&);
    int3& operator *= (int);

    friend int3 operator + (const int3&, const int3&);
    friend int3 operator - (const int3&, const int3&);
    friend int3 operator * (const int3&, int);
    friend int3 operator * (int, const int3&);
    friend int3 operator / (const int3&, int);
    friend int operator * (const int3&, const int3&); // inner product
    friend int operator , (const int3&, const int3&); // inner product
    friend int3 Cross (const int3&, const int3&);     // cross product
    friend int3 operator^(const int3&, const int3&);  // cross product
    friend int operator == (const int3&, const int3&);
    friend int operator != (const int3&, const int3&);
    friend double Dist(const int3&, const int3&);
    friend std::ostream& operator<<(std::ostream& s, const int3& p);
    friend std::istream& operator>>(std::istream& s, int3& p);
};

inline int3::int3 ()
{ X[0]=X[1]=X[2]=0; }

inline int3::int3 (int x0, int x1, int x2)
{ X[0]=x0; X[1]=x1; X[2]=x2; }

inline int3::int3 (const int3& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; }

inline void int3::operator = (const int3& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; }

inline int3 int3::operator - () const
{ return int3(-X[0], -X[1], -X[2]); }

inline int3& int3::operator += (const int3& p)
{ X[0]+=p[0]; X[1]+=p[1]; X[2]+=p[2]; return *this; }

inline int3& int3::operator -= (const int3& p)
{ X[0]-=p[0]; X[1]-=p[1]; X[2]-=p[2]; return *this; }

inline int3& int3::operator *= (int t)
{ X[0]*=t; X[1]*=t; X[2]*=t; return *this; }

inline int int3::Length2() const
{ return X[0]*X[0] + X[1]*X[1] + X[2]*X[2]; }

inline double int3::Length () const
{ return sqrt( (*this).Length2() ); }

inline int3 operator + (const int3& a, const int3& b)
{ return int3(a[0]+b[0], a[1]+b[1], a[2]+b[2]); }

inline int3 operator - (const int3& a, const int3& b)
{ return int3(a[0]-b[0], a[1]-b[1], a[2]-b[2]); }

inline double int3::Dist(const int3& p) const
{ return (*this - p).Length(); }

inline int3 operator * (const int3& a, int t)
{ return int3(t*a[0], t*a[1], t*a[2]); }

inline int3 operator * (int t, const int3& a)
{ return int3(t*a[0], t*a[1], t*a[2]); }

inline int operator * (const int3& a, const int3& b)
{ return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]; }

inline int operator , (const int3& a, const int3& b)
{ return a*b; }

inline int3 Cross(const int3& a, const int3& b)
{
  return int3(a[1]*b[2] - a[2]*b[1],
              a[2]*b[0] - a[0]*b[2],
              a[0]*b[1] - a[1]*b[0] );
}

inline int3 operator^(const int3& a, const int3& b)
{ return Cross(a,b); }

inline int operator == (const int3& a, const int3& b)
{ return (a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]); }

inline int operator != (const int3& a, const int3& b)
{ return !(a == b); }

inline double Dist(const int3& a, const int3& b)
{ return a.Dist(b); }

inline void int3::MakeZero()
{ X[0] = X[1] = X[2] = 0; }

inline bool int3::IsZero() const
{ return X[0] == 0 && X[1] == 0 && X[2] == 0; }

inline void int3::Negate()
{ X[0] = -X[0]; X[1] = -X[1]; X[2] = -X[2]; }

inline std::ostream& operator<<(std::ostream& s, const int3& a)
{ return s << '(' << a[0] << ',' << a[1] << ',' << a[2] << ')'; }

inline std::istream& operator>>(std::istream& s, int3& a)
{
  return s >> std::ws >> skipbracket >> a[0]
           >> skipcomma >> a[1]
           >> skipcomma >> a[2]
           >> skipbracket;
}

#endif

