// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#ifndef __INT4_H
#define __INT4_H

#include <cmath>
#include <iostream>
#include <mymanip.h>

//-----------------------------------------------------------------------//
//                          int4                                         //
//-----------------------------------------------------------------------//
// this struct represents a vector containing 4 integers

struct int4
{
  protected:
    int X[4];

  public:
    // constructors
    int4();
    int4(int, int, int, int);
    int4(const int4&);

    int Elem(int i) const { return X[i-1]; }
    // returns the i-th component (i = 1..4)

    int& Elem(int i) { return X[i-1]; }
    // returns a reference to the i-th component (i = 1..4)

    int operator[](int i) const { return X[i]; }
    // returns the i-th component (i = 0..3)

    int& operator[](int i) { return X[i]; }
    // returns a reference to the i-th component (i = 0..3)

    int& x() { return X[0]; }
    int& y() { return X[1]; }
    int& z() { return X[2]; }
    int& w() { return X[3]; }
    int x() const { return X[0]; }
    int y() const { return X[1]; }
    int z() const { return X[2]; }
    int w() const { return X[3]; }

    int Length2() const;
    // returns the squared length of the vector

    double Length() const;
    // returns the length of the vector

    double Dist(const int4& p) const;
    // returns the Euclidian distance from the point X to p

    int4 Unit() const;
    // returns the vector X / Length(X)

    void MakeZero();
    // makes the vector equal to (0,0,0,0)

    bool IsZero() const;
    // tests if the vector is equal to (0,0,0,0)

    void Negate();
    // replaces the vector with -X

    void  operator = (const int4&);
    int4  operator - () const;
    int4& operator += (const int4&);
    int4& operator -= (const int4&);
    int4& operator *= (int);

    friend int4 operator + (const int4&, const int4&);
    friend int4 operator - (const int4&, const int4&);
    friend int4 operator * (const int4&, int);
    friend int4 operator * (int, const int4&);
    friend int4 operator / (const int4&, int);
    friend int operator * (const int4&, const int4&); // inner product
    friend int operator , (const int4&, const int4&); // inner product
    friend int operator == (const int4&, const int4&);
    friend int operator != (const int4&, const int4&);
    friend double Dist(const int4&, const int4&);
    friend std::ostream& operator<<(std::ostream& s, const int4& p);
    friend std::istream& operator>>(std::istream& s, int4& p);
};

inline int4::int4 ()
{ X[0]=X[1]=X[2]=X[3]=0; }

inline int4::int4 (int x0, int x1, int x2, int x3)
{ X[0]=x0; X[1]=x1; X[2]=x2; X[3]=x3; }

inline int4::int4 (const int4& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; X[3]=p[3]; }

inline void int4::operator = (const int4& p)
{ X[0]=p[0]; X[1]=p[1]; X[2]=p[2]; X[3]=p[3]; }

inline int4 int4::operator - () const
{ return int4(-X[0], -X[1], -X[2], -X[3]); }

inline int4& int4::operator += (const int4& p)
{ X[0]+=p[0]; X[1]+=p[1]; X[2]+=p[2]; X[3]+=p[3]; return *this; }

inline int4& int4::operator -= (const int4& p)
{ X[0]-=p[0]; X[1]-=p[1]; X[2]-=p[2]; X[3]-=p[3]; return *this; }

inline int4& int4::operator *= (int t)
{ X[0]*=t; X[1]*=t; X[2]*=t; X[3]*=t; return *this; }

inline int int4::Length2() const
{ return X[0]*X[0] + X[1]*X[1] + X[2]*X[2] + X[3]*X[3]; }

inline double int4::Length () const
{ return sqrt( (*this).Length2() ); }

inline int4 operator + (const int4& a, const int4& b)
{ return int4(a[0]+b[0], a[1]+b[1], a[2]+b[2], a[3]+b[3]); }

inline int4 operator - (const int4& a, const int4& b)
{ return int4(a[0]-b[0], a[1]-b[1], a[2]-b[2], a[3]-b[3]); }

inline double int4::Dist(const int4& p) const
{ return (*this - p).Length(); }

inline int4 operator * (const int4& a, int t)
{ return int4(t*a[0], t*a[1], t*a[2], t*a[3]); }

inline int4 operator * (int t, const int4& a)
{ return int4(t*a[0], t*a[1], t*a[2], t*a[3]); }

inline int operator * (const int4& a, const int4& b)
{ return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3]; }

inline int operator , (const int4& a, const int4& b)
{ return a*b; }

inline int operator == (const int4& a, const int4& b)
{ return (a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]) && (a[3] == b[3]); }

inline int operator != (const int4& a, const int4& b)
{ return !(a == b); }

inline double Dist(const int4& a, const int4& b)
{ return a.Dist(b); }

inline void int4::MakeZero()
{ X[0] = X[1] = X[2] = X[3] = 0; }

inline bool int4::IsZero() const
{ return X[0] == 0 && X[1] == 0 && X[2] == 0 && X[3] == 0; }

inline void int4::Negate ()
{ X[0] = -X[0]; X[1] = -X[1]; X[2] = -X[2]; X[3] = -X[3]; }

inline std::ostream& operator<<(std::ostream& s, const int4& a)
{ return s << '(' << a[0] << ',' << a[1] << ',' << a[2] << ',' << a[3] << ')'; }

inline std::istream& operator>>(std::istream& s, int4& a)
{
  return s >> std::ws >> skipbracket >> a[0]
           >> skipcomma >> a[1]
           >> skipcomma >> a[2]
           >> skipcomma >> a[3]
           >> skipbracket;
}

#endif
