// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Mon 05-27-1996

#ifndef __INTEGR_H
#define __INTEGR_H

#include <double2.h>
#include <listtmpl.h>
#include <geotyp.h>
#include <polygon.h>

// This module contains coefficients (weights) and points (nodes) for
// some numerical integration formulas given in [Stroud].

struct IntPoint1D
{
  private:
    double weight;
    double point;
  public:
    IntPoint1D(double w, double p): weight(w), point(p) {}
    double Weight() const { return weight; }
    double Point() const { return point; }
};    

struct IntPoint2D
{
  private:
    double weight;
    double2 point;
  public:
    IntPoint2D(double w, const double2& p): weight(w), point(p) {}
    double Weight() const { return weight; }
    const double2& Point() const { return point; }
    double x() { return point.x(); }
    double y() { return point.y(); }
};    

typedef List<IntPoint1D> IntPoint1DList;
typedef List<IntPoint2D> IntPoint2DList;

class TIntPointEvaluator
{
  public:
    TIntPointEvaluator();
    // default constructor

    IntPoint1DList GetIntervalIntPoints(const double2& interval, int d);
    // computes a list of weights w_i and points x_i of a degree d
    // numerical integration formula
    //
    //	     -					 
    //	    |	 f(x) dx =~ Sum  w_i f(x_i)
    //	   -                 i
    //	interval
    //
    // valid degrees are 1, 3, 5, 7 and 9

    IntPoint2DList GetTriangleIntPoints(const Triangle2D& triangle, int d);
    // computes a list of weights w_i and points (x_i,y_i) of a degree d
    // numerical integration formula
    //
    //	   - -					 
    //	  | |	 f(x,y) dx dy =~ Sum  w_i f(x_i,y_i)
    //	 - -                      i
    //	triangle
    //
    // valid degrees are 2, 3, 5 and 7
    
    IntPoint2DList GetSimpleTriangleIntPoints(const Triangle2D& triangle, int level);
    // computes a list of weights w_i and points (x_i,y_i) of a degree d
    // numerical integration formula
    //
    //	   - -					 
    //	  | |	 f(x,y) dx dy =~ Sum  w_i f(x_i,y_i)
    //	 - -                      i
    //	triangle
    //
    // the integration points are arranged in a 'regelmatig' grid:
    //
    //                       x
    //                    x     x
    //                 x     x     x
    //              x     x     x     x
    //           x     x     x     x     x
    //
    // the number of integration points is level*(level+1)/2
    
    IntPoint2DList GetRectangleIntPoints(const Rectangle2D& rectangle, int d);
    // computes a list of weights w_i and points (x_i,y_i) of a degree d
    // numerical integration formula
    //
    //	   - -					 
    //	  | |	 f(x,y) dx dy =~ Sum  w_i f(x_i,y_i)
    //	 - -                      i
    //	rectangle
    //
    // valid degrees are 3, 5, 7 and 9

    IntPoint2DList GetPolygonIntPoints(const Polygon2D& polygon, int d);
    // computes a list of weights w_i and points (x_i,y_i) of a degree d
    // numerical integration formula
    //
    //	   - -					 
    //	  | |	 f(x,y) dx dy =~ Sum  w_i f(x_i,y_i)
    //	 - -                      i
    //	polygon
    //
    // valid degrees are 2, 3, 5 and 7
      
  private:
    IntPoint1DList Interval1_1;
    // contains weights and points for degree 1 numerical integration formula
  
    IntPoint1DList Interval3_1;
    // contains weights and points for degree 3 numerical integration formula
  
    IntPoint1DList Interval5_1;
    // contains weights and points for degree 5 numerical integration formula
  
    IntPoint1DList Interval7_1;
    // contains weights and points for degree 7 numerical integration formula
  
    IntPoint1DList Interval9_1;
    // contains weights and points for degree 9 numerical integration formula
  
    IntPoint2DList Simplex2_2;
    // contains weights and points for degree 2 numerical integration formula
  
    IntPoint2DList Simplex2_3;
    // contains weights and points for degree 3 numerical integration formula
    
    IntPoint2DList Simplex2_5;
    // contains weights and points for degree 5 numerical integration formula
    
    IntPoint2DList Simplex2_7;
    // contains weights and points for degree 7 numerical integration formula
    
    IntPoint2DList Cube3_1;
    // contains weights and points for degree 3 numerical integration formula
    
    IntPoint2DList Cube5_1;
    // contains weights and points for degree 5 numerical integration formula
    
    IntPoint2DList Cube7_1;
    // contains weights and points for degree 7 numerical integration formula
    
    IntPoint2DList Cube9_1;
    // contains weights and points for degree 9 numerical integration formula
    
    void InitIntervalPoints();
    void InitCubePoints();
    void InitSimplexPoints();
};

extern TIntPointEvaluator IntPointEvaluator;

#endif
