// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 12-13-1995

#ifndef __IVECTOR_H
#define __IVECTOR_H

#include <iostream>
#include <cmath>
extern "C" {
#include <matrix.h>
}

//-----------------------------------------------------------------------//
//                          IntVector                                    //
//-----------------------------------------------------------------------//
// The IntVector class is a m-dimensional vector of ints.
// The components of the vector can be retrieved in two ways:
//   numbered from 0..m-1, with the method operator[](int)
//   numbered from 0..m-1, with the method operator()(int)

class IntVector
{
  private:
    IVEC* x;

  protected:
    int& elem(int i) { return x->ive[i]; }
    int elem(int i) const { return x->ive[i]; }
    int size() const { return x->dim; }
    IVEC* vec() const { return x; }

  public:
    IntVector() { x = iv_get(0); }
    // constructor

    IntVector(int m) { x = iv_get(m); }
    // constructor (creates an uninitialized vector of size m)

    IntVector(const IntVector& y) { x = iv_copy(y.vec(), IVNULL); }
    // copy constructor

    IntVector(const IntVector& X, const IntVector& Y, const IntVector& Z);
    // constructor (creates a concatenation vector xyzxyz...)

    ~IntVector() { IV_FREE(x); }
    // destructor

    int NumRows() const { return size(); }
    // returns the size of the vector

    int Size() const { return size(); }
    // returns the size of the vector

    int Contains(int k) const;
    // returns true if k is an element of x

    int& operator()(int i) { return x->ive[i]; }
    // returns the i-th component; 0 <= i <= m-1

    int operator()(int i) const { return x->ive[i]; }
    // returns the i-th component; 0 <= i <= m-1

    int& operator[](int i) { return x->ive[i]; }
    // returns the i-th component; 0 <= i <= m-1

    int operator[](int i) const { return x->ive[i]; }
    // returns the i-th component; 0 <= i <= m-1

    int Length2() const;
    // returns the square of the Euclidian norm of the vector

    double Length() const { return sqrt(Length2()); }
    // returns the Euclidian norm of the vector

    void Sort() { x = iv_sort(x, PNULL); }
    // sorts the components of the vector

    void Split(int k, IntVector& y, IntVector& z) const;
    // returns two vectors y and z, where y contains the first k and z
    // contains the last m-k components of this vector

    IntVector SubVector(int i0, int i1) const;
    // returns the subvector with components i0 .. i1, where the indexing
    // of i0 and i1 starts at 0!!

    IntVector Concat(const IntVector&) const;
    // returns the concatenation of the 2 vectors

    void SetSize(int m) { x = iv_resize(x,m); }
    // sets the size of the vector to m
    // N.B. the contents of the vector are lost??

    void Fill(int fill);
    // makes all elements of the vector equal to fill

    void FillRandom(int min, int max);
    // fills the vector with random elements in the interval [min,max]

    void Swap(int i, int j)
      { int temp = elem(i); elem(i) = elem(j); elem(j) = temp; }
    // swaps the i-th and j-th component (i = 0..m-1, j = 0..m-1)

    void MakeZero() { iv_zero(x); }
    // makes all elements of the vector equal to 0

    void MakeSequence();
    // creates the vector (0,1,2,...)

    void operator  = (const IntVector& y) { x = iv_copy(y.vec(), x); }
    void operator += (const IntVector& y) { iv_add(x, y.vec(), x); }
    void operator -= (const IntVector& y) { iv_sub(x, y.vec(), x); }
    void operator *= (int t);

    friend IntVector operator - (const IntVector&);
    friend IntVector operator + (const IntVector&, const IntVector&);
    friend IntVector operator - (const IntVector&, const IntVector&);
    friend IntVector operator * (int t, const IntVector& x);
    friend IntVector operator * (const IntVector& x, int t);
    friend int operator * (const IntVector&, const IntVector&);
    friend int operator , (const IntVector&, const IntVector&);
    // returns the inner product of x and y
 
    friend std::ostream& operator<<(std::ostream& s, const IntVector&);
    friend std::istream& operator>>(std::istream& s, IntVector&);
};

#endif
