// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 17-7-95

#ifndef __NUMALG_H
#define __NUMALG_H

#include <numtyp.h>
#include <function.h>
#include <double2.h>
#include <iostream>

//-----------------------------------------------------------------------//
//                          ConjGradSolve                                //
//-----------------------------------------------------------------------//
// The function ConjGradSolve solves the equation
//
//                           f(x) = 0
//
// for an arbitrary function f: R^n -> R, using the nonlinear version of the
// conjugate gradient method. (See [Luenberger p. 252])

int ConjGradSolve(Function& f, DoubleVector& x0, DoubleVector& x, 
                  int MaxIter, double Eps);

// x0 is the startvector, MaxIter is the maximum number of iterations
// and the stopcriterion is
//
//                       |grad f(x)| < Eps
//
// If the function finds a solution, 1 is returned and the solution is x.

//-----------------------------------------------------------------------//
//                          ConjGradSolve                                //
//-----------------------------------------------------------------------//
// The function ConjGradSolve solves the linear system
//
//                             A x = b  
//
// using the linear version of the conjugate gradient method, where A is
// a positive definite symmetric matrix (See [Shewchuk])

int ConjGradSolve(SparseMatrix& A, DoubleVector& b, DoubleVector& x,
		  int imax, double eps);

// imax is the maximum number of iterations
// and the stopcriterion is
//                                 
//                           (r,r) < eps (r_0,r_0)
//
// with r the residu (i.e. r(x) = b - Ax).
//
// If the function finds a solution, 1 is returned and the solution is x.

//-----------------------------------------------------------------------//
//                          StedescSolve                                 //
//-----------------------------------------------------------------------//
// The function StedescSolve solves the linear system
//
//                             A x = b  
//
// using the linear version of the steepest descent method, where A is
// a positive definite symmetric matrix (See [Shewchuk])

int StedescSolve(SparseMatrix& A, DoubleVector& b, DoubleVector& x,
                 int imax, double eps);

// imax is the maximum number of iterations
// and the stopcriterion is
//                                 
//                           (r,r) < eps (r_0,r_0)
//
// with r the residu (i.e. r(x) = b - Ax).
//
// If the function finds a solution, 1 is returned and the solution is x.

//-----------------------------------------------------------------------//
//                          LUSolve                                      //
//-----------------------------------------------------------------------//
// The function LUSolve solves the linear system
//
//                             A x = b  
//
// using LU decomposition, where A is a positive definite symmetric matrix.

int LUSolve(const DoubleMatrix& A, const DoubleVector& b, DoubleVector& x);

//-----------------------------------------------------------------------//
//                          QuadMinLUSolve                               //
//-----------------------------------------------------------------------//
// The function QuadMinLUSolve is used to solve the constrained minimization
// problem
//
//           minimize     f(x) = x^T A x + b^T x
//           such that    Cx = d
//
// where A is a symmetric positive definite matrix.
//
// If the function finds a solution, 1 is returned and the solution is x.

int QuadMinLUSolve(SparseMatrix& A, DoubleVector& b, DoubleMatrix& C, DoubleVector& d, DoubleVector& x);

//-----------------------------------------------------------------------//
//                          RewriteEquations                             //
//-----------------------------------------------------------------------//
// The function RewriteEquations is used to write the equation
//
//                     Ax = b
//
// into the form
//
//                     x = Cz + d
//
// with A an mxn-matrix and C an nx(n-m)-matrix.
// The following relations hold: AC = 0 and Ad = b.
//
// (see also the routine gaussj from "Numerical Recipes in C")
//
// N.B. Afterwards, the matrix A and the vector b are changed!!
//
// First, Ax = b is rewritten as
//                            _
//                [ I(m) F ]  x  = g,     (*)
//
// where I(m) is the m-dimensional unit matrix, F is an m x (n-m) matrix,
//                   _
// g an m-vector and x = T x a permutation of the vector x, such that
//                      _
//                      x  =  x
//                       i     sigma(i)
//             _
// If we write x = [y z] then it follows from (*) that
//              _
//              x  =  [ -F     ]  z   +   [ g ]
//                    [ I(n-m) ]          [ 0 ]
//
// From this it follows that
//	           _
//	x = T(inv) x = T(inv) [ -F     ] z  +  T(inv) [ g ] = C z + d
//                        [	I(n-m) ]
//
// The first version of RewriteEquations returns the matrix C as a DoubleMatrix
// and the second version returns C as a SparseMatrix.

void RewriteEquations(DoubleMatrix& A, DoubleVector& b, DoubleMatrix& C, DoubleVector& d);
/*
void RewriteEquations(DoubleMatrix& A, DoubleVector& b, RewriteMatrix& C, DoubleVector& d);
*/

//-----------------------------------------------------------------------//
//                          GivensMatrix                                 //
//-----------------------------------------------------------------------//
// This function computes a GivensMatrix (See [Golub p.201]).

DoubleMatrix GivensMatrix(int n, int i, int k, double c, double s);

//-----------------------------------------------------------------------//
//                          FastGivens                                   //
//-----------------------------------------------------------------------//
// See [Golub p. 207]
// given x in R^2 and positive d in R^2 the following algorithm
// computes a 2-by-2 fast Givens transformation M such that the second
// component of M^T x is zero and M^T DM = D_1 is diagonal where
// D = diag(d_1, d_2).

void FastGivens(const double2& x, const double2& d, double& alpha, double& beta, int& type);

//-----------------------------------------------------------------------//
//                          FastGivensQR                                 //
//-----------------------------------------------------------------------//
// See [Golub p. 216]
// Given A in R[m,n] with m >= n, the following algorithm computes
// nonsingular M in R[m,m] and positive d(1:m) such that
// M^T A = T is upper triangular and M^T M = diag(d_1, ..., d_m)
// A is overwritten by T

void FastGivensQR(DoubleMatrix& A, DoubleVector& d);

//-----------------------------------------------------------------------//
//                          Givens                                       //
//-----------------------------------------------------------------------//
// See [Golub p. 202].
// given scalars a and b, this algorithm computes c = cos(theta) and s = sin(theta)
// so
//             T
//   [  c  s ]   [a]  =  [r]
//   [ -s  c ]   [b]     [0]

void Givens(double a, double b, double& c, double& s);

//-----------------------------------------------------------------------//
//                          GivensQR                                     //
//-----------------------------------------------------------------------//
// See [Golub p. 216].
// Given A in R[m,n] with m >= n, the following algorithm overwrites
// A with Q^T A = R, and B with Q^T, where R is upper triangular and
// Q is orthogonal

void GivensQR(DoubleMatrix& QT, DoubleMatrix& R);

#endif
