// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Fri 06-21-1996

#ifndef __NUMTYP_H
#define __NUMTYP_H

#include <iostream>
#include <fstream>
#include <cmath>
#include <ivector.h>
#include <dvector.h>
#include <dmatrix.h>
#include <listtmpl.h>

//-----------------------------------------------------------------------//
//                          SparseVector                                 //
//-----------------------------------------------------------------------//
// SparseVector is an abstract class that represents a double vector x.
// A descendant must take care of the internal representation of the
// components x[i].

class SparseVector
{
  public:
    SparseVector() {}
    // constructor
    
    virtual ~SparseVector() {}
    // destructor

    virtual void AddToVector(DoubleVector&) const = 0;
    // adds x to a DoubleVector

    virtual void MultAddToVector(double t, DoubleVector&) const = 0;
    // adds t*x to a DoubleVector

    void GetVector(DoubleVector&) const;
    // returns x (is implemented using AddToVector)

    virtual double GetInnerProduct(const DoubleVector&) const = 0;
    // return the inner product of x with a DoubleVector of the same size;
    // N.B. This method is used in 'operator,' , which is the operator for
    // the inner product.

    virtual void operator*= (double t) {}
    // multiplies the vector x by t

    virtual int NumRows() const = 0;
    // returns the size of the vector x

    virtual void Show() const { std::cout << "SparseVector::Show()" << std::endl; }
    // writes the contents of x to standard out

    virtual void Read(std::ifstream& from) { }
    // reads the vector x from a stream;
    // N.B. this method is used in the stream operator 'operator>>'.

    virtual void Write(std::ofstream& to) const { }
    // writes the vector x to a stream
    // N.B. this method is used in the stream operator 'operator<<'.

    void SelfTest() const;
    // this method tests if the virtual methods are implemented correctly

    friend double operator,(const SparseVector&, const DoubleVector&); // inner product
    friend double operator,(const DoubleVector&, const SparseVector&); // inner product
    friend std::ifstream& operator>>(std::ifstream&, SparseVector&);
    friend std::ofstream& operator<<(std::ofstream&, const SparseVector&);
};

//-----------------------------------------------------------------------//
//                          SparseDoubleVector                           //
//-----------------------------------------------------------------------//
// The class SparseDoubleVector is just a SparseVector representation
// of an ordinary DoubleVector.

class SparseDoubleVector: public SparseVector
{
  protected:
    DoubleVector& X;

  public:
    SparseDoubleVector(DoubleVector& X0): X(X0) {}
    void AddToVector(DoubleVector& Y) const { Y += X; }
    void MultAddToVector(double t, DoubleVector& Y) const { Y += t*X; }
    double GetInnerProduct(const DoubleVector& Y) const { return (X,Y); }
    void operator*= (double t) { X *= t; }
    int NumRows() const { return X.NumRows(); }
    void Show() const;
    void Read(std::ifstream& from);
    void Write(std::ofstream& to) const;
};

inline void SparseDoubleVector::Show() const { std::cout << X << std::endl; }
inline void SparseDoubleVector::Read(std::ifstream& from) { from >> X; }
inline void SparseDoubleVector::Write(std::ofstream& to) const { to << X; }

//-----------------------------------------------------------------------//
//                          IndexedVector                                //
//-----------------------------------------------------------------------//
// The class IndexedVector is a SparseVector x for which the nonzero elements
// are explicitly stored.
//
//         x(index(i)) = b(i)    0 <= i < b.Size() 
//                        0      otherwise

class IndexedVector: public SparseVector
{
  protected:
    int numrows;
    IntVector index;
    DoubleVector b;

  public:
    IndexedVector(int nr, int n): numrows(nr), index(n), b(n)
       { b.MakeZero(); index.MakeZero(); }

    int& Index(int i) { return index[i]; }
    double& Coeff(int i) { return b[i]; }
    
    void AddToVector(DoubleVector& Y) const;
    void MultAddToVector(double t, DoubleVector& Y) const;
    double GetInnerProduct(const DoubleVector& Y) const;
    void operator*= (double t) { b *= t; }
    int NumRows() const { return numrows; }
    void Show() const;
    void Read(std::ifstream& from);
    void Write(std::ofstream& to) const;
};

//-----------------------------------------------------------------------//
//                          QuadVector                                   //
//-----------------------------------------------------------------------//
// A QuadVector is a vector that corresponds to the following expression:
//                                    n-1               T
//  f(X) = f(P_0, ..., P_numrows-1) = sum  b[index(i)] P_index(i) P
//                                    i=0 

class QuadVector: public SparseVector
{
  protected:
    int numrows;
    IntVector index;
    DoubleVector b;
    double3 P;

  public:
    QuadVector(int nr, int n, const double3& P0): numrows(nr), index(n), b(n), P(P0)
       { b.MakeZero(); index.MakeZero(); }

    QuadVector(int nr, const IntVector& ind, const DoubleVector& b0, const double3& P0): numrows(nr), index(ind), b(b0), P(P0) {}
   
    int& Index(int i) { return index[i]; }
    double& Coeff(int i) { return b[i]; }

    void AddToVector(DoubleVector& Y) const;
    void MultAddToVector(double t, DoubleVector& Y) const;
    double GetInnerProduct(const DoubleVector& Y) const;
    void operator*= (double t) { P *= t; }
    int NumRows() const { return 3*numrows; }
    void Show() const;
    void Read(std::ifstream& from);
    void Write(std::ofstream& to) const;
};

//-----------------------------------------------------------------------//
//                          SparseMatrix                                 //
//-----------------------------------------------------------------------//
// SparseMatrix is an abstract double Matrix class. A descendant must take
// care of its internal representation of the components.

class SparseMatrix
{
  public:
    SparseMatrix() {}
    // constructor
    
    virtual ~SparseMatrix() {}
    // destructor

    virtual void AddToMatrix(DoubleMatrix&) const = 0;
    // adds the matrix to a DoubleMatrix

    void GetMatrix(DoubleMatrix&) const;
    // returns the actual matrix (is implemented using AddToMatrix)

    virtual DoubleVector GetVectorMatrixProduct(const DoubleVector&) const = 0;
    // computes the vector matrix product with a DoubleVector
    // NB This method is used in operator*(DoubleVector&, DoubleMatrix&)

    virtual DoubleVector GetMatrixVectorProduct(const DoubleVector&) const = 0;
    // computes the matrix vector product with a DoubleVector
    // NB This method is used in operator*(DoubleMatrix&, DoubleVector&)

    virtual void operator*= (double t) {}
    // multiplies the matrix by t

    virtual int NumRows() const = 0;
    // returns the number of rows

    virtual int NumCols() const = 0;
    // returns the number of columns

    virtual void Show() const { std::cout << "SparseMatrix::Show()" << std::endl; }
    // writes the contents to standard out

    virtual void Read(std::ifstream& from) { }
    // reads the matrix from a stream;
    // N.B. this method is used in the stream operator 'operator>>'.

    virtual void Write(std::ofstream& to) const { }
    // writes the matrix to a stream
    // N.B. this method is used in the stream operator 'operator<<'.

    void SelfTest() const;
    // this method tests if the virtual methods are implemented correctly

    void SolveBKP(const DoubleVector& b, DoubleVector& x) const;
    // Solves the system Ax=b using BKP factorization
    
    void SolveCH(const DoubleVector& b, DoubleVector& x) const;
    // Solves the system Ax=b using Cholesky factorization
    
    void SolveLDL(const DoubleVector& b, DoubleVector& x) const;
    // Solves the system Ax=b using LDL factorization
    
    void SolveLU(const DoubleVector& b, DoubleVector& x) const;
    // Solves the system Ax=b using LU factorization
    
    void SolveQR(const DoubleVector& b, DoubleVector& x) const;
    // Solves the system Ax=b using QR factorization
    
    int SolveSteepestDescent(const DoubleVector& b, DoubleVector& x, int imax, double eps) const;
    // Solves the system Ax=b using the steepest descent method
    
    int SolveConjugateGradient(const DoubleVector& b, DoubleVector& x, int imax, double eps) const;
    // Solves the system Ax=b using the conjugate gradient method
    
    int SolveConjugateResidual(const DoubleVector& b, DoubleVector& x, int imax, double eps) const;
    // Solves the system Ax=b using the conjugate residual method
    
    int SolveSTOD(const DoubleVector& b, DoubleVector& x, int imax, double eps) const;
    // Solves the system Ax=b using stabilized orthogonal directions

    friend DoubleVector operator*(const SparseMatrix& A, const DoubleVector& X);
    friend DoubleVector operator*(const DoubleVector& X, const SparseMatrix& A);
    friend std::ifstream& operator>>(std::ifstream&, SparseMatrix&);
    friend std::ofstream& operator<<(std::ofstream&, const SparseMatrix&);
};

//-----------------------------------------------------------------------//
//                          SparseDoubleMatrix                           //
//-----------------------------------------------------------------------//
// The class SparseDoubleMatrix is just a SparseMatrix representation
// of an ordinary DoubleMatrix.

class SparseDoubleMatrix: public SparseMatrix
{
  protected:
    DoubleMatrix& A;

  public:
    SparseDoubleMatrix(DoubleMatrix& A0): A(A0) {}
    void AddToMatrix(DoubleMatrix& B) const { B += A; }
    DoubleVector GetVectorMatrixProduct(const DoubleVector& X) const { return X*A; }
    DoubleVector GetMatrixVectorProduct(const DoubleVector& X) const { return X*A; }
    void operator*= (double t) { A *= t; }
    int NumRows() const { return A.NumRows(); }
    int NumCols() const { return A.NumCols(); }
};

//-----------------------------------------------------------------------//
//                          SparseMatrixList                             //
//-----------------------------------------------------------------------//
// The class SparseMatrixList represents the sum of a list of SparseMatrix's

class SparseMatrixList: public SparseMatrix
{
  protected:
    List<SparseMatrix> f;

  public:
    SparseMatrixList() {}

    void AddToMatrix(DoubleMatrix& B) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector& X) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector& X) const;
    void operator*= (double t);
    int NumRows() const;
    int NumCols() const;

    void operator+= (SparseMatrix* f0) { f += f0; }
};

//-----------------------------------------------------------------------//
//                          BlockMatrix                                  //
//-----------------------------------------------------------------------//
// A BlockMatrix is a square 3nx3n matrix A that corresponds to the following
// quadratic function:
//                              m-1 m-1                                 T
//  f(X) = f(P_0, ..., P_n-1) = sum sum M[index(i)][index(j)] P_index(i) P_index(j)
//                              i=0 j=0

class BlockMatrix: public SparseMatrix
{
  protected:
    int numrows;     // numrows=n
    IntVector index;
    DoubleMatrix M;

  public:
    BlockMatrix(int n, int nM): numrows(n), index(nM), M(nM,nM)
      { M.MakeZero(); index.MakeSequence(); }
    // constructor (N.B. after calling the constructor, M and index have to be set!)

    BlockMatrix(int n, const IntVector& ind, const DoubleMatrix& M0): numrows(n), index(ind), M(M0) {}
    // constructor
    
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector&) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector&) const;
    void operator*= (double t) { M *= t; }
    int NumRows() const { return 3*numrows; }
    int NumCols() const { return 3*numrows; }
    void Show() const;

    void SetIndices(const IntVector& ind) { index = ind; }
    void SetCoeff(int i, int j, double mij) { M(i,j) = mij; }
};

//-----------------------------------------------------------------------//
//                          SharedBlockMatrix                            //
//-----------------------------------------------------------------------//
// A SharedBlockMatrix is a square 3nx3n matrix A that corresponds to a list of
// QuadMatrices, that share the same matrix M
//

class SharedBlockMatrix: public SparseMatrix
{
  protected:
    int numrows;              // numrows=n
    List<IntVector> indices;
    DoubleMatrix M;

  public:
    SharedBlockMatrix(int n, int nM): numrows(n), M(nM,nM) { M.MakeZero(); }
    // constructor (N.B. after calling the constructor, M and indices have to be set!)

    SharedBlockMatrix(int n, const DoubleMatrix& M0): numrows(n), M(M0) {}
    // constructor
    
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector&) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector&) const;
    void operator*= (double t) { M *= t; }
    int NumRows() const { return 3*numrows; }
    int NumCols() const { return 3*numrows; }
    void Show() const;

    void AddBlock(const IntVector& index) { indices += new IntVector(index); }
    // a new block is added by specifying a vector of indices
    
    void SetCoeff(int i, int j, double mij) { M(i,j) = mij; }
};

//-----------------------------------------------------------------------//
//                          Quad2Matrix                                  //
//-----------------------------------------------------------------------//
// A Quad2Matrix is a square matrix that corresponds to the following
// quadratic function:
//                                    m-1  m-1                                 T         
//  f(X) = f(P_0, ..., P_numrows-1) = sum  sum M[index(i)][index(j)] P_index(i) A P_index(j)
//                                    i=0  j=0

class Quad2Matrix: public SparseMatrix
{
  protected:
    int numrows;
    IntVector index;
    DoubleMatrix M;
    double3x3 A;

  public:
    Quad2Matrix(int nr, int n, const double3x3& A0): numrows(nr), index(n), M(n,n), A(A0)
       { M.MakeZero(); index.MakeSequence(); }
    // constructor (N.B. after calling the constructor, M and index have to be set!)

    Quad2Matrix(int nr, const IntVector& ind, const DoubleMatrix& M0, const double3x3& A0): numrows(nr), index(ind), M(M0), A(A0) {}
    // constructor
    
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector&) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector&) const; 
    void operator*= (double t) { A *= t; }
    int NumRows() const { return 3*numrows; }
    int NumCols() const { return 3*numrows; }
    void Show() const;
    
    int& Index(int i) { return index[i]; }
    double& Coeff(int i, int j) { return M[i][j]; }
};

//-----------------------------------------------------------------------//
//                          LagrangeMatrix                               //
//-----------------------------------------------------------------------//
// The class LagrangeMatrix represents the matrix
//
//          | A   B^T |
//          |         |
//          | B    0  |
//
// with A an nxn-matrix and B an mxn-matrix. It is used for iteratively
// solving the quadratic programming problem
//
// minimize         1/2 x^T A x + b^T x + c
// subject to       Bx = d

class LagrangeMatrix: public SparseMatrix
{
  protected:
    SparseMatrix* A;
    SparseMatrix* B;

  public:
    LagrangeMatrix(SparseMatrix* A0, SparseMatrix* B0): A(A0), B(B0) {}
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector& X) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector& X) const;
    void operator*= (double t) { (*A) *= t; (*B) *= t; }
    int NumCols() const { return (*A).NumRows() + (*B).NumRows(); }
    int NumRows() const { return (*A).NumRows() + (*B).NumRows(); }
};

//-----------------------------------------------------------------------//
//                          PermutationMatrix                            //
//-----------------------------------------------------------------------//
// The class PermutationMatrix represents a permutation matrix A' of the
// matrix A such that the i-th column of A' is the sigma[i]-th column of A

class PermutationMatrix: public SparseMatrix
{
  protected:
    SparseMatrix& A;
    IntVector sigma;

  public:
    PermutationMatrix(SparseMatrix& A0, IntVector& sigma0): A(A0), sigma(sigma0) {}
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector& X) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector& X) const;
    void operator*= (double t) { A *= t; }
    int NumCols() const  { return A.NumCols(); }
    int NumRows() const { return A.NumRows(); }

    IntVector& Sigma()     { return sigma; }
};

//-----------------------------------------------------------------------//
//                          BTABMatrix                                   //
//-----------------------------------------------------------------------//
// The class BTABMatrix is a SparseMatrix representation of the matrixproduct
//            T
//           B  A  B

class BTABMatrix: public SparseMatrix
{
  protected:
    SparseMatrix& A;
    SparseMatrix& B;

  public:
    BTABMatrix(SparseMatrix& A0, SparseMatrix& B0): A(A0), B(B0) {}
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector& X) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector& X) const;
    void operator*= (double t) { A *= t; }
    int NumRows() const { return B.NumCols(); }
    int NumCols() const { return B.NumCols(); }
};

//-----------------------------------------------------------------------//
//                          BandMatrix                                   //
//-----------------------------------------------------------------------//
// The class BandMatrix is a SparseMatrix representation of a band matrix.
//
//              x x x 0 0
//              x x x x 0        upper bandwidth = 2
//              0 x x x x        lower bandwidth = 1
//              0 0 x x x
//              0 0 0 x x
//              0 0 0 0 x
//              0 0 0 0 0

class BandMatrix: public SparseMatrix
{
  protected:
    int numrows;
    int numcols;
    int lbw;           // lower bandwidth 
    int ubw;           // upper bandwidth 
    DoubleMatrix band;

    double& elem(int i, int j) { return band(i-j+ubw,i); }
    double elem(int i, int j) const { return band(i-j+ubw,i); }

  // band has the following meaning: a(i,j) = band(i,j-i+ubw)
  // for all (i,j) that fall inside the band.
    
  public:
    BandMatrix() {}
    BandMatrix(int nr, int nc, int l, int u): numrows(nr), numcols(nc), lbw(l), ubw(u), band(l+u+1,nr) {}
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector& X) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector& X) const;
    void operator*= (double t) { band *= t; }
    int NumRows() const { return numrows; }
    int NumCols() const { return numcols; }
    void Show() const;

    DoubleVector SolveCholesky(const DoubleVector& b);
    // Solves the linear system Ax=b using a Cholesky decomposition. (see Golub, p155)
    // PRE: A must be positive definite and symmetric!

    DoubleVector LowerTriangularSolveLeft(const DoubleVector& b);
    // Solves the system x^T A=b^T, where the upper triangular elements are neglected
    
    DoubleVector LowerTriangularSolve(const DoubleVector& b);
    // Solves the system Ax=b, where the upper triangular elements are neglected
    
    double& Band(int i, int j)       { return band(i,j); }
    double& operator()(int i, int j) { return elem(i,j); }
    double operator()(int i, int j) const { return elem(i,j); }
    // there is no error checking here!!
};

//--- INLINE FUNCTIONS ------------------------------------------------//

inline void SparseVector::GetVector(DoubleVector& X) const
{
  X.SetSize(NumRows());
  X.MakeZero();
  AddToVector(X);
}

inline double operator,(const SparseVector& X, const DoubleVector& Y) // inner product
{
  return X.GetInnerProduct(Y);
}

inline double operator,(const DoubleVector& Y, const SparseVector& X) // inner product
{
  return X.GetInnerProduct(Y);
}

inline std::ifstream& operator>>(std::ifstream& from, SparseVector& X)
{
  X.Read(from);
  return from;
}

inline std::ofstream& operator<<(std::ofstream& to, const SparseVector& X)
{
  X.Write(to);
  return to;
}

inline void SparseMatrix::GetMatrix(DoubleMatrix& A) const
{
  A.SetSize(NumRows(), NumCols());
  A.MakeZero();
  AddToMatrix(A);
}

inline DoubleVector operator*(const SparseMatrix& A, const DoubleVector& X)
{
  if (X.NumRows() != A.NumCols())
    std::cerr << "dimension error in operator*(SparseMatrix&, DoubleVector&)" << std::endl;
  return A.GetMatrixVectorProduct(X);
}

inline DoubleVector operator*(const DoubleVector& X, const SparseMatrix& A)
{
  if (X.NumRows() != A.NumRows())
    std::cerr << "dimension error in operator*(DoubleVector&, SparseMatrix&)" << std::endl;
  return A.GetVectorMatrixProduct(X);
}

inline std::ifstream& operator>>(std::ifstream& from, SparseMatrix& A)
{
  A.Read(from);
  return from;
}

inline std::ofstream& operator<<(std::ofstream& to, const SparseMatrix& A)
{
  A.Write(to);
  return to;
}



// The following classes should disappear in the future:


//-----------------------------------------------------------------------//
//                          QuadMatrix                                   //
//-----------------------------------------------------------------------//
// A QuadMatrix is a square 3nx3n matrix A that corresponds to the following
// quadratic function:
//                              n-1 n-1                                 T
//  f(X) = f(P_0, ..., P_n-1) = sum sum M[index(i)][index(j)] P_index(i) P_index(j)
//                              i=0 j=0

class QuadMatrix: public SparseMatrix
{
  protected:
    int numrows;     // numrows=n
    IntVector index;
    DoubleMatrix M;

  public:
    QuadMatrix(int nA, int nM): numrows(nA), index(nM), M(nM,nM)
       { M.MakeZero(); index.MakeSequence(); }
    // constructor (N.B. after calling the constructor, M and index have to be set!)
    void AddToMatrix(DoubleMatrix&) const;
    DoubleVector GetVectorMatrixProduct(const DoubleVector&) const;
    DoubleVector GetMatrixVectorProduct(const DoubleVector&) const;
    void operator*= (double t) { M *= t; }
    int NumRows() const { return 3*numrows; }
    int NumCols() const { return 3*numrows; }
    void Show() const;

    int& Index(int i) { return index[i]; }
    double& Coeff(int i, int j) { return M[i][j]; }
};

#endif

