// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sun 06-09-1996

#ifndef __POLY_H
#define __POLY_H

#include <iostream>

//-----------------------------------------------------------------------//
//                          Polynomial                                   //
//-----------------------------------------------------------------------//
// The class Polynomial represents a univariate polynomial f, with
//
//           f(t) = a0 + a1 t + ... + a[n] t^n

class Polynomial
{
  protected:
    double* a;
    int n;      // degree
    int maxn;   // maximum degree that the array coefs can hold

    void allocate(int m, int CopyContents=0);
    // maxn is set to max(maxn,m); if necessary new memory is allocated for a
    // if (CopyContents) is true, then the current contents are copied
    
    void raise_degree(int d);
    // the degree is set to max(d,n) and the coefficients a[n+1]..a[d] are set to 0
    // if necessary, new memory is allocated

    int numcoefs(int degree) const { return degree+1; }
    
  public:
    Polynomial(): n(0), maxn(10) { a = new double[maxn]; a[0]=1; }
    // default constructor; creates the polynomial f(t) = 1

    Polynomial(int n0);
    // constructor; creates the degree d polynomial f(x) = 0

    Polynomial(double* c, int degree): a(c), n(degree) { maxn = numcoefs(degree); }
    // constructor

    Polynomial(const Polynomial&);
    // copy constructor

    ~Polynomial() { delete[] a; }
    // destructor

    void RaiseDegree(int d) { raise_degree(d); }
    // the degree is set to max(d,n) and the coefficients a[n+1]..a[d] are set to 0
    // if necessary, new memory is allocated
    
    void SetDegree(int n0);
    // sets the degree of f to n0; the current contents are lost

    int Degree() const { return n; }
    // returns the degree of the polynomial

    double& operator[](int i) { return a[i]; }
    // returns the i-th coefficient (0 <= i <= n)

    double operator[](int i) const { return a[i]; }
    // returns the i-th coefficient (0 <= i <= n)

    void Show() const;
    // writes the polynomial to standard out

    double operator()(double t) const;
    // returns f(t)

    Polynomial Primitive(int k=1) const;
    // returns a primitive polynomial F, such that the k-th derivative
    // of F is equal to f

    Polynomial Derivative(int k=1) const;
    // returns the k-th derivative of f
    
    double Derivative (double t, int k=1) const;
    // returns the k-th derivative of f, evaluated at t

    double Integrate(double t0, double t1) const;
    // returns F(t1) - F(t0), where F is a primitive of f
    
    double Integrate(double x0, double x1, double a0, double a1, double b0, double b1) const;
    // returns the following integral:
    //
    //      x1   b0 + b1 x
    //      -       -
    //     |       |        f(t) dt  dx
    //    -       -
    //    x0   a0 + a1 x

    void operator  = (double c) { n=0; a[0]=c; }
    void operator  = (const Polynomial&);
    void operator += (const Polynomial&);
    void operator -= (const Polynomial&);
    void operator *= (double);
    void operator /= (double);

    friend Polynomial operator - (const Polynomial&);
    friend Polynomial operator + (const Polynomial&, const Polynomial&);
    friend Polynomial operator - (const Polynomial&, const Polynomial&);
    friend Polynomial operator * (const Polynomial&, const Polynomial&);
    friend Polynomial operator * (double, const Polynomial&);
    friend Polynomial operator * (const Polynomial&, double);
    friend Polynomial operator / (const Polynomial&, double);

    friend std::ostream& operator<<(std::ostream&, const Polynomial&);
    friend std::istream& operator>>(std::istream&, Polynomial&);
    // format: degree+1 a[0] a[1] ... a[degree]
};

#endif
