// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Wed 06-12-1996

#ifndef __POLYGON_H
#define __POLYGON_H

#include <geotyp.h>
#include <trapezd.h>

//-----------------------------------------------------------------------//
//                          Polygon2D                                    //
//-----------------------------------------------------------------------//

class Polygon2D {
  protected:
    List<double2> points;

    int is_empty() const { return (points.Count == 0); }
    double x(int i) const { return points[i].x(); }
    double y(int i) const { return points[i].y(); }
    int size() const { return points.Count; }

  public:
    Polygon2D() {}
    // default constructor

    Polygon2D(const Polygon2D& polygon);
    // copy constructor
    
    ~Polygon2D() { points.DeleteAll(); }

    void Transform(const double2x2& A, const double2& b);
    // applies the mapping x -> Ax+b to all the vertices

    void Insert(const double2& point) { points.Insert(new double2(point)); }
    void Insert(int i, const double2& point) { points.AtInsert(i, new double2(point)); }

    int NumPoints() const { return points.Count; }
    double2& Point(int i) { return points[i]; }
    const double2& Point(int i) const { return points[i]; }

    double Area() const;
    // returns the area of the polygon

    int IsConvex(double precision);
    // returns true if the polygon is convex

    int IsClockwise() const;
    // PRECONDITION: the polygon is convex
    // returns true if the vertices of the polygon are ordered clockwise
    
    void GetLeftRightTopBottom(int& left, int& right, int& top, int& bottom) const;
    // returns the indices of the point with the smallest/largest (x/y)-coordinate

    HorizontalTrapezoid2D GetHorizontalTrapezoid(double ymin, double ymax, const int2& left, const int2& right) const;
    // returns the trapezoid bounded by the lines
    //
    //               y=ymin
    //               y=ymax
    //
    // and the line segments
    //
    //      [ Point(left[0]), Point(left[1]) ]
    //      [ Point(right[0]), Point(right[1]) ]
    //
    
    VerticalTrapezoid2D GetVerticalTrapezoid(double xmin, double xmax, const int2& bottom, const int2& top) const;
    // returns the trapezoid bounded by the lines
    //
    //               x=xmin
    //               x=xmax
    //
    // and the line segments
    //
    //      [ Point(bottom[0]), Point(bottom[1]) ]
    //      [ Point(top[0]), Point(top[1]) ]
    //
    
    List<HorizontalTrapezoid2D> GetHorizontalTrapezoids(int SplitAtIntegers = 0, double precision = 1e-10) const;
    // PRECONDITION: the polygon is convex
    // splits the polygon in horizontal trapezoids by cutting it with
    // horizontal lines through the vertices
    //
    // SplitAtIntegers: if true, then the polygon is also cut at horizontal
    //                  lines y=yi, with yi an integer
    // 
    // precision: trapezoids with (height < precision) are neglected

    List<VerticalTrapezoid2D> GetVerticalTrapezoids(int SplitAtIntegers = 0, double precision = 1e-10) const;
    // PRECONDITION: the polygon is convex
    // splits the polygon in vertical trapezoids by cutting it with
    // vertical lines through the vertices
    //
    // SplitAtIntegers: if true, then the polygon is also cut at vertical
    //                  lines x=xi, with xi an integer
    // 
    // precision: trapezoids with (width < precision) are neglected

    List<VerticalTrapezoid2D> GetTrapezoids(double precision) const;
    // PRECONDITION: the polygon is convex
    // splits the polygon in vertical trapezoids such that any piece
    // lies inside a square [i,i+1] x [j,j+1] with i and j integers

    void GetTriangulation(List<Triangle2D>& triangles) const;
    // PRECONDITION: the polygon is convex
    // returns a triangulation of the polygon

    double2 Center() const;
    // returns the center of the vertices of the polygon
    
	void Show() const;
	// writes the contents to standard out

    void Scale(const double2& P, double t);
    // scales the polygon with a factor t w.r.t. the point P

    friend std::ostream& operator<<(std::ostream& os, const Polygon2D& polygon);
    friend std::istream& operator>>(std::istream& is, Polygon2D& polygon);
};

#endif
