// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Tue 06-04-1996

#ifndef __SAMPLE_H
#define __SAMPLE_H 1

#include <iostream>

#include <double2.h>
#include <double2x2.h>
#include <double3.h>

//-----------------------------------------------------------------------//
//                          CurvSample
//-----------------------------------------------------------------------//
// The class CurvSample represents all kind of information of a parametric
// curve F in a point with parameter t.
//
// F   = F(t)
// Ft  = Ft(t)
// Ftt = Ftt(t)
//
class CurvSample {
  public:
    double3 F, Ft, Ftt;

  public:
    CurvSample() {}
    // default constructor

    CurvSample(const double3& F0, const double3& Ft0, const double3& Ftt0):
      F(F0), Ft(Ft0), Ftt(Ftt0) {}
    // constructor
    
    CurvSample(const CurvSample& sample) {
      F   = sample.F;
      Ft  = sample.Ft;
      Ftt = sample.Ftt;
    }
    // copy constructor

    ~CurvSample() {}
    // destructor

    void MakeZero() {
      F.MakeZero();
      Ft.MakeZero();
      Ftt.MakeZero();
    }

    void SetValues(const double3& F0, const double3& Ft0, const double3& Ftt0) {
      F   = F0;
      Ft  = Ft0;
      Ftt = Ftt0;
    }

    double Curvature() const;
    // returns the curvature in F(t)

    void Show() const;
    // writes the contents to standard out

    friend std::ostream& operator<<(std::ostream&, const CurvSample&);
    friend std::istream& operator>>(std::istream&, CurvSample&);
};

//-----------------------------------------------------------------------//
//                          FuncSample1D
//-----------------------------------------------------------------------//
// The class FuncSample1D represents all kind of information of a 1D scalar
// function f in a specific point t.
//
// f   = f(t)
// ft  = f'(t)
// ftt = f''(t)
//

class FuncSample1D {
  public:
    double f, ft, ftt;

  public:
    FuncSample1D() {}
    // default constructor

    FuncSample1D(double f0, double ft0, double ftt0):
      f(f0), ft(ft0), ftt(ftt0) {}
    // constructor
    
    FuncSample1D(const FuncSample1D& sample) {
      f   = sample.f;
      ft  = sample.ft;
      ftt = sample.ftt;
    }
    // copy constructor

    ~FuncSample1D() {}
    // destructor

    void MakeZero() {
      f = ft = ftt = 0;
    }

    void SetValues(const double& f0, const double& ft0, const double& ftt0) {
      f   = f0;
      ft  = ft0;
      ftt = ftt0;
    }

    void Show() const;
    // writes the contents to standard out

    friend std::ostream& operator<<(std::ostream&, const FuncSample1D&);
    friend std::istream& operator>>(std::istream&, FuncSample1D&);
};

//-----------------------------------------------------------------------//
//                          FuncSample2D
//-----------------------------------------------------------------------//
// The class FuncSample2D represents all kind of information of a scalar
// function f in a specific point (u,v).
//
// f   = f(u,v)
// fu  = fu(u,v)
// fv  = fv(u,v)
// fuu = fuu(u,v)
// fuv = fuv(u,v)
// fvv = fvv(u,v)
//

class FuncSample2D {
  public:
    double f, fu, fv, fuu, fuv, fvv;

  public:
    FuncSample2D() {}
    // default constructor

    FuncSample2D(double f0, double fu0, double fv0, double fuu0, double fuv0, double fvv0):
      f(f0), fu(fu0), fv(fv0), fuu(fuu0), fuv(fuv0), fvv(fvv0) {}
    // constructor

    FuncSample2D(const FuncSample2D& sample) {
      f   = sample.f;
      fu  = sample.fu;
      fv  = sample.fv;
      fuu = sample.fuu;
      fuv = sample.fuv;
      fvv = sample.fvv;
    }
    // copy constructor

    ~FuncSample2D() {}
    // destructor

    void MakeZero() {
      f = fu = fv = fuu = fuv = fvv = 0;
    }

    void SetValues(const double& f0, const double& fu0, const double& fv0,
                   const double& fuu0, const double& fuv0, const double& fvv0) {
      f   = f0;
      fu  = fu0;
      fv  = fv0;
      fuu = fuu0;
      fuv = fuv0;
      fvv = fvv0;
    }

    double2x2 Hessian() const { return double2x2(fuu, fuv, fuv, fvv); }

    void Show() const;
    // writes the contents to standard out

    friend std::ostream& operator<<(std::ostream&, const FuncSample2D&);
    friend std::istream& operator>>(std::istream&, FuncSample2D&);
};

//-----------------------------------------------------------------------//
//                          SurfSample
//-----------------------------------------------------------------------//
// The class SurfSample represents all kind of information of a parametric
// surface F in a specific point (u,v).
//
// F   = F(u,v)
// Fu  = Fu(u,v)
// Fv  = Fv(u,v)
// Fuu = Fuu(u,v)
// Fuv = Fuv(u,v)
// Fvv = Fvv(u,v)
//

class SurfSample {
  public:
    double3 F, Fu, Fv, Fuu, Fuv, Fvv;

  public:
    SurfSample() {}
    // default constructor

    SurfSample(const double3& F0, const double3& Fu0, const double3& Fv0,
               const double3& Fuu0, const double3& Fuv0, const double3& Fvv0);
    // constructor
    
    SurfSample(const SurfSample&);
    // copy constructor

    ~SurfSample() {}
    // destructor

    void MakeZero() {
      F.MakeZero();
      Fu.MakeZero();
      Fv.MakeZero();
      Fuu.MakeZero();
      Fuv.MakeZero();
      Fvv.MakeZero();
    }

    void SetValues(const double3& F0, const double3& Fu0, const double3& Fv0,
                   const double3& Fuu0, const double3& Fuv0, const double3& Fvv0);

    void GetCurvatures(double& kappa1, double& kappa2) const;
    // computes the principal curvature values

    double2x2 FirstFundamentalForm() const;
    // returns the first fundamental form

    double2x2 SecondFundamentalForm() const;
    // returns the second fundamental form

    double3 Normal() const;
    // returns the unit normal vector

    double TotalCurvature() const;
    // returns the value kappa1^2 + kappa2^2

    double MeanCurvature() const;
    // returns the mean curvature (kappa1 + kappa2)/2

    double GaussianCurvature() const;
    // returns the Gaussian curvature kappa1*kappa2

    double SimpleArea() const
      { return (Fu,Fu) + (Fv,Fv); }
    
    double SimpleTotalCurvature() const
      { return (Fuu,Fuu) + 2*(Fuv,Fuv) + (Fvv,Fvv); }

    double SimpleMeanCurvature() const
      { return 0.5*(Fuu + Fvv, Fuu + Fvv); }

    double SimpleGaussianCurvature() const
      { return (Fuu,Fvv) - (Fuv,Fuv); }

    void Show() const;
    // writes the contents to standard out

    friend std::ostream& operator<<(std::ostream&, const SurfSample&);
    friend std::istream& operator>>(std::istream&, SurfSample&);
};

//-----------------------------------------------------------------------//
//                          SurfSampleExt
//-----------------------------------------------------------------------//
// The class SurfSampleExt is a descendant of SurfSample, in which most of
// the values are stored as attributes
//
// F   = F(u,v)
// Fu  = Fu(u,v)
// Fv  = Fv(u,v)
// Fuu = Fuu(u,v)
// Fuv = Fuv(u,v)
// Fvv = Fvv(u,v)
//
// n     = Fu(u,v) x Fv(u,v) / ||Fu(u,v) x Fv(u,v)||    (unit normal)
//
// I     = first fundamental form
//
// g     = determinant of I
//
// I_inv = inverse of first fundamental form
//
// II    = second fundamental form
//
// Christoffel symbols:
//
//    gamma_1[i][j] = Gamma            gamma_2[i][j] = Gamma
//                         ij1                              ij2
//
//                          1                                2
//    gamma1[i][j]  = Gamma            gamma2[i][j]  = Gamma
//                         ij                               ij
//

class SurfSampleExt: public SurfSample {
  public:
    double3 n;
    double2x2 gamma1, gamma2, gamma_1, gamma_2, I, I_inv, II;
    double g;

  public:
    SurfSampleExt() {}
    // default constructor

    SurfSampleExt(const double3& F0, const double3& Fu0, const double3& Fv0,
               const double3& Fuu0, const double3& Fuv0, const double3& Fvv0);
    // constructor
    
    SurfSampleExt(const SurfSampleExt&);
    // copy constructor

    ~SurfSampleExt() {}
    // destructor

    void SetValues(const double3& F0, const double3& Fu0, const double3& Fv0,
                   const double3& Fuu0, const double3& Fuv0, const double3& Fvv0);

    void GetCurvatures(double& kappa1, double& kappa2) const;
    // computes the principal curvature values

    double TotalCurvature() const;
    // returns the value kappa1^2 + kappa2^2

    double MeanCurvature() const;
    // returns the mean curvature (kappa1 + kappa2)/2

    double GaussianCurvature() const;
    // returns the Gaussian curvature kappa1*kappa2

    void Show() const;
    // writes the contents to standard out
};

#endif
