// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <iostream>
#include <sstream>
#include <fstream>
#include <bvalues.h>
#include <cstdlib>
#include "vbspline.h"
#include "potent.h"

const double FACTOR = 5.0;

std::ofstream report("tst3.report", std::ios::out | std::ios::app);

void Test(int NumControlPoints, double alpha)
{
  std::cout << "n, alpha = " << NumControlPoints << " " << alpha << std::endl;
  VariationalBSpline B;

  for (int i=0; i<NumControlPoints; i++) {
    double3* Point = new double3(
      double3((100.0*i)/NumControlPoints, 0, 0) + FACTOR*double3(drand48(), drand48(), drand48())
    );
    B.InsertControlPoint(Point);
  }

  double MinPar = B.MinPar();
  double MaxPar = B.MaxPar();

  // define internal energy of the curve
  double Weight = 1;
  BendEnergyPotential* BendEnergy =
    new BendEnergyPotential(MinPar, MaxPar, alpha*Weight);
  StretchEnergyPotential* StretchEnergy =
    new StretchEnergyPotential(MinPar, MaxPar, (1-alpha)*Weight);

  // define two point constraints at the endpoints of the curve
  double3 BeginPoint(0, 0, 0);
  double3 EndPoint(100, 0, 0);
  PointConstraint* Constr1 = new PointConstraint(MinPar, BeginPoint);
  PointConstraint* Constr2 = new PointConstraint(MaxPar, EndPoint);

  B.InsertPotential(BendEnergy);
  B.InsertPotential(StretchEnergy);
  B.InsertConstraint(Constr1);
  B.InsertConstraint(Constr2);

  // minimize energy using conjugate gradient method
  if (B.MinimizeConjugateGradient(10000, 1e-3))
    B.CheckConstraints();

  DoubleMatrix A;
  DoubleVector b;

  std::ifstream from("conjgrad.mtx");
  from >> b >> A;
  from.close();

  DoubleVector ev = A.EigenValues();
  ev.Sort();

  // create a filename
  std::stringstream out;
  out << "eig" << std::setw(3) << std::setfill('0') << NumControlPoints 
      << '_' << std::setw(2) << (int) 100*alpha << ".dat" << std::ends;
  ev.WriteToFile(out.str());
  
  // read conjgrad.info
  std::ifstream info("conjgrad.info");
  double starterror, enderror;
  int numiter;
  info >> starterror >> enderror >> numiter;
  info.close();
  std::cout << "sen = " << starterror << " " << enderror << " " << numiter << std::endl;
  
  double lmin = ev[0];
  double lmax = ev[ev.NumRows()-1];
  report << std::setw(4) << setprecision(3) << alpha 
         << std::setw(8) << 3*NumControlPoints 
         << std::setw(10) << numiter 
         << setprecision(4) << std::setw(16) << lmin 
         << std::setw(12) << lmax
         << std::setw(12) << lmax/lmin 
         << std::setw(13) << starterror 
         << std::setw(14) << enderror 
         << std::endl;
}

int main()
{
  ReadEnergyMatrices();

  report << "alpha    dim    #iteraties     l_min        l_max      cond. getal    startfout    eindfout" << std::endl;

  std::ifstream in("tst3.in");
  for (;;) {
    int n;
    double alpha;
    in >> n;
    in >> alpha;
    if (!in) break;
    Test(n, alpha);
  }
  
  report.close();
}
