// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#include <iostream>
#include <sstream>
#include <fstream>
#include <bvalues.h>
#include <cstdlib>
#include <cmath>
#include <string>
#include <equation.h>
#include <dvector.h>
#include <dmatrix.h>
#include "vbspline.h"

//-----------------------------------------------------------------------//
//                          GetFileName                                  //
//-----------------------------------------------------------------------//
std::string GetFileName(std::string InFile, int i)
// creates the name InFile_i
{
  std::stringstream out;
  out << InFile << std::setw(3) << std::setfill('0') << i << ".dat" << std::ends;
  return out.str();
}

//-----------------------------------------------------------------------//
//                          WriteInfo                                    //
//-----------------------------------------------------------------------//

void WriteInfo(std::ostream& to, const BSpline& x)
{
  to << "  exact stretch energy   = " << x.ComputeStretchEnergy() << std::endl;
  to << "  simple stretch energy  = " << x.ComputeSimpleStretchEnergy() << std::endl;
  to << "  exact bend energy      = " << x.ComputeBendEnergy() << std::endl;
  to << "  simple bend energy     = " << x.ComputeSimpleBendEnergy() << std::endl;
}

//-----------------------------------------------------------------------//
//                          IterateCurveDirect                           //
//-----------------------------------------------------------------------//
void IterateCurveDirect(std::string BaseName, int NumIterations)
{
  std::string InFileName = BaseName + ".dat";
  std::string OutFileName;
  std::string ReportFileName = BaseName + ".rep";

  std::ofstream report(ReportFileName.c_str());
  if (!report)
    std::cerr << "Could not open file " << ReportFileName;

  // read the curve from a file
  VariationalBSpline x(InFileName);

  LinearEquationSystem constraints;
  double t0 = x.MinPar();
  double t2 = x.MaxPar();
  double t1 = (t0+t2)/2;

std::cout << "t012 = " << t0 << " " << t1 << " " << t2 << std::endl;

  double3 x0 = x.GetPoint(t0);
  double3 x1 = x.GetPoint(t1);
  double3 x2 = x.GetPoint(t2);
  constraints += x.GetPointToPointConstraint(t0,x0);
  constraints += x.GetPointToPointConstraint(t1,x1);
  constraints += x.GetPointToPointConstraint(t2,x2);

  DoubleMatrix D;
  DoubleVector E;
  constraints.GetMatrix(D);
  constraints.GetVector(E);
  DoubleVector lambda(constraints.NumRows(),0);

  DoubleMatrix F;
  DoubleVector G;
  IntVector index;
  constraints.GetEliminateForm(F,G,index);
  std::cout << "constraints computed!" << std::endl;
  DoubleVector X = x.GetControlVector();
 
  report << "------------------------------------------------------" << std::endl;
  report << "                Curve " << 0 << std::endl;
  report << "------------------------------------------------------" << std::endl;
  WriteInfo(report,x);
  OutFileName = GetFileName(BaseName, 0);
  x.SaveFile(OutFileName);
//  constraints.Check(X);

  for (int i=0; i<NumIterations; i++) {
    std::cout << "iteration " << i << std::endl;
    report << "------------------------------------------------------" << std::endl;
    report << "                Curve " << i+1 << std::endl;
    report << "------------------------------------------------------" << std::endl;

    BSpline y = x;

    QuadraticFunctionList f_total = x.GetDataStretchEnergy(y);

    report << "voor: f(X) = " << f_total(X) << std::endl;
    f_total.GetConstrainedMinimumLagrange(D, E, X, lambda, BKP_method);
    report << "na  : f(X) = " << f_total(X) << std::endl;

    x.SetControlVector(X);
    WriteInfo(report,x);
    OutFileName = GetFileName(BaseName, i+1);
    x.SaveFile(OutFileName);
//    constraints.Check(X);

    f_total.DeleteAll();
  }
}

int main(int argc, char* argv[])
// usage: FileName, maxiter
{
  int maxiter;
  char BaseName[80];

  std::cout << "Iterating curve" << std::endl;

  if (argc > 1) {
    strcpy(BaseName, argv[1]);
  }
  else {
    std::cout << "input file: "; std::cin >> BaseName;
  }

  if (argc > 2) {
    maxiter = atoi(argv[2]);
  }
  else {
    std::cout << "number of iterations: "; std::cin >> maxiter;
  }

  if (maxiter <= 0) {
    std::cerr << "Number of iterations must be >= 0!" << std::endl;
    exit(0);
  }

  IterateCurveDirect(BaseName, maxiter);

  return 0;
}

