// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#include <cmath>
#include <iostream>
#include <splinval.h>

//--- NUBSplineValues ----------------------------------------------//
/*
    Carry out the Cox-deBoor recurrence to compute the nonzero B-spline
    values at a given value of the parameter.

  Input:
    parval --- parameter value for evaluation
    order ---- order of spline
    knots ---- knot vector portion s.t. knots[index] <= parval < knots[index+1]
    index ---- index for knots indicating position of parval

  Output:
    bvals[i] ---- for i=0,...,order-1 containing the nonzero B-spline values
                  B[index,order](parval),...,B[index-order+1,order](parval),
                  where bvals[j] = B[index-j,order](parval)

  Caution:
    No validity checking is done!
    Make sure parval is not to the left of the first segment or
    to the right of the last.  Make sure that knots[index] < knots[index+1],
    and make sure there are enough knots to the left of the first segment
    and to the right of the last.  If you don't, you'll be sooorrreeee.
*/
void NUBSplineValues(double parval, int order, const double* knots, int index, double* bvals)
{ int r,i,s;
  double omega;

  if( order<=0 ) return;

  bvals[0] = 1.0;
  for ( r=2; r<=order; r++ ) {
    i = index-r+1;
    bvals[r-1] = 0.0;
    for ( s=r-2; s>=0; s-- ) {
      i++;
      omega = ( parval - knots[i] ) / ( knots[i+r-1] - knots[i] );
      bvals[s+1] += ( 1.0 - omega ) * bvals[s];
      bvals[s] *= omega;
    }
  }
}

//--- NUBSplineDerivs ----------------------------------------------//
/*
    Carry out Sablionnier's tetrahedral recurrence to compute the
    derivatives of the nonzero B-spline values at a given parameter.

  Input:
    parval --- parameter value for evaluation
    order ---- order of spline
    deriv ---- the desired derivative order
    knots ---- knot vector portion s.t. knots[index] <= parval < knots[index+1]
    index ---- index for knots indicating position of parval

  Output:
    dvals[i] ---- for i=0,...,order-1 containing the deriv-th
                  derivatives of the nonzero bspline values
                  B[index,order](parval),...,B[index-order+1,order](parval),
                  where dvals[j] = D^deriv B[index-j,order](parval)

  Caution:
    No validity checking is done!
    Read the cautionary note for NUBSplineValues,
    and consider well what thou doist.
*/
void NUBSplineDerivs(double parval, int order, int deriv, const double *knots, int index, double *dvals)
{ int r,i,s;
  double omega;

  for( r=0; r<order; r++ ) {
    dvals[r] = 0.0;
  }
  if( (deriv>=order) || (deriv<0) ) return;

  NUBSplineValues( parval, order-deriv, knots+deriv, index-deriv, dvals );

//std::cout << "NUBSplineValues(): " << std::endl;
//for (int y=0; y<4; y++) std::cout << dvals[y] << " ";
//std::cout << std::endl;

  for( r=order-deriv+1; r<=order; r++ ) {
    i = index-r+1;
    for( s=r-2; s>=0; s-- ) {
      i++;
      omega = ((double)(r-1)) / (knots[i+r-1] - knots[i]);
      dvals[s+1] += -omega * dvals[s];
      dvals[s] *= omega;
    }
  }

//std::cout << "NUBSplineDerivs(): " << std::endl;
//for (y=0; y<4; y++) std::cout << dvals[y] << " ";
//std::cout << std::endl;
}

//--- BSplineValues ------------------------------------------------//
double* BSplineValues(double u, int degree, const double* knots, int /*index*/)
//  this function calculates B-spline values for a fixed parameter u
//
//  input:
//    u -------- parameter value for evaluation
//    degree --- degree of spline
//    knots ---- portion of knot vector: knots[i] = u[index+i], i = 0..degree
//    index ---- indicates which values are wanted
//               (this will generally be the segment number of u, which is defined
//                by u[index+degree-1] <= u < u[index+degree] (???) )
//
//  output:
//    bvals[i] = N[index+i, degree](u),       i = 0..degree
// Based on Mansfield, de Boor, Cox recursion (Farin, p. 185)
//
//                 N[l-1,n+1]
//              /                              u - u[l-1]
//        1-t /                        t =  ---------------
//          /                               u[l+n] - u[l-1]
//  N[l,n] - - - - N[l,n+1]
//             t
{
  int i,j;

  double* bvals = new double[degree+1];
  for (i=0; i<degree; i++) bvals[i] = 0;
  bvals[degree] = 1;

  for (i=0; i<degree; i++) {
    for (j=degree-i; j<=degree; j++) {
      int n = i;
      int l = j;
      double t = (u - knots[l-1]) / (knots[l+n] - knots[l-1]);
      bvals[j-1] += (1-t)*bvals[j];
      bvals[j] *= t;
    }
  }

  return bvals;
}

//--- BSplinePolynomials -------------------------------------------//
Polynomial* BSplinePolynomials(int degree, const double* knots, int /*index*/)
//  this function calculates B-spline polynomials
//
//  input:
//    degree --- degree of spline
//    knots ---- portion of knot vector: knots[i] = u[index+i], i = 0..degree
//    index ---- indicates which values are wanted
//
//  output:
//    bpols[i],      i = 0..degree
//       with bpols[i](u) = N[index+i, degree](u)
//
// Based on Mansfield, de Boor, Cox recursion (Farin, p. 185)
//
//                 N[l-1,n+1]
//              /                              u - u[l-1]
//        1-t /                        t =  ---------------
//          /                               u[l+n] - u[l-1]
//  N[l,n] - - - - N[l,n+1]
//             t
{
  int i,j,k;

  double** pols = new double*[degree+1];
  for (i=0; i<=degree; i++) {
    pols[i] = new double[degree+1];
    for (j=0; j<=degree; j++) pols[i][j] = 0;
  }
  pols[degree][0] = 1;

  double* temp = new double[degree+1];

  for (i=0; i<degree; i++) {
    for (j=degree-i; j<=degree; j++) {
      int n = i;
      int l = j;

      // t = a*u + b
      double a = 1 / (knots[l+n] - knots[l-1]);
      double b = - knots[l-1]*a;

      // temp = t*pols[j]
      temp[0] = b*pols[j][0];
      for (k=1; k<=i; k++) temp[k] = a*pols[j][k-1] + b*pols[j][k];
      temp[i+1] = a*pols[j][i];

      // pols[j-1] += (1-t)*pols[j];
      for (k=0; k<=i; k++) pols[j-1][k] += (pols[j][k] - temp[k]);
      pols[j-1][i+1] -= temp[i+1];

      // pols[j] *= t;
      for (k=0; k<=i+1; k++) pols[j][k] = temp[k];
    }
  }

  Polynomial* bpols = new Polynomial[degree+1];
  for (i=0; i<=degree; i++) bpols[i] = Polynomial(pols[i], degree);

  delete pols;
  delete temp;

  return bpols;
}

//--- DeBoor -------------------------------------------------------//
double DeBoor(int degree, double* coeff, const double* knot, double u, int /*index*/)
// u[index] <= u < u[index+1]
// coeff[i] = d[index-degree+1+i] (i=0..degree)
// knot[i] = u[index-degree+1+i]  (i=0..2*degree-1)
// de Boor algorithm: (Farin p. 166)
//
// d[i,k] = (u[i+n-k]-u)/(u[i+n-k)-u[i-1]) d[i-1,k-1] + (u-u[i-1])/(u[i+n-k]-u[i-1]) d[i,k-1]
//
{
  for (int i=1; i<=degree; i++)
    for (int j=degree; j>=i; j--) {
      double t = (knot[j-i+degree]-u)/(knot[j-i+degree]-knot[j-1]);
      coeff[j] = t*coeff[j-1] + (1-t)*coeff[j]; // coeff[j] = d[index-degree+1+j,i]
   }
 return coeff[degree];
}

/*
//--- BSplineDerivs ------------------------------------------------//
double* BSplineDerivs(double u, int degree, double* knots, int index, int deriv)
//  this function calculates derivatives of B-spline values for a fixed parameter u
//
//  input:
//    u -------- parameter value for evaluation
//    degree --- degree of spline
//    knots ---- knot vector
//    index ---- indicates which values are wanted
//    deriv ---- the desired derivative order
//
//  output:
//    bvals[i] = N(deriv)[index+i, degree](u),       i = 0..degree
//
//                   n                             n
//  N'[l,n] = ----------------- N[l,n-1]  -  ------------- N[l+1,n-1]
//            u[n+l-1] - u[l-1]              u[l+n] - u[l]
//
{
  int i,j;

  double* bvals = BSplineValues(u, degree, knots, index);

  for (i=0; i<degree; i++) {
    for (j=degree; j>=degree-i; j--) {
      int n = i;
      int l = index + j;
      double t = (u - knots[l-1]) / (knots[l+n] - knots[l-1]);
      bvals[j-1] += (1-t)*bvals[j];
      bvals[j] *= t;
    }
  }

  return bvals;
}
*/
