// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sat 06-22-1996

#include "tbspline.h"

#include <iostream>
#include <fstream>
#include <math.h>
#include <mystuff.h>
#include <bvalues.h>
#include <integr.h>

#include <pause.h>

//-----------------------------------------------------------------------//
//                          MatIndex                                     //
//-----------------------------------------------------------------------//
inline static int MatIndex(int i, int j)
{
  return 4*i + j;
}

//-----------------------------------------------------------------------//
//                          TensorBSpline                                //
//-----------------------------------------------------------------------//

TensorBSpline::TensorBSpline(): numRows(0), numCols(0)
{
  for (int i=0; i<MAXKNOTS; i++)
    uknots[i] = vknots[i] = i;
}

TensorBSpline::TensorBSpline(int nrows, int ncols): numRows(nrows), numCols(ncols)
{
  for (int i=0; i<MAXKNOTS; i++)
    uknots[i] = vknots[i] = i;
}

TensorBSpline::TensorBSpline(char* FileName)
{
  for (int i=0; i<MAXKNOTS; i++)
    uknots[i] = vknots[i] = i;
  LoadFile(FileName);
}

void TensorBSpline::operator=(TensorBSpline& G)
{
  for (int i=0; i<MAXKNOTS; i++)
    uknots[i] = vknots[i] = i;

  for (int i=0; i<G.NumControlPoints(); i++)
    SetControlPoint(i, G.GetControlPoint(i));
}

int TensorBSpline::UIndex(double u) const
{
  int result;

  if (u < MinUPar()) result = 2;
  else if (u >= MaxUPar()) result = NumUKnots()-4;
    else result = int(floor(u));

  return result;
}

int TensorBSpline::VIndex(double v) const
{
  int result;

  if (v < MinVPar()) result = 2;
  else if (v >= MaxVPar()) result = NumVKnots()-4;
    else result = int(floor(v));

  return result;
}

Rectangle2D TensorBSpline::GetDomain() const
{
  double umin = MinUPar();
  double umax = MaxUPar();
  double vmin = MinVPar();
  double vmax = MaxVPar();

  return Rectangle2D(double2(umin,vmin), umax-umin, vmax-vmin);
}

double TensorBSpline::GetSizeMeasure() const
// returns a rough approximation of the size of the surface
{
  double umin = MinUPar();
  double umax = MaxUPar();
  double vmin = MinVPar();
  double vmax = MaxVPar();

  double3 P0 = GetPoint(double2(umin,vmin));
  double3 P1 = GetPoint(double2(umax,vmin));
  double3 P2 = GetPoint(double2(umax,vmax));
  double3 P3 = GetPoint(double2(umin,vmax));
  
  double Result = 0;
  if ((P0-P1).Length() > Result) Result = (P0-P1).Length();
  if ((P1-P2).Length() > Result) Result = (P1-P2).Length();
  if ((P2-P3).Length() > Result) Result = (P2-P3).Length();
  if ((P3-P0).Length() > Result) Result = (P3-P0).Length();
  
  return Result;
}

void TensorBSpline::GetSurfaceMesh(int numrows, int numcols, SurfaceMesh& mesh) const
// generates a mesh of the surface by splitting the domain in numcols*numrows rectangles
//
// Numbering scheme of the mesh points.
//
//       |
//       |
//      nc+1 -- nc+2 -- nc+1 ---               2nc+1
//       |       |       |                       |
//       |       |       |                       |
//       0  ---  1  ---  2 ---       ...    ---  nc
//
//    (umin,vmin)
//
//
//      k+nc+1 ------ k+nc+2
//        |          .  |
//        |       .     |
//        |    .        |
//        | .           |
//        k ---------- k+1
//

{
  // compute the samples
  double umin = MinUPar();
  double umax = MaxUPar();
  double vmin = MinVPar();
  double vmax = MaxVPar();
  double du = (umax-umin)/numrows;
  double dv = (vmax-vmin)/numcols;

  for (int i=0; i<=numrows; i++)
    for (int j=0; j<=numcols; j++) {
      double v = vmin + i*dv;
      double u = umin + j*du;
      mesh.Insert(GetSurfSample(double2(u,v)));
    }

  // compute the quadrangles
  for (int i=0; i<numrows; i++)
    for (int j=0; j<numcols; j++) {
      int k = (numcols+1)*i + j;
      mesh.Insert(int4(k, k+1, k+numcols+2, k+numcols+1));
    }
}

void TensorBSpline::GetSurfaceMesh(const Polygon2D& gamma, SurfaceMesh& mesh, int resolution) const
// generates a mesh of the surface on the domain gamma
{
  List<Triangle2D> triangles;
  gamma.GetTriangulation(triangles);
  std::cout << "polygon triangulated" << std::endl;

  for (int i=0; i<triangles.Count; i++) {
    Triangle2D triangle = triangles[i];
    List<double2> gridpoints = triangle.GetGridPoints(resolution);
    List<int3> gridtriangles = triangle.GetGridTriangles(resolution);

    // do not change the order of these two loops!!
    int diff = mesh.NumSamples();

    for (int j=0; j<gridtriangles.Count; j++)
      mesh.Insert(gridtriangles[j] + int3(diff,diff,diff));

    for (int j=0; j<gridpoints.Count; j++)
      mesh.Insert(GetSurfSample(gridpoints[j]));
  }
}

void TensorBSpline::GetControlPointMesh(PolygonMesh& mesh) const
// generates a mesh of the control polyhedron
{
  for (int i=0; i<NumControlPoints(); i++)
    mesh.Insert(GetControlPoint(i));
    
  // compute the quadrangles
  for (int i=0; i<NumRows()-1; i++)
    for (int j=0; j<NumCols()-1; j++) {
      int k = i*NumCols() + j;
      mesh.Insert(int4(k, k+1, k+NumCols()+1, k+NumCols()));
    }
}

IntVector TensorBSpline::GetPatchIndices(int p, int q) const
// returns the indices of the control points
//
//          ControlPoint(i+p,j+q),      0 <= i,j <= 3
//
// that belong to the patch with domain [p+2,p+3] x [q+2,q+3]
{
  IntVector Result(16);

  int k=0;
  
  for (int i=p; i<=p+3; i++)
    for (int j=q; j<=q+3; j++)
      Result(k++) = ControlPointIndex(i,j);
      
  return Result;
}          

void TensorBSpline::SetCylinder(int NumRows, int NumCols, double h, double R, double Angle)
{
  numRows = NumRows;
  numCols = NumCols;

  double du = (M_PI*Angle/180)/(NumRows-3);
  double dv = h/(NumCols-3);

  for (int i=0; i<NumRows; i++)
    for (int j=0; j<NumCols; j++) {
      double u = (i-1)*du;
      double v = (j-1)*dv;
      SetControlPoint(i, j, double3(R*cos(u), R*sin(u), v));
    }
}

void TensorBSpline::SetTorus(int NumRows, int NumCols, double R, double RAngle, double r, double rAngle)
{
  numRows = NumRows;
  numCols = NumCols;

  double du = (M_PI*RAngle/180)/(NumRows-3);
  double dv = (M_PI*rAngle/180)/(NumCols-3);

  for (int i=0; i<NumRows; i++)
    for (int j=0; j<NumCols; j++) {
      double u = (i-1)*du;
      double v = (j-1)*dv;
      SetControlPoint(i, j, double3( (R + r*cos(v))*cos(u), (R + r*cos(v))*sin(u), r*sin(v) ));
    }
}

void TensorBSpline::SetPlane(int NumRows, int NumCols, double Width, double Height)
{
  numRows = NumRows;
  numCols = NumCols;

  double du = Width/(NumRows-3);
  double dv = Height/(NumCols-3);

  for (int i=0; i<NumRows; i++)
    for (int j=0; j<NumCols; j++) {
      double u = (i-1)*du;
      double v = (j-1)*dv;
      SetControlPoint(i, j, double3(u, v, 0));
    }
}

SurfSample TensorBSpline::GetSurfSample(const double2& x) const
{
  double NU[4], NUu[4], NUuu[4], NV[4], NVv[4], NVvv[4];

  double u = x.x();
  double v = x.y();

  if ( (u < MinUPar()) || (u > MaxUPar()) || (v < MinVPar()) || (v > MaxVPar()) ) {
    std::cerr << "error in TensorBSpline::GetSurfSample(): parameter outside domain!" << std::endl;
  }

  int k = USegment(u);
  int l = VSegment(v);

  BSplineEvaluator.GetNValues(u-k, NU);
  BSplineEvaluator.GetNxValues(u-k, NUu);
  BSplineEvaluator.GetNxxValues(u-k, NUuu);

  BSplineEvaluator.GetNValues(v-l, NV);
  BSplineEvaluator.GetNxValues(v-l, NVv);
  BSplineEvaluator.GetNxxValues(v-l, NVvv);

  double3 F, Fu, Fv, Fuu, Fuv, Fvv;
  F.MakeZero();
  Fu.MakeZero();
  Fv.MakeZero();
  Fuu.MakeZero();
  Fuv.MakeZero();
  Fvv.MakeZero();

  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++) {
      F   += NU[i]   * NV[j]   * GetControlPoint(i+k,j+l);
      Fu  += NUu[i]  * NV[j]   * GetControlPoint(i+k,j+l);
      Fv  += NU[i]   * NVv[j]  * GetControlPoint(i+k,j+l);
      Fuu += NUuu[i] * NV[j]   * GetControlPoint(i+k,j+l);
      Fuv += NUu[i]  * NVv[j]  * GetControlPoint(i+k,j+l);
      Fvv += NU[i]   * NVvv[j] * GetControlPoint(i+k,j+l);
    }

  return SurfSample(F, Fu, Fv, Fuu, Fuv, Fvv);
}

void TensorBSpline::GetBoundaryCurves(BSpline& umincurve, BSpline& umaxcurve, BSpline& vmincurve, BSpline& vmaxcurve) const
// returns the curves that represent the boundary of the surface
{
  //                  nc-1    3
  // umincurve f(v) = sum  ( sum N_i(umin) P_ij ) N_j(v)
  //                  j=0    i=0
  //
  //                  nc-1   nr-1
  // umaxcurve f(v) = sum  ( sum N_i(umax) P_ij ) N_j(v)
  //                  j=0   i=nr-4
  //
  //                  nr-1    3
  // vmincurve f(u) = sum  ( sum N_j(vmin) P_ij ) N_i(v)
  //                  i=0    j=0
  //
  //                  nr-1   nc-1
  // vmaxcurve f(u) = sum  ( sum N_j(vmax) P_ij ) N_i(v)
  //                  i=0   j=nc-4
  
  int nc = NumCols();
  int nr = NumRows();
  double N[4];

  BSplineEvaluator.GetNValues(2,N); 
  for (int j=0; j<nc; j++) {
    double3 Q(0,0,0);
    for (int i=0; i<4; i++)
      Q += N[i]*GetControlPoint(i,j);
    umincurve.InsertControlPoint(new double3(Q));  
  }

  BSplineEvaluator.GetNValues(3,N); 
  for (int j=0; j<nc; j++) {
    double3 Q(0,0,0);
    for (int i=nr-4; i<nr; i++)
      Q += N[i-nr+4]*GetControlPoint(i,j);
    umaxcurve.InsertControlPoint(new double3(Q));  
  }

  BSplineEvaluator.GetNValues(2,N); 
  for (int i=0; i<nr; i++) {
    double3 Q(0,0,0);
    for (int j=0; j<4; j++)
      Q += N[j]*GetControlPoint(i,j);
    vmincurve.InsertControlPoint(new double3(Q));  
  }

  BSplineEvaluator.GetNValues(3,N); 
  for (int i=0; i<nr; i++) {
    double3 Q(0,0,0);
    for (int j=nc-4; j<nc; j++)
      Q += N[j-nc+4]*GetControlPoint(i,j);
    vmaxcurve.InsertControlPoint(new double3(Q));  
  }
}

int TensorBSpline::LoadFile(char *FileName)
{
  std::ifstream from(FileName);

  if (!from) {
    std::cerr << "TensorBSpline::LoadFile(): could not open file " << FileName << std::endl;
    return 0;
  }

  from >> numRows >> numCols;

  if (!(from) || (numRows < 0) || (numRows > MAXROWS) || (numCols < 0) || (numCols > MAXCOLS)) {
    std::cerr << "TensorBSpline::LoadFile(): dimension error!" << std::endl;
  }

  for (int i=0; i<numRows*numCols; i++)
    from >> V[i];

  from.close();
  return 1;
}

int TensorBSpline::SaveFile(char *FileName) const
{
  std::ofstream to(FileName);

  if (!to) {
    std::cerr << "TensorBSpline::SaveFile(): could not open file " << FileName << std::endl;
    return 0;
  }

  to << numRows << " " << numCols << std::endl;

  // save control points
  for (int i=0; i<NumControlPoints(); i++)
    to << V[i] << std::endl;

  to.close();
  return 1;
}

void TensorBSpline::Show() const
{
  std::cout << "TensorBSpline::Show():" << std::endl;
  std::cout << "  parameter domain: [" << MinUPar() << "," << MaxUPar() << "] x ["
       <<                            MinVPar() << "," << MaxVPar() << "]" << std::endl;

  std::cout << numRows << " x " << numCols << " control points" << std::endl;
  for (int i=0; i<numRows; i++) {
    for (int j=0; j<numCols; j++)
      std::cout << GetControlPoint(i,j) << " ";
    std::cout << std::endl;
  }
}

