// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

//-----------------------------------------------------------------------//
//                          BiPolynomial                                 //
//-----------------------------------------------------------------------//

#include <bipoly.h>
#include <cmath>
#include <pause.h>

static double Factorial(int n)
{
  double fac = 1;
  for (int i=2; i<=n; i++)
    fac *= n;

  return fac;
}

void BiPolynomial::allocate(int degree, int CopyContents)
// maxn is set to max(maxn,d);
// if necessary new memory is allocated for coefs;
//   if (CopyContents) is true, then the current contents of coefs are copied
{
  if (maxn < degree) {
    maxn = degree;
    double* b = new double[numcoefs(degree)];
    if (CopyContents)
      for (int i=0; i<numcoefs(n); i++) b[i] = coefs[i];
    delete[] coefs;
    coefs=b;
  }
}

void BiPolynomial::raise_degree(int degree)
// n is set to max(degree,n) and the coefficients
// a[i,j], i+j > n are set to 0
// if necessary, new memory is allocated
{
  allocate(degree, 1);
  if (n < degree) {
    for (int i=numcoefs(n); i<numcoefs(degree); i++)
      coefs[i]=0;
    n=degree;
  }
}

BiPolynomial::BiPolynomial(int d)
// constructor; creates the degree d polynomial f(x,y) = 0
{
  maxn = d;
  n = maxn;
  coefs = new double[numcoefs(maxn)];
  for (int i=0; i<numcoefs(n); i++) coefs[i]=0;
}

BiPolynomial::BiPolynomial(const Polynomial& g, const Polynomial& h)
// constructor; creates the polynomial f(x,y) = g(x) h(y)
{
  maxn = g.Degree() + h.Degree();
  n = maxn;  
  coefs = new double[numcoefs(maxn)];
  for (int i=0; i<numcoefs(n); i++) coefs[i]=0;

  for (int i=0; i<=g.Degree(); i++)
    for (int j=0; j<=h.Degree(); j++)
      a(i,j) = g[i]*h[j];
}

BiPolynomial::BiPolynomial(const BiPolynomial& g)
// copy constructor
{
  maxn = g.maxn;
  n = g.n;
  coefs = new double[numcoefs(g.n)];
  for (int i=0; i<numcoefs(g.n); i++) coefs[i] = g.coefs[i];
}

void BiPolynomial::SetDegree(int d)
// sets the degree of f to d; the current contents are lost
{
  allocate(d);
  n=d;
}

void BiPolynomial::Show() const
// writes the polynomial to standard out
{
  std::cout << "degree = " << n << " max degree = " << maxn << std::endl;
  for (int i=0; i<=n; i++) {
    for (int j=0; i+j<=n; j++) {
      std::cout << a(i,j);
      if (i>=1) std::cout << " x";
      if (i>1)  std::cout << "^" << i;
      if (j>=1) std::cout << " y";
      if (j>1)  std::cout << "^" << j;
      std::cout << " + ";
    }
    std::cout << std::endl;
  }
  std::cout << std::endl;
}

double BiPolynomial::operator()(double x, double y)
// returns f(x,y):
//
//               n                             n-i
//     f(x,y) = sum  b(i) x^i   with   b(i) =  sum  a(i,j) y^j
//              i=0                            j=0
{                  
  double fxy = a(n,0); // fxy = b(n)

  for (int i=n-1; i>=0; i--) {
    double bi = a(i,n);
    for (int j=n-i-1; j>=0; j--)
      bi = y*bi + a(i,j);
    fxy = x*fxy + bi;
  }
  
  return fxy;
}

BiPolynomial BiPolynomial::Primitive(int k, int l) const
// returns a polynomial F, such that
//
//             d^(k+l) F
//             ---------   =   f
//             dx^k dy^l
//
// The following polynomial is returned:
//
//           n+k  n-i+k+l  (i-k)! (j-l)!
//  F(x,y) = sum    sum    ------ ------ a(i-k,j-l) x^i y^j
//           i=k    j=l      i!     j!
{
  double *b = new double[numcoefs(n+k+l)];

  for (int i=0; i<numcoefs(n+k+l); i++) b[i]=0;
  
  double ifactor = 1 / Factorial(k);

  for (int i=k; i<=n+k; i++) {
    if (i!=k) ifactor *= (double(i-k)/i); // xfactor = (i-k)!/i!
    double factor = ifactor / Factorial(l);
    for (int j=l; j<=n-i+k+l; j++) {
      if (j!=l) factor *= (double(j-l)/j);  // factor = (i-k)! (j-l)! / (i! j!)
      b[(i+j)*(i+j+1)/2 + i] = a(i-k,j-l)*factor;
    }
  }
        
  return BiPolynomial(b,n+k+l);
}

BiPolynomial BiPolynomial::Derivative(int k, int l) const
// returns the polynomial F, such that
//
//             d^(k+l) f
//             ---------   =   F
//             dx^k dy^l
//
//           n-k  n-i-k-l  (i+k)! (j+l)!
//  F(x,y) = sum    sum    ------ ------ a(i+k,j+l) x^i y^j
//           i=0    j=0      i!     j!
{
  if (n-k-l<0) return BiPolynomial(0);
    
  double* b = new double[numcoefs(n-k-l)];

  double ifactor = Factorial(k);

  for (int i=0; i<=n-k; i++) {
    if (i!=0) ifactor *= (double(i+k)/i); // ifactor = (i+k)!/i!
    double factor = ifactor * Factorial(l);
    for (int j=0; j<=n-i-k-l; j++) {
      if (j!=0) factor *= (double(j+l)/j);  // factor = (i+k)! (j+l)! / (i! j!)
      b[(i+j)*(i+j+1)/2 + i] = a(i+k,j+l)*factor;
    }
  }

  return BiPolynomial(b,n-k-l);
}

double BiPolynomial::Derivative (double x, double y, int k, int l) const
// returns the value
//
//             d^(k+l) f
//             --------- (x,y)
//             dx^k dy^l
{
  BiPolynomial df = Derivative(k,l);
  return df(x,y);
}

double BiPolynomial::Integrate(double f, double t, double a0, double a1, double b0, double b1) const
// returns the following integral:
//
//      t     b0 + b1 x
//      -       -
//     |       |        f(x,t) dt  dx
//    -       -
//    f    a0 + a1 x
//
// The following algorithm is from Michael Franssen.
{
  const double EPS = 1e-50;
  int degree=n;

  double res=0,L=f*f,R=t*t;
  double af=0,bf=0,ff=1,tf=1;
  if (fabs(a1)>EPS) af=a0/a1;
  if (fabs(b1)>EPS) bf=b0/b1;
  if (fabs(f)>EPS) ff=f;
  if (fabs(t)>EPS) tf=t;
  for(int i=0;i<=degree;i++) {
    // L=f^(i+2) R=t^(i+2)
    double A=a1,B=b1,LA=a0,LB=b0;
    double LL=L,RR=R;
    for(int j=0;j<=degree-i;j++) {
      // A=a1^(j+1)    B=b1^(j+1)
      // LA=a0^(j+1)   LB=b0^(j+1)
      // LL=f^(i+j+2)  RR=t^(i+j+2)
      int jf=1;
      double X=0,AA=A,BB=B,LLL=LL,RRR=RR;
      for(int k=0;k<=j;k++) {
        // AA=a0^k*a1^(j+1-k)  BB=b0^k*b1^(j+1-k)
        // jf=j!               f2=k!(j+1-k)!(i+j+2-k)
        // LLL=f^(i+j+2-k)     RRR=t^(i+j+2-k)
        X=X+(BB-AA)*jf*(RRR-LLL)/(i+j+2-k);
        AA*=af;BB*=bf;LLL/=ff;RRR/=tf;
        jf=((j+1-k)*jf)/(k+1);
      }
      X=X+(LB-LA)*(RRR-LLL)/(i+1); // last case : k=j+1
      A*=a1;B*=b1;
      LA*=a0;LB*=b0;
      LL*=f;RR*=t;
      res=res+X*a(i,j)/(j+1);
    }
    L*=f;R*=t;
  }
  
  return res;
}

void BiPolynomial::operator = (const BiPolynomial& g)
{
  n = g.n;
  allocate(n);
  for (int i=0; i<numcoefs(n); i++) coefs[i] = g.coefs[i];
}

void BiPolynomial::operator += (const BiPolynomial& g)
{
  raise_degree(g.n);
  for (int i=0; i<numcoefs(g.n); i++) coefs[i] += g.coefs[i];
}

void BiPolynomial::operator -= (const BiPolynomial& g)
{
  raise_degree(g.n);
  for (int i=0; i<numcoefs(g.n); i++) coefs[i] -= g.coefs[i];
}

void BiPolynomial::operator *= (double t)
{
  for (int i=0; i<numcoefs(n); i++) coefs[i] *= t;
}

void BiPolynomial::operator /= (double t)
{
  for (int i=0; i<numcoefs(n); i++) coefs[i] /= t;
}

BiPolynomial operator - (const BiPolynomial& f)
{
  double* coefs = new double[f.numcoefs(f.n)];
  for (int i=0; i<f.numcoefs(f.n); i++) coefs[i] = -f.coefs[i];
  return BiPolynomial(coefs,f.n);
}

BiPolynomial operator + (const BiPolynomial& f, const BiPolynomial& g)
// a quick hack...
{
  BiPolynomial result = f;
  result += g;
 
  return result;
}

BiPolynomial operator - (const BiPolynomial& f, const BiPolynomial& g)
// a quick hack...
{
  BiPolynomial result = f;
  result -= g;
 
  return result;
}

BiPolynomial operator * (const BiPolynomial& f, const BiPolynomial& g)
{
  BiPolynomial Result(f.Degree() + g.Degree());

  for (int i=0; i<=f.Degree(); i++)
    for (int j=0; j<=f.Degree()-i; j++)
      for (int k=0; k<=g.Degree(); k++)
        for (int l=0; l<=g.Degree()-k; l++)
          Result.a(i+k,j+l) += (f.a(i,j)*g.a(k,l));
  
  return Result;
}

BiPolynomial operator * (double t, const BiPolynomial& f)
// a quick hack...
{
  BiPolynomial result = f;
  result *= t;
 
  return result;
}

BiPolynomial operator * (const BiPolynomial& f, double t)
{
  return t*f;
}

BiPolynomial operator / (const BiPolynomial& f, double t)
// a quick hack...
{
  BiPolynomial result = f;
  result /= t;
 
  return result;
}

std::ostream& operator<<(std::ostream& to, const BiPolynomial& f)
// format: degree coefs[0] coefs[1] ... coefs[numcoefs-1]
{
  to << f.n << " ";
  for (int i=0; i<f.numcoefs(f.n); i++) to << f.coefs[i] << ' ';
  return to;
}

std::istream& operator>>(std::istream& from, BiPolynomial& f)
// format: degree coefs[0] coefs[1] ... coefs[numcoefs-1]
{
  int degree;
  from >> degree;

  f.raise_degree(degree);

  for (int i=0; i<f.numcoefs(degree); i++) from >> f.coefs[i];

  return from;
}

