// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#include <cmath>
#include <iostream>
#include <double3.h>
#include <double3x3.h>

//-----------------------------------------------------------------------//
//                          double3x3                                     //
//-----------------------------------------------------------------------//
double3x3::double3x3(double a00, double a01, double a02,
							double a10, double a11, double a12,
							double a20, double a21, double a22)
{
  A[0][0] = a00;
  A[0][1] = a01;
  A[0][2] = a02;
  A[1][0] = a10;
  A[1][1] = a11;
  A[1][2] = a12;
  A[2][0] = a20;
  A[2][1] = a21;
  A[2][2] = a22;
}

double3x3::double3x3(const double3& row0,
						 const double3& row1,
                   const double3& row2 )
{
  for (int i=0; i<3; i++) {
    A[0][i] = row0[i];
    A[1][i] = row1[i];
    A[2][i] = row2[i];
  }
}               
    
double double3x3::Det() const
{
  return  A[0][0]*A[1][1]*A[2][2] +
          A[0][1]*A[1][2]*A[2][0] +
          A[0][2]*A[1][0]*A[2][1] -
          A[2][0]*A[1][1]*A[0][2] -
          A[2][1]*A[1][2]*A[0][0] -
          A[2][2]*A[1][0]*A[0][1];
}

double3x3 double3x3::Transpose() const
{
  return double3x3(Col(0), Col(1), Col(2));
}

double double3x3::FrobeniusNorm() const
{
  double norm=0;

  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      norm += A[i][j]*A[i][j];
  
  return sqrt(norm);
}

void double3x3::Fill(double fill)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      A[i][j] = fill;
}

void double3x3::MakeSymmetric()
{
  for (int i=0; i<3; i++)
    for (int j=i+1; j<3; j++) {
      double t = (A[i][j] + A[j][i])/2;
	  A[i][j] = A[j][i] = t;
    }
}

void double3x3::SwapCols(int col1, int col2)
{
  if (col1 != col2) {
    for (int i=0; i<3; i++) {
      double temp = A[i][col1];
      A[i][col1] = A[i][col2];
      A[i][col2] = temp;
    }
  }
}

void double3x3::SwapRows(int row1, int row2)
{
  if (row1 != row2) {
    for (int i=0; i<3; i++) {
       double temp = A[row1][i];
       A[row1][i] = A[row2][i];
       A[row2][i] = temp;
    }
  }
}

void double3x3::MultCol(int col, double t)
{
  for (int i=0; i<3; i++) A[i][col] *= t;
}

void double3x3::MultRow(int row, double t)
{
  for (int i=0; i<3; i++) A[row][i] *= t;
}

void double3x3::operator=(const double3x3& B)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      A[i][j] = B.A[i][j];
}

void double3x3::operator+=(const double3x3& B)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      A[i][j] += B.A[i][j];
}

void double3x3::operator-=(const double3x3& B)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      A[i][j] -= B.A[i][j];
}

void double3x3::operator*=(double t)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      A[i][j] *= t;
}

void double3x3::operator/=(double t)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      A[i][j] /= t;
}

double3x3 operator-(const double3x3& A)
{
  double3x3 B;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      B[i][j] = A[i][j];
  return B;
}

double3x3 operator+(const double3x3& A, const double3x3& B)
{
  double3x3 C;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C[i][j] = A[i][j] + B[i][j];
  return C;
}

double3x3 operator-(const double3x3& A, const double3x3& B)
{
  double3x3 C;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C[i][j] = A[i][j] - B[i][j];
  return C;
}

double3x3 operator*(double t, const double3x3& A)
{
  double3x3 C;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C[i][j] = t*A[i][j];
  return C;
}

double3x3 operator/(const double3x3& A, double t)
{
  double3x3 C;
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      C[i][j] = A[i][j]/t;
  return C;
}

double3x3 operator*(const double3x3& A, const double3x3& B)
{
  double3x3 C;
  C.MakeZero();
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      for (int k=0; k<3; k++)
        C[i][j] += A[i][k]*B[k][j];
  return C;
}

std::ostream& operator<<(std::ostream& s, const double3x3& A)
{
  s << std::endl;
  for (int i=0; i<3; i++) {
    int j;
    for (j=0; j<2; j++)
      s << A[i][j] << ", ";
    s << A[i][j] << std::endl;
  }
  return s;
}

std::istream& operator>>(std::istream& s, double3x3& A)
{
  for (int i=0; i<3; i++) 
    for (int j=0; j<3; j++)
      s >> A[i][j];
  return s;
}

double3x3 DiagonalMatrix(const double3& x)
{
  double3x3 A;
  A.MakeZero();
  for (int i=0; i<3; i++)
    A[i][i] = x[i];

  return A;
}

double3x3 DyadeMatrix(const double3& x, const double3& y)
{
  double3x3 A;
  for (int i=0; i<3; i++) 
    for (int j=0; j<3; j++)
      A[i][j] = x[i]*y[j];

  return A;
}



// A set to axial rotation matrix
// precondition: |v|=1 
double3x3& double3x3::AxRot(const double3& v, double phi)
{ double S[3][3], sst[3][3], cosphi, sinphi, omcosphi;

  cosphi=(double)cos((double)phi);
  sinphi=(double)sin((double)phi);
  omcosphi=1.0-cosphi;

  S[0][0]=  0.0;  S[0][1]= -v[2]; S[0][2]=  v[1];
  S[1][0]=  v[2]; S[1][1]=  0.0;  S[1][2]= -v[0];
  S[2][0]= -v[1]; S[2][1]=  v[0]; S[2][2]=  0.0;

  sst[0][0]= v[0]*v[0]; sst[0][1]= v[0]*v[1]; sst[0][2]= v[0]*v[2];
  sst[1][0]= v[1]*v[0]; sst[1][1]= v[1]*v[1]; sst[1][2]= v[1]*v[2];
  sst[2][0]= v[2]*v[0]; sst[2][1]= v[2]*v[1]; sst[2][2]= v[2]*v[2];

  A[0][0]= cosphi + omcosphi*sst[0][0] + sinphi*S[0][0];
  A[0][1]=          omcosphi*sst[0][1] + sinphi*S[0][1];
  A[0][2]=          omcosphi*sst[0][2] + sinphi*S[0][2];

  A[1][0]=          omcosphi*sst[1][0] + sinphi*S[1][0];
  A[1][1]= cosphi + omcosphi*sst[1][1] + sinphi*S[1][1];
  A[1][2]=          omcosphi*sst[1][2] + sinphi*S[1][2];

  A[2][0]=          omcosphi*sst[2][0] + sinphi*S[2][0];
  A[2][1]=          omcosphi*sst[2][1] + sinphi*S[2][1];
  A[2][2]= cosphi + omcosphi*sst[2][2] + sinphi*S[2][2];

  return(*this);
}
