// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// 8-5-95

#include <cmath>
#include <iostream>
#include <double4.h>
#include <double4x4.h>

//-----------------------------------------------------------------------//
//                          double4x4                                     //
//-----------------------------------------------------------------------//

double4x4::double4x4(const double4& row0, 
                   const double4& row1, 
                   const double4& row2, 
                   const double4& row3 )
{
  for (int i=0; i<4; i++) {
    A[0][i] = row0[i];
    A[1][i] = row1[i];
    A[2][i] = row2[i];
    A[3][i] = row3[i];
  }
}               
    
double4x4 double4x4::Transpose() const
{
  return double4x4(Col(0), Col(1), Col(2), Col(3));
}

double double4x4::FrobeniusNorm() const
{
  double norm = 0;

  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      norm += A[i][j]*A[i][j];

  return sqrt(norm);
}

void double4x4::Fill(double fill)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      A[i][j] = fill;
}

void double4x4::MakeSymmetric()
{
  for (int i=0; i<4; i++)
    for (int j=i+1; j<4; j++) {
      double t = (A[i][j] + A[j][i])/2;
	  A[i][j] = A[j][i] = t;
    }
}

void double4x4::SwapCols(int col1, int col2)
{
  if (col1 != col2) {
    for (int i=0; i<4; i++) {
      double temp = A[i][col1];
      A[i][col1] = A[i][col2];
      A[i][col2] = temp;
    }
  }
}

void double4x4::SwapRows(int row1, int row2)
{
  if (row1 != row2) {
    for (int i=0; i<4; i++) {
       double temp = A[row1][i];
       A[row1][i] = A[row2][i];
       A[row2][i] = temp;
    }
  }
}

void double4x4::MultCol(int col, double t)
{
  for (int i=0; i<4; i++) A[i][col] *= t;
}

void double4x4::MultRow(int row, double t)
{
  for (int i=0; i<4; i++) A[row][i] *= t;
}

void double4x4::operator=(const double4x4& B)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      A[i][j] = B.A[i][j];
}

void double4x4::operator+=(const double4x4& B)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      A[i][j] += B.A[i][j];
}

void double4x4::operator-=(const double4x4& B)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      A[i][j] -= B.A[i][j];
}

void double4x4::operator*=(double t)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      A[i][j] *= t;
}

void double4x4::operator/=(double t)
{
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      A[i][j] /= t;
}

double4x4 operator-(const double4x4& A)
{
  double4x4 B;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      B[i][j] = A[i][j];
  return B;
}

double4x4 operator+(const double4x4& A, const double4x4& B)
{
  double4x4 C;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      C[i][j] = A[i][j] + B[i][j];
  return C;
}

double4x4 operator-(const double4x4& A, const double4x4& B)
{
  double4x4 C;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      C[i][j] = A[i][j] - B[i][j];
  return C;
}

double4x4 operator*(double t, const double4x4& A)
{
  double4x4 C;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      C[i][j] = t*A[i][j];
  return C;
}

double4x4 operator/(const double4x4& A, double t)
{
  double4x4 C;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      C[i][j] = A[i][j]/t;
  return C;
}

double4x4 operator*(const double4x4& A, const double4x4& B)
{
  double4x4 C;
  C.MakeZero();
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      for (int k=0; k<4; k++)
        C[i][j] += A[i][k]*B[k][j];
  return C;
}

std::ostream& operator<<(std::ostream& s, const double4x4& A)
{
  s << std::endl;
  for (int i=0; i<4; i++) {
    int j;
    for (j=0; j<3; j++)
      s << A[i][j] << ", ";
    s << A[i][j] << std::endl;
  }
  return s;
}

std::istream& operator>>(std::istream& s, double4x4& A)
{
  for (int i=0; i<4; i++) 
    for (int j=0; j<4; j++)
      s >> A[i][j];
  return s;
}

double4x4 DiagonalMatrix(const double4& x)
{
  double4x4 A;
  A.MakeZero();
  for (int i=0; i<4; i++)
    A[i][i] = x[i];

  return A;
}

double4x4 DyadeMatrix(const double4& x, const double4& y)
{
  double4x4 A;
  for (int i=0; i<4; i++) 
    for (int j=0; j<4; j++)
      A[i][j] = x[i]*y[j];

  return A;
}


// A set to axial rotation matrix
// precondition: |v|=1 
double4x4& double4x4::AxRot(const double3& v, double phi)
{ double S[3][3], sst[3][3], cosphi, sinphi, omcosphi;

  cosphi=(double)cos((double)phi);
  sinphi=(double)sin((double)phi);
  omcosphi=1.0-cosphi;

  S[0][0]=  0.0;  S[0][1]= -v[2]; S[0][2]=  v[1];
  S[1][0]=  v[2]; S[1][1]=  0.0;  S[1][2]= -v[0];
  S[2][0]= -v[1]; S[2][1]=  v[0]; S[2][2]=  0.0;

  sst[0][0]= v[0]*v[0]; sst[0][1]= v[0]*v[1]; sst[0][2]= v[0]*v[2];
  sst[1][0]= v[1]*v[0]; sst[1][1]= v[1]*v[1]; sst[1][2]= v[1]*v[2];
  sst[2][0]= v[2]*v[0]; sst[2][1]= v[2]*v[1]; sst[2][2]= v[2]*v[2];

  A[0][0]= cosphi + omcosphi*sst[0][0] + sinphi*S[0][0];
  A[0][1]=          omcosphi*sst[0][1] + sinphi*S[0][1];
  A[0][2]=          omcosphi*sst[0][2] + sinphi*S[0][2];
  A[0][3]= 0.0;

  A[1][0]=          omcosphi*sst[1][0] + sinphi*S[1][0];
  A[1][1]= cosphi + omcosphi*sst[1][1] + sinphi*S[1][1];
  A[1][2]=          omcosphi*sst[1][2] + sinphi*S[1][2];
  A[1][3]= 0.0;

  A[2][0]=          omcosphi*sst[2][0] + sinphi*S[2][0];
  A[2][1]=          omcosphi*sst[2][1] + sinphi*S[2][1];
  A[2][2]= cosphi + omcosphi*sst[2][2] + sinphi*S[2][2];
  A[2][3]= 0.0;

  A[3][0]= 0.0;
  A[3][1]= 0.0;
  A[3][2]= 0.0;
  A[3][3]= 1.0;

  return(*this);
}
