// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Thu 05-30-1996

#include <iostream>
#include <cmath>
#include <equation.h>

const double PRECISION = 1e-50;

//-----------------------------------------------------------------------//
//                          LinearEquation                               //
//-----------------------------------------------------------------------//

void LinearEquation::GetMatrix(DoubleMatrix& A) const
// returns the matrix C
{
  A.SetSize(NumRows(), NumCols());
  AddToMatrix(A, 0);
}

void LinearEquation::GetVector(DoubleVector& b) const
// returns the vector d
{
  b.SetSize(NumRows());
  AddToVector(b, 0);
}

void LinearEquation::Check(const DoubleVector& x) const
// checks if the equations are satisfied for x and reports this
{
  DoubleMatrix C;
  DoubleVector d;
  GetMatrix(C);
  GetVector (d);
  
  std::cout << "|Cx-d| = " << (C*x-d).Length() << std::endl;
}

void LinearEquation::Show() const
// writes the contents to standard out
{
  DoubleMatrix C;
  DoubleVector d;
  GetMatrix(C);
  GetVector (d);
  
  std::cout << "C = " << C << std::endl;
  std::cout << "d = " << d << std::endl;
}

//-----------------------------------------------------------------------//
//                          PointConstraintEquation                      //
//-----------------------------------------------------------------------//

int PointConstraintEquation::NumRows() const
{
  return 3;
}

int PointConstraintEquation::NumCols() const
{
  return 3*n;
}

void PointConstraintEquation::AddToMatrix(DoubleMatrix& A, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    A(row  , col  ) += a(i);
    A(row+1, col+1) += a(i);
    A(row+2, col+2) += a(i);
  }
}

void PointConstraintEquation::AddToVector(DoubleVector& b, int row) const
{
  b(row  ) += P.x();
  b(row+1) += P.y();
  b(row+2) += P.z();
}

void PointConstraintEquation::AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(row  ) += a(i) * x(col  );
    result(row+1) += a(i) * x(col+1);
    result(row+2) += a(i) * x(col+2);
  }
}

void PointConstraintEquation::AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(col  ) += a(i) * x(row  );
    result(col+1) += a(i) * x(row+1);
    result(col+2) += a(i) * x(row+2);
  }
}

//-----------------------------------------------------------------------//
//                          TangentConstraintEquation                    //
//-----------------------------------------------------------------------//
// This class represents the linear system of equations
//
//             sum       a(i) (P(i)^P) = 0,
//          i in index
//
// where P(i), 0 <= i < n is a set of 3D points.
// and v0^v1 parallel to P

int TangentConstraintEquation::NumRows() const
{
  return 2;
}

int TangentConstraintEquation::NumCols() const
{
  return 3*n;
}

void TangentConstraintEquation::AddToMatrix(DoubleMatrix& A, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    A(row, col  )   += a(i) * v0.x();
    A(row, col+1)   += a(i) * v0.y();
    A(row, col+2)   += a(i) * v0.z();
    A(row+1, col  ) += a(i) * v1.x();
    A(row+1, col+1) += a(i) * v1.y();
    A(row+1, col+2) += a(i) * v1.z();
  }
}

void TangentConstraintEquation::AddToVector(DoubleVector& b, int row) const
{
}

void TangentConstraintEquation::AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(row)   += a(i) * (v0.x()*x(col) + v0.y()*x(col+1) + v0.z()*x(col+2));
    result(row+1) += a(i) * (v1.x()*x(col) + v1.y()*x(col+1) + v1.z()*x(col+2));
  }
}
 
void TangentConstraintEquation::AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(col  ) += a(i) * ( x(row)*v0.x() + x(row+1)*v1.x() );
    result(col+1) += a(i) * ( x(row)*v0.y() + x(row+1)*v1.y() );
    result(col+2) += a(i) * ( x(row)*v0.z() + x(row+1)*v1.z() );
  }
}

//-----------------------------------------------------------------------//
//                          ControlPointConstraintEquation               //
//-----------------------------------------------------------------------//

int ControlPointConstraintEquation::NumRows() const
{
  return 3*index.Size();
}
 
int ControlPointConstraintEquation::NumCols() const
{
  return 3*n;
}
 
void ControlPointConstraintEquation::AddToMatrix(DoubleMatrix& A, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index(i);
    A(row++, col  ) += 1;
    A(row++, col+1) += 1;
    A(row++, col+2) += 1;
  }
}
 
void ControlPointConstraintEquation::AddToVector(DoubleVector& b, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int j = row + 3*i;
    b(j)   += points[i].x();
    b(j+1) += points[i].y();
    b(j+2) += points[i].z();
  }
}

void ControlPointConstraintEquation::AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index(i);
    result(row++) += x(col  );
    result(row++) += x(col+1);
    result(row++) += x(col+2);
  }
}
 
void ControlPointConstraintEquation::AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index(i);
    result(col  ) += x(row++);
    result(col+1) += x(row++);
    result(col+2) += x(row++);
  }
}
 
//-----------------------------------------------------------------------//
//                          NormalConstraintEquation                     //
//-----------------------------------------------------------------------//

int NormalConstraintEquation::NumRows() const
{
  return 2;
}

int NormalConstraintEquation::NumCols() const
{
  return 3*n;
}

void NormalConstraintEquation::AddToMatrix(DoubleMatrix& A, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    A(row  , col  ) += a(i) * normal.x();
    A(row  , col+1) += a(i) * normal.y();
    A(row  , col+2) += a(i) * normal.z();
    A(row+1, col  ) += b(i) * normal.x();
    A(row+1, col+1) += b(i) * normal.y();
    A(row+1, col+2) += b(i) * normal.z();
  }
}

void NormalConstraintEquation::AddToVector(DoubleVector&, int row) const
{
}

void NormalConstraintEquation::AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(row  ) += a(i) * (normal.x()*x(col) + normal.y()*x(col+1) + normal.z()*x(col+2));
    result(row+1) += b(i) * (normal.x()*x(col) + normal.y()*x(col+1) + normal.z()*x(col+2));
  }
}

void NormalConstraintEquation::AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(col  ) += (a(i)*x(row)*normal.x() + b(i)*x(row+1)*normal.x());
    result(col+1) += (a(i)*x(row)*normal.y() + b(i)*x(row+1)*normal.y());
    result(col+2) += (a(i)*x(row)*normal.z() + b(i)*x(row+1)*normal.z());
  }
}

//-----------------------------------------------------------------------//
//                          LineConstraintEquation                       //
//-----------------------------------------------------------------------//

int LineConstraintEquation::NumRows() const
{
  return 2;
}

int LineConstraintEquation::NumCols() const
{
  return 3*n;
}

void LineConstraintEquation::AddToMatrix(DoubleMatrix& A, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    A(row  , col  ) += a(i) * V0.n.x();
    A(row  , col+1) += a(i) * V0.n.y();
    A(row  , col+2) += a(i) * V0.n.z();
    A(row+1, col  ) += a(i) * V1.n.x();
    A(row+1, col+1) += a(i) * V1.n.y();
    A(row+1, col+2) += a(i) * V1.n.z();
  }
}

void LineConstraintEquation::AddToVector(DoubleVector& b, int row) const
{
  b(row  ) += V0.d;
  b(row+1) += V1.d;
}

void LineConstraintEquation::AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(row  ) += a(i) * (V0.n.x()*x(col) + V0.n.y()*x(col+1) + V0.n.z()*x(col+2));
    result(row+1) += a(i) * (V1.n.x()*x(col) + V1.n.y()*x(col+1) + V1.n.z()*x(col+2));
  }
}

void LineConstraintEquation::AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(col  ) += a(i) * (x(row)*V0.n.x() + x(row+1)*V1.n.x());
    result(col+1) += a(i) * (x(row)*V0.n.y() + x(row+1)*V1.n.y());
    result(col+2) += a(i) * (x(row)*V0.n.z() + x(row+1)*V1.n.z());
  }
}

//-----------------------------------------------------------------------//
//                          PlaneConstraintEquation                      //
//-----------------------------------------------------------------------//

int PlaneConstraintEquation::NumRows() const
{
  return 1;
}

int PlaneConstraintEquation::NumCols() const
{
  return 3*n;
}

void PlaneConstraintEquation::AddToMatrix(DoubleMatrix& A, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    A(row, col  ) += a(i) * plane.n.x();
    A(row, col+1) += a(i) * plane.n.y();
    A(row, col+2) += a(i) * plane.n.z();
  }
}

void PlaneConstraintEquation::AddToVector(DoubleVector& b, int row) const
{
  b(row) += plane.d;
}

void PlaneConstraintEquation::AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(row) += a(i) * (plane.n.x()*x(col) + plane.n.y()*x(col+1) + plane.n.z()*x(col+2));
  }
}

void PlaneConstraintEquation::AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<index.Size(); i++) {
    int col = 3*index[i];
    result(col  ) += a(i) * x(row)*plane.n.x();
    result(col+1) += a(i) * x(row)*plane.n.y();
    result(col+2) += a(i) * x(row)*plane.n.z();
  }
}

//-----------------------------------------------------------------------//
//                          CurveConstraintEquation                      //
//-----------------------------------------------------------------------//

int CurveConstraintEquation::NumRows() const
{
  return 3*index.Count;
}

int CurveConstraintEquation::NumCols() const
{
  return 3*n;
}

void CurveConstraintEquation::AddToMatrix(DoubleMatrix& A, int row) const
{
  for (int i=0; i<Q.Count; i++)
    for (int j=0; j<a.Size(); j++) {
      int col = 3*index[i][j];
      A(row+3*i  , col  ) += a(j);
      A(row+3*i+1, col+1) += a(j);
      A(row+3*i+2, col+2) += a(j);
    }
}

void CurveConstraintEquation::AddToVector(DoubleVector& b, int row) const
{
  for (int i=0; i<Q.Count; i++) {
    b(row+3*i  ) += Q[i].x();
    b(row+3*i+1) += Q[i].y();
    b(row+3*i+2) += Q[i].z();
  }
}

void CurveConstraintEquation::AddToMatrixVector(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<Q.Count; i++)
    for (int j=0; j<a.Size(); j++) {
      int col = 3*index[i][j];     
      result(row+3*i  ) += a(j) * x(col  );
      result(row+3*i+1) += a(j) * x(col+1);
      result(row+3*i+2) += a(j) * x(col+2);
    }
}

void CurveConstraintEquation::AddToVectorMatrix(const DoubleVector& x, DoubleVector& result, int row) const
{
  for (int i=0; i<Q.Count; i++)
    for (int j=0; j<a.Size(); j++) {
      int col = 3*index[i][j];     
      result(col  ) += a(j) * x(row+3*i  );
      result(col+1) += a(j) * x(row+3*i+1);
      result(col+2) += a(j) * x(row+3*i+2);
    }
}

void CurveConstraintEquation::Show() const
{
  std::cout << "CurveConstraint: " << std::endl;
  std::cout << "  n = " << n << std::endl;
  std::cout << "  a = " << a << std::endl;
  for (int i=0; i<index.Count; i++)
     std::cout << index[i] << "  " << Q[i] << std::endl;
}

//-----------------------------------------------------------------------//
//                          LinearEquationSystem                         //
//-----------------------------------------------------------------------//

int LinearEquationSystem::NumRows() const
{
  int Result = 0;
  for (int i=0; i<eqn.Count; i++)
    Result += eqn[i].NumRows();
  
  return Result;
}

int LinearEquationSystem::NumCols() const
{
  if (eqn.Count>0)
    return eqn[0].NumCols();

  else
    return 0;
}

void LinearEquationSystem::AddToMatrix(DoubleMatrix& A) const
{
  int row=0;
  for (int i=0; i<eqn.Count; i++) {
    eqn[i].AddToMatrix(A,row);
    row += eqn[i].NumRows();
  }  
}

void LinearEquationSystem::GetVector(DoubleVector& b) const
// returns the vector d
{
  b.SetSize(NumRows());
  
  int row=0;
  for (int i=0; i<eqn.Count; i++) {
    eqn[i].AddToVector(b,row);
    row += eqn[i].NumRows();
  }  
}

DoubleVector LinearEquationSystem::GetMatrixVectorProduct(const DoubleVector& x) const
// returns the vector x^T C
{
  DoubleVector Result(NumRows());
  Result.MakeZero();
  
  int row=0;
  for (int i=0; i<eqn.Count; i++) {
    eqn[i].AddToMatrixVector(x, Result, row);
    row += eqn[i].NumRows();
  }  

  return Result;
}

DoubleVector LinearEquationSystem::GetVectorMatrixProduct(const DoubleVector& x) const
// returns the vector Cx
{
  DoubleVector Result(NumCols());
  Result.MakeZero();
  
  int row=0;
  for (int i=0; i<eqn.Count; i++) {
    eqn[i].AddToVectorMatrix(x, Result, row);
    row += eqn[i].NumRows();
  }  

  return Result;
}  

void LinearEquationSystem::GetEliminateForm(DoubleMatrix& F, DoubleVector& G, IntVector& index) const
// Computes a matrix F and a vector G, such that the original system
//
//                          Cx = d
// is equivalent to
//                           x = Fy + G
//
// The vector index contains the indices of the remaining columns,
// so y = (x(index(0)), x(index(1)), ...);
{
  LinearSystem sys;
  GetMatrix(sys.A);
  GetVector(sys.b);
  sys.GetEliminateForm(F, G, index);
}

void LinearEquationSystem::TestEliminateForm(const DoubleVector& x) const
// checks if F and g satisfy the following relations:
//
//           CF = 0   &&   Cg = d
{
  LinearSystem sys;
  GetMatrix(sys.A);
  GetVector(sys.b);
  sys.TestEliminateForm(x);
}

void LinearEquationSystem::Check(const DoubleVector& x) const
{
  for (int i=0; i<eqn.Count; i++) {
    std::cout << "equation " << i << " ";
    eqn[i].Check(x);
  }  
}

void LinearEquationSystem::Show() const
// writes the contents to standard out
{
  DoubleMatrix C0;
  GetMatrix(C0);
  DoubleVector d0;
  GetVector(d0);
  
  std::cout << "C = " << C0 << std::endl;
  std::cout << "d = " << d0 << std::endl;
}

//-----------------------------------------------------------------------//
//                          LinearSystem                                 //
//-----------------------------------------------------------------------//

int LinearSystem::GetPivotElement(int& row, int& col, const IntVector& done)
// Searches for the element in the matrix A with maximum absolute value
// The vector done has the following interpretation:
//   if (done[i] >= 0), the element a(i,done[i]) has already been used
//   as a pivot element, so this row will be skipped
// If a nonzero pivot element is found true is returned.
// The vector done is not updated!
{
  int m = A.NumRows();
  int n = A.NumCols();

  double max = 0;
  for (int i=0; i<m; i++) {
    if (done[i] >= 0)         // row i has already been used
      continue;
    for (int j=0; j<n; j++) {
      if (fabs(A(i,j)) > max) {
        max = fabs(A(i,j));
        row = i;
        col = j;
      }
    }
  }

  return (max > PRECISION);
}

void LinearSystem::TestClearColumn(int row, int col, const DoubleVector& x)
// by elementary row operations the column col is transformed into the
// unitvector e[row]
//
// precondition: |A(row,col)| > 0
{
std::cout << "TestClearColumn: |Ax-b| = " << (A*x-b).Length() << std::endl;

  int m = A.NumRows();

  double factor = 1/A(row,col);
  A.MultRow(factor, row);
  b(row) *= factor;

  for (int i=0; i<m; i++) {
    if ((i == row) || (fabs(A(i,col)) < PRECISION))
      continue;
    double factor = -A(i,col);
    A.MultAddRow(factor, row, i);
    b(i) += (factor*b(row));
  }
std::cout << "result of TestClearColumn: |Ax-b| = " << (A*x-b).Length() << std::endl;
}

void LinearSystem::ClearColumn(int row, int col)
// by elementary row operations the column col is transformed into the
// unitvector e[row]
//
// precondition: |A(row,col)| > 0
{
  int m = A.NumRows();
 
  double factor = 1/A(row,col);
  A.MultRow(factor, row);
  b(row) *= factor;
 
  for (int i=0; i<m; i++) {
    if ((i == row) || (fabs(A(i,col)) < PRECISION))
      continue;
    double factor = -A(i,col);
    A.MultAddRow(factor, row, i);
    b(i) += (factor*b(row));
  }
}

int LinearSystem::Clear(IntVector& done)
// By elementary row operations and permutations of columns, the matrix
// is transformed into a matrix that contains the m columns of the
// m-dimensional unit matrix.
//
// postcondition:
// The vector done contains the positions of these columns.
//
// if something goes wrong false is returned
{
  int m = A.NumRows();
  
  if (m < 1) {
    std::cerr << "cannot clear an empty system!" << std::endl;
    return (0==1);
  }
 
  done.SetSize(m);
  done.Fill(-1);
 
  int row, col;
 
  for (int i=0; i<m; i++) {

    if (!GetPivotElement(row,col,done)) {
      std::cerr << "Could not find a suitable pivot element in LinearSystem::Clear()!" << std::endl;
      return (0==1);
    }
 
    ClearColumn(row, col);
    done[row] = col;
  }
 
  return (0==0);
}

int LinearSystem::TestClear(IntVector& done, const DoubleVector& x)
// Performs the same actions as Clear and prints inbetween results
{
std::cout << "TestClear: |Ax-b| = " << (A*x-b).Length() << std::endl;
  int m = A.NumRows();
  
  if (m < 1) {
    std::cerr << "cannot clear an empty system!" << std::endl;
    return (0==1);
  }

  done.SetSize(m);
  done.Fill(-1);

  int row, col;

  for (int i=0; i<m; i++) {
    if (!GetPivotElement(row,col,done)) {
      std::cerr << "Could not find a suitable pivot element in LinearEquationSystem::Clear()!" << std::endl;
      return (0==1);
    }
std::cout << "pivot = " << row << " " << col << " " << A(row,col) << std::endl;
    TestClearColumn(row, col, x);
    done[row] = col;
  }

std::cout << "result of TestClear: |Ax-b| = " << (A*x-b).Length() << std::endl;

  return (0==0);
}

void LinearSystem::GetEliminateForm(DoubleMatrix& F, DoubleVector& G, IntVector& index)
// Computes a matrix F and a vector G, such that the original system
//
//                          Ax = b
// is equivalent to
//                           x = Fy + G
//
// The vector index contains the indices of the remaining columns,
// so y = (x(index(0)), x(index(1)), ...);
{
  int m = A.NumRows();
  int n = A.NumCols();

  IntVector done;
  Clear(done);

  IntVector keep = done;
  keep.Sort();

  // compute the indices of the remaining columns
  index.SetSize(n-m);
  {
    int idx = 0;
    int j;

    for (j=0; j<keep(0); j++)
      index(idx++) = j;

    for (int i=0; i<m-1; i++)
      for (j=keep(i)+1; j<keep(i+1); j++)
        index(idx++) = j;

    for (j=keep(m-1)+1; j<n; j++)
      index(idx++) = j;
  }

  // build the matrix F and the vector G
  F.SetSize(n,n-m);
  G.SetSize(n);
  {
    F.MakeZero();
    G.MakeZero();

    for (int i=0; i<m; i++) {
      int row = done(i);
      for (int j=0; j<index.Size(); j++)
        F(row,j) = -A(i,index(j));
      G(row) = b(i);
    }

    for (int i=0; i<index.Size(); i++)
      F(index(i),i) = 1;
  }
}

void LinearSystem::TestEliminateForm(const DoubleVector& x)
// Performs the same actions as GetSimpleForm and does some testing.
// x is a vector that satisfies Ax=b
{
  int m = A.NumRows();
  int n = A.NumCols();

  DoubleMatrix A0 = A;
  DoubleVector b0 = b;

  IntVector done;
  TestClear(done,x);

std::cout << "done = " << done << std::endl;

  IntVector keep = done;
  keep.Sort();
std::cout << "keep = " << keep << std::endl;

  // compute the indices of the remaining columns
  IntVector index(n-m);
  {
    int idx = 0;
    int j;

    for (j=0; j<keep(0); j++)
      index(idx++) = j;

    for (int i=0; i<m-1; i++)
      for (j=keep(i)+1; j<keep(i+1); j++)
        index(idx++) = j;

    for (j=keep(m-1)+1; j<n; j++)
      index(idx++) = j;
  }
std::cout << "index = " << index << std::endl;

  // build the matrix F and the vector G
  DoubleMatrix F(n,n-m);
  DoubleVector G(n);
  {
    F.MakeZero();
    G.MakeZero();

    for (int i=0; i<m; i++) {
      int row = done(i);
      for (int j=0; j<index.Size(); j++)
        F(row,j) = -A(i,index(j));
      G(row) = b(i);
    }

    for (int i=0; i<index.Size(); i++)
      F(index(i),i) = 1;
  }

  // perform the tests
  std::cout << "||AF|| = " << (A0*F).FrobeniusNorm() << std::endl;
  std::cout << "||AG - b|| = " << (A0*G - b0).Length() << std::endl;
}

void LinearSystem::Show() const
// writes the contents to standard out
{
  std::cout << "Linear system Ax=b with" << std::endl;
  std::cout << "A = " << A << std::endl;
  std::cout << "b = " << b << std::endl;
}

