// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sun 06-09-1996

#include <poly.h>
#include <cmath>

//-----------------------------------------------------------------------//
//                          Polynomial                                   //
//-----------------------------------------------------------------------//

static double Factorial(int n)
{
  double fac = 1;
  for (int i=2; i<=n; i++)
    fac *= n;

  return fac;
}

void Polynomial::allocate(int degree, int CopyContents)
// maxn is set to max(maxn,d);
// if necessary new memory is allocated for a;
//   if (CopyContents) is true, then the current contents of a are copied
{
  if (maxn < degree) {
    maxn = degree;
    double* b = new double[numcoefs(degree)];
    if (CopyContents)
      for (int i=0; i<numcoefs(n); i++) b[i] = a[i];
    delete[] a;
    a=b;
  }
}

void Polynomial::raise_degree(int degree)
// n is set to max(degree,n) and the coefficients
// a[i], i>n are set to 0
// if necessary, new memory is allocated
{
  allocate(degree, 1);
  if (n < degree) {
    for (int i=numcoefs(n); i<numcoefs(degree); i++)
      a[i]=0;
    n=degree;
  }
}

Polynomial::Polynomial(int d)
// constructor; creates the degree d polynomial f(x) = 0
{
  maxn = d;
  n = d;
  a = new double(numcoefs(d));
  for (int i=0; i<numcoefs(d); i++) a[i]=0;
}

Polynomial::Polynomial(const Polynomial& g)
// copy constructor
{
  n = g.n;
  allocate(n); 
  for (int i=0; i<numcoefs(n); i++) a[i] = g.a[i];
}

void Polynomial::SetDegree(int d)
// sets the degree of f to d; the current contents are lost
{
  allocate(d);
  n=d;
}

void Polynomial::Show() const
// writes the polynomial to standard out
{
  std::cout << a[0];
  for (int i=1; i<=n; i++) {
    std::cout << " + " << a[i] << " t^" << i;
  }
  std::cout << std::endl;
}

double Polynomial::operator()(double t) const
// returns f(t)
{
  double result = a[n];
  
  for (int i=n-1; i>=0; i--)
    result = t*result + a[i];
    
  return result;
}

Polynomial Polynomial::Primitive(int k) const
// returns a polynomial F, such that
//
//             d^k F
//             ------   =   f
//             dx^k
//
// The following polynomial is returned:
//
//         n+k   (i-k)!
//  F(t) = sum   ------  a[i-k]  t^i
//         i=k     i!
{
  double *b = new double[numcoefs(n+k)];

  for (int i=0; i<k; i++) b[i]=0;
  
  double factor = 1 / Factorial(k);
  b[k] = a[0]*factor;
  for (int i=k+1; i<=n+k; i++) {
    factor *= (double(i-k)/i);  // factor = (i-k)!/i!
    b[i] = factor*a[i-k];
  }

  return Polynomial(b,n+k);
}

Polynomial Polynomial::Derivative(int k) const
// returns the k-th derivative of f:
//
//   (k)      n-k    (i+k)!
//  f   (t) = sum    ------  a[i+k]  t^i
//            i=0      i!
{
  if (k>n) return Polynomial(0);

  double* b = new double[numcoefs(n-k)];

  double factor = Factorial(k);
  
  b[0] = factor*a[k];
  for (int i=1; i<=n-k; i++) {
    factor *= (double(i+k)/i);  // factor = (i+k)!/i!
    b[i] = factor*a[i+k];
  }
  
  return Polynomial(b,n-k);
}

double Polynomial::Derivative (double t, int k) const
//   (k)      n-k    (i+k)!
//  f   (t) = sum    ------  a[i+k] t^i
//            i=0      i!
{
  if (k>n) return 0;

  double factor = Factorial(k);
  double result = factor*a[k];
  for (int i=1; i<=n-k; i++) {
    factor *= (t*double(i+k)/i);  // factor = t^i*(i+k)!/i!
    result += a[i+k]*factor;
  }
  
  return result;
}

double Polynomial::Integrate(double t0, double t1) const
//        n+1  a[i-1]
// F(t) = sum  ------ t^i
//        i=1    i
{
  double F0 = a[n]/(n+1);
  double F1 = F0;
  
  for (int i=n; i>=1; i--) {
    F0 *= t1; F0 += a[i-1]/i;
    F1 *= t0; F1 += a[i-1]/i;
  }

  F0 *= t0; F1 *= t1;

  return F1 -F0;
}

double Polynomial::Integrate(double x0, double x1, double a0, double a1, double b0, double b1) const
// returns the following integral:
//
//          x1   b0 + b1 x
//          -       -
//  I =    |       |        f(t) dt  dx
//        -       -
//        x0   a0 + a1 x
//
// Let F be such that F''=f. Then, if a1<>0 and b1<>0:
//
//       1                                   1
//  I = --- |F(b0+b1 x1) - F(b0+b1 x0)|  -  --- |F(a0+a1 x1) - F(a0+a1 x0)|
//      b1                                  a1

{
  double aterm, bterm;
  Polynomial F = Primitive(2);
  
  if (fabs(a1) < 1e-50)
    aterm = (x1-x0)*F.Derivative(a0);
  else
    aterm = ( F(a0+a1*x1) - F(a0+a1*x0) ) / b1;
  
  if (fabs(b1) < 1e-50)
    bterm = (x1-x0)*F.Derivative(b0);
  else
    bterm = ( F(b0+b1*x1) - F(b0+b1*x0) ) / b1;
  
  return bterm - aterm;
}

void Polynomial::operator = (const Polynomial& g)
{
  n = g.n;
  allocate(n+1); 
  for (int i=0; i<=n; i++) a[i] = g.a[i];
}

void Polynomial::operator += (const Polynomial& g)
{
  raise_degree(g.n);
  for (int i=0; i<=g.n; i++) a[i] += g.a[i];
}

void Polynomial::operator -= (const Polynomial& g)
{
  raise_degree(g.n);
  for (int i=0; i<=g.n; i++) a[i] -= g.a[i];
}

void Polynomial::operator *= (double t)
{
  for (int i=0; i<=n; i++) a[i] *= t;
}

void Polynomial::operator /= (double t)
{
  for (int i=0; i<=n; i++) a[i] /= t;
}

Polynomial operator - (const Polynomial& f)
{
  double* a = new double[f.n+1];
  for (int i=0; i<=f.n; i++) a[i] = -f.a[i];
  return Polynomial(a,f.n);
}

Polynomial operator + (const Polynomial& p, const Polynomial& q)
{
  int degree = (p.n > q.n) ? p.n : q.n;
  double* a = new double[degree+1];

  for (int i=0; i<=degree; i++) a[i]=0;
  for (int i=0; i<=p.n; i++) a[i] += p.a[i];
  for (int i=0; i<=q.n; i++) a[i] += q.a[i];

  return Polynomial(a, degree);
}

Polynomial operator - (const Polynomial& p, const Polynomial& q)
{
  int degree = (p.n > q.n) ? p.n : q.n;
  double* a = new double[degree+1];

  for (int i=0; i<=degree; i++) a[i]=0;
  for (int i=0; i<=p.n; i++) a[i] += p.a[i];
  for (int i=0; i<=q.n; i++) a[i] -= q.a[i];

  return Polynomial(a, degree);
}

Polynomial operator * (const Polynomial& p, const Polynomial& q)
{
  int i,j;

  int degree = p.n + q.n;
  double* a = new double[degree+1];
  for (i=0; i<=degree; i++) a[i]=0;
  
  for (i=0; i<=p.n; i++)
    for (j=0; j<=q.n; j++)
      a[i+j] += p.a[i]*q.a[j];
      
  return Polynomial(a, degree);
}

Polynomial operator * (double c, const Polynomial& f)
{
  double* a = new double[f.n+1];
  for (int i=0; i<=f.n; i++) a[i] = c*f.a[i];
  return Polynomial(a,f.n);
}

Polynomial operator * (const Polynomial& f, double c)
{
  return c*f;
}

Polynomial operator / (const Polynomial& f, double c)
{
  double* a = new double[f.n+1];
  for (int i=0; i<=f.n; i++) a[i] = f.a[i]/c;
  return Polynomial(a,f.n);
}

std::ostream &operator<<(std::ostream& to, const Polynomial& p)
// format: degree+1 a[0] a[1] ... a[degree]
{
  to << (p.n + 1) << " ";
  for (int i=0; i<=p.n; i++) to << p.a[i] << ' ';
  return to;
}

std::istream &operator>>(std::istream &from, Polynomial &p)
// format: degree+1 a[0] a[1] ... a[degree]
{
  int num;
  from >> num;

  p.n = num-1;

  if (num > 0) {
    p.a = new double[num];
    for (int i=0; i<num; i++) from >> p.a[i];
  }

  return from;
}

