// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

//

#include <function.h>
#include <quadfunc.h>
#include <double3.h>
#include <cmath>
#include <fstream>

//-----------------------------------------------------------------------//
//                          QuadraticFunction                            //
//-----------------------------------------------------------------------//

int QuadraticFunction::dimension() const
{    
  if (A) return A->NumRows();
  if (B) return B->NumRows();
  return 0;
}

void QuadraticFunction::GetMatrix(DoubleMatrix& A0) const
// returns the matrix A
{
  int n = dimension();
  
  A0.SetSize(n,n);
  if (A)
    A->GetMatrix(A0);
  else
    A0.MakeZero();
}

void QuadraticFunction::GetVector(DoubleVector& B0) const
// returns the vector B
{
  int n = dimension();
  
  B0.SetSize(n);
  if (B)
    B->GetVector(B0);
  else
    B0.MakeZero();
}


double QuadraticFunction::val(const DoubleVector& x) const
{
  double Result = C;
  if (A) Result += (x,(*A)*x);
  if (B) Result += ((*B),x);

  return Result;
};

DoubleVector QuadraticFunction::grad(const DoubleVector& x) const
// gradient is equal to
//
//                   T       T     T
//                  x ( A + A ) + B
//
{
  int n = dimension(); 
  DoubleVector Result(n);
  
  if (B)
    B->GetVector(Result);
  else
    Result.MakeZero();

  if (A) {
    Result += A->GetVectorMatrixProduct(x);
    Result += A->GetMatrixVectorProduct(x);
  }

  return Result;
}
   
DoubleMatrix QuadraticFunction::hess(const DoubleVector&) const
//
// hess is equal to
//
//                          T
//                     A + A
//
{
  int n = dimension();
  DoubleMatrix Result(n,n);
  
  if (A) {
    A->GetMatrix(Result);
    Result.MakeSymmetric();
    Result *= 2;
  }
  else
  {
    Result.MakeZero();
  }

  return Result;
}

double QuadraticFunction::hessleftright(const DoubleVector& x, const DoubleVector& d, const DoubleVector& e) const
{
  if (A)
    return (d*(*A) + (*A)*d, e);
  else
    return 0;
}

void QuadraticFunction::GetMinimum(DoubleVector& X, SolveMethod method) const
{
  if (A == 0) {
    std::cerr << "A==0 in QuadraticFunction::GetMinimum()" << std::endl;
  }

  DoubleMatrix AA;
  GetMatrix(AA);
  DoubleVector BB;
  GetVector(BB);

  switch (method) {
    case BKP_method: {
      X = AA.SolveBKP(-BB/2);
      break;
    }
    case CH_method: {
      X = AA.SolveCH(-BB/2);
      break;
    }
    case LDL_method: {
      X = AA.SolveLDL(-BB/2);
      break;
    }
    case LU_method: {
      X = AA.SolveLU(-BB/2);
      break;
    }
    case QR_method: {
      X = AA.SolveQR(-BB/2);
      break;
    }   
  }
}

void QuadraticFunction::GetConstrainedMinimum(const DoubleMatrix& F, const DoubleVector& G, DoubleVector& X, SolveMethod method) const
// unconstrained system:
// g(Y)  =  f(FY + G)  =  Y^T P Y + Q Y + R    with
//      P = F^T A F
//      Q = (G^T (A + A^T) + B^T) F
//      R = G^T A G + B^T G + C
{
std::cout << "computing FTAF..." << std::flush;
  DoubleMatrix AA;
  GetMatrix(AA);
  DoubleVector BB;
  GetVector(BB);
  
  DoubleMatrix P = F.Transpose()*AA*F;
  DoubleVector Q = (G*AA + AA*G + BB)*F;
  //double R = (G,AA*G) + (BB,G) + C;
std::cout << " done!" << std::endl;

  DoubleVector Y(F.NumCols());

  switch (method) {
    case BKP_method: {
      Y = (P+P.Transpose()).SolveBKP(-Q);
      break;
    }
    case CH_method: {
      Y = (P+P.Transpose()).SolveCH(-Q);
      break;
    }
    case LDL_method: {
      Y = (P+P.Transpose()).SolveLDL(-Q);
      break;
    }
    case LU_method: {
      Y = (P+P.Transpose()).SolveLU(-Q);
      break;
    }
    case QR_method: {
      Y = (P+P.Transpose()).SolveQR(-Q);
      break;
    }   
  }

//  std::cout << "|PY+Q/2| = " << (P*Y + Q/2).Length() << std::endl;
//  std::cout << "minimum: Y^T P Y + Q^T Y + R = " << (Y,P*Y) + (Q,Y) + R << std::endl;
//  std::ofstream to("quadfunc.mtx");
//  to << P << std::endl << Q << std::endl << R << std::endl << Y << std::endl;
//  to.close();

  X = F*Y + G;
}

void QuadraticFunction::GetConstrainedMinimumLagrange(const DoubleMatrix& D,
  const DoubleVector& E, DoubleVector& X, DoubleVector& Lambda, SolveMethod method) const
{
  int m = D.NumRows();
  int n = D.NumCols();

  X.SetSize(n);
  Lambda.SetSize(m);

  DoubleMatrix AA;
  GetMatrix(AA);
  DoubleVector BB;
  GetVector(BB);

  DoubleMatrix P(m+n,m+n);
  P.MakeZero();
  DoubleVector Q(m+n);
  Q.MakeZero();

  for (int i=0; i<n; i++)
    for (int j=0; j<n; j++)
      P(i,j) = AA(i,j) + AA(j,i);

  for (int i=0; i<m; i++)
    for (int j=0; j<n; j++) {
      P(j,i+n) = D(i,j);
      P(i+n,j) = D(i,j);
    }

  for (int i=0; i<n; i++)
    Q(i) = -BB(i);
  for (int i=0; i<m; i++)
    Q(i+n) = E(i);

// Solve the system PZ=Q
  DoubleVector Z;
  
  switch (method) {
    case BKP_method: {
      Z = P.SolveBKP(Q);
      break;
    }
    case CH_method: {
      Z = P.SolveCH(Q);
      break;
    }
    case LDL_method: {
      Z = P.SolveLDL(Q);
      break;
    }
    case LU_method: {
      Z = P.SolveLU(Q);
      break;
    }
    case QR_method: {
      Z = P.SolveQR(Q);
      break;
    }   
  }

  std::cout << "|PZ-Q| = " << (P*Z - Q).Length() << std::endl;
  
  for (int i=0; i<n; i++)
    X(i) = Z(i);
  for (int i=0; i<m; i++)
    Lambda(i) = Z(n+i);
}

//-----------------------------------------------------------------------//
//                          ListQuadraticFunctionVector                  //
//-----------------------------------------------------------------------//

int ListQuadraticFunctionVector::NumRows() const
{
  if (f.Count > 0)
    return f[0].dimension();
  else
    return 0;
}

void ListQuadraticFunctionVector::AddToVector(DoubleVector& Y) const
{
  for (int i=0; i<f.Count; i++) {
    SparseVector* B = f[i].GetB();
    if (B) B->AddToVector(Y);
  }
}

void ListQuadraticFunctionVector::MultAddToVector(double t, DoubleVector& Y) const
{
  for (int i=0; i<f.Count; i++) {
    SparseVector* B = f[i].GetB();
    if (B) B->MultAddToVector(t,Y);
  }
}

double ListQuadraticFunctionVector::GetInnerProduct(const DoubleVector& Y) const
{
  double Result = 0;
  
  for (int i=0; i<f.Count; i++) {
    SparseVector* B = f[i].GetB();
    if (B) Result += B->GetInnerProduct(Y);
  }

  return Result;
}

void ListQuadraticFunctionVector::operator*= (double t)
{
  for (int i=0; i<f.Count; i++) {
    SparseVector* B = f[i].GetB();
    if (B) (*B) *= t;
  }
}

void ListQuadraticFunctionVector::Show() const
{
  std::cout << "ListQuadraticFunctionVector::Show()" << std::endl;
}

//-----------------------------------------------------------------------//
//                          ListQuadraticFunctionMatrix                  //
//-----------------------------------------------------------------------//

int ListQuadraticFunctionMatrix::NumCols() const
{
  if (f.Count > 0)
    return f[0].dimension();
  else
    return 0;
}

int ListQuadraticFunctionMatrix::NumRows() const
{
  if (f.Count > 0)
    return f[0].dimension();
  else
    return 0;
}

void ListQuadraticFunctionMatrix::AddToMatrix(DoubleMatrix& B) const
{
  for (int i=0; i<f.Count; i++) {
    SparseMatrix* A = f[i].GetA();
    if (A) A->AddToMatrix(B);
  }
}

DoubleVector ListQuadraticFunctionMatrix::GetVectorMatrixProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumCols());
  Result.MakeZero();
  
  for (int i=0; i<f.Count; i++) {
    SparseMatrix* A = f[i].GetA();
    if (A) Result += A->GetVectorMatrixProduct(X);
  }
  
  return Result;
}

DoubleVector ListQuadraticFunctionMatrix::GetMatrixVectorProduct(const DoubleVector& X) const
{
  DoubleVector Result(NumRows());
  Result.MakeZero();
  
  for (int i=0; i<f.Count; i++) {
    SparseMatrix* A = f[i].GetA();
    if (A) Result += A->GetMatrixVectorProduct(X);
  }
  
  return Result;
}

void ListQuadraticFunctionMatrix::operator*= (double t)
{
  for (int i=0; i<f.Count; i++) {
    SparseMatrix* A = f[i].GetA();
    if (A) (*A) *= t;
  }
}

//-----------------------------------------------------------------------//
//                          QuadraticFunctionList                        //
//-----------------------------------------------------------------------//

double QuadraticFunctionList::GetC() const
// returns the constant C
{
  double Result = 0;

  for (int i=0; i<f.Count; i++)
    Result += f[i].GetC();
    
  return Result;
}

int QuadraticFunctionList::dimension() const
{
  if (f.Count > 0)
    return f[0].dimension();
  else
    return 0;
}

double QuadraticFunctionList::val(const DoubleVector& x) const
{
  double result = 0;
  for (int i=0; i<f.Count; i++)
    result += f[i](x);

  return result;
}

DoubleVector QuadraticFunctionList::grad(const DoubleVector& x) const
{
  int n = dimension();
  DoubleVector result(n);

  for (int i=0; i<f.Count; i++)
    result += f[i].grad(x);

  return result; 
}

DoubleMatrix QuadraticFunctionList::hess(const DoubleVector& x) const
{
  int n = dimension();
  DoubleMatrix result(n,n);

  for (int i=0; i<f.Count; i++)
    result += f[i].hess(x);

  return result; 
}

void QuadraticFunctionList::operator*= (double t)
{
  for (int i=0; i<f.Count; i++)
    f[i] *= t;
}

