// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sat 06-15-1996

#include <vrml.h>
#include <polyline.h>

static double GetAngleFromSinCos(double sina, double cosa)
// This function returns a value a such that sin(a)=sina and cos(a) = cosa.
{
  double a = acos(cosa);

  // this is not very efficient...
  if ( ((sin(a) > 0) && (sina < 0)) || ((sin(a) < 0) && (sina > 0)) )
    a = 2*M_PI - a;

  return a;
}

void VRMLWriteBeginSeparator(std::ostream& out, int numspaces)
{
  for (int i=0; i<numspaces; i++)
    out << ' ';
  out << "Separator {" << std::endl;
}

void VRMLWriteEndSeparator(std::ostream& out, int numspaces)
{
  for (int i=0; i<numspaces; i++)
    out << ' ';
  out << "}" << std::endl << std::endl;
}

void VRMLWriteFileHeader(std::ostream& out)
{
  out << "#VRML V1.0 ascii" << std::endl;
}

void VRMLWriteColor(std::ostream& out, const double3& rgb)
// This function writes an RGB color.
{
  out << "  Material {" << std::endl;
  out << "    diffuseColor " << rgb[0] << " " << rgb[1] << " " << rgb[2] << std::endl;
  out << "    shininess 0.2" << std::endl;
  out << "    transparency 0" << std::endl;
  out << "  }" << std::endl;
}

void VRMLWriteTranslation(std::ostream& out, const double3& x)
// This function writes a translation over the vector x.
{
  out << "  Translation { translation " << x[0] << " " << x[1] << " "
        << x[2] << " }" << std::endl;
}

void VRMLWriteAxialRotation(std::ostream& out, const double3& n, double phi)
// This function writes a rotation around the axis n over the angle phi.
{
  out << "  Rotation { rotation " << n[0] << " " << n[1] << " "
        << n[2] << " " << phi << " }" << std::endl;
}

void VRMLWriteVectorRotation(std::ostream& out, const double3& a, const double3& b)
// This function writes a rotation that maps a/a.Unit() to b/b.Unit().
{
  // compute a rotation from axis a to axis b
  double3 n = (a^b).Unit();
  double sinphi = (a^b).Length() / (a.Length()*b.Length());
  double cosphi = (a,b) / (a.Length()*b.Length());
  double phi = GetAngleFromSinCos(sinphi, cosphi);

  VRMLWriteAxialRotation(out, n, phi);
}

void VRMLWriteCylinder(std::ostream& out, const double3& x0, const double3& x1, double r)
// This function writes a cylinder with central axis (x0,x1) and radius r.
{
  double3 a(0,1,0);
  double3 b = x1-x0;

  VRMLWriteBeginSeparator(out, 2);
  out << "  "; VRMLWriteTranslation(out, (x0+x1)/2);
  out << "  "; VRMLWriteVectorRotation(out, a, b);
  out << "    Cylinder {" << std::endl;
  out << "      parts ALL" << std::endl;
  out << "      radius " << r << std::endl;
  out << "      height " << (x1-x0).Length() << std::endl;
  out << "    }" << std::endl;
  VRMLWriteEndSeparator(out, 2);
}

void VRMLWriteCylinder(std::ostream& out, const double3& x0, const double3& x1, double r, const double3& color)
// This function writes a cylinder with central axis (x0,x1) and radius r.
{
  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteColor(out, color);
  VRMLWriteCylinder(out, x0, x1, r);
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWriteSphere(std::ostream& out, const double3& m, double r)
// This function writes a sphere with center m and radius r.
{
  VRMLWriteBeginSeparator(out, 2);
  out << "  "; VRMLWriteTranslation(out, m);
  out << "    Sphere { radius " << r << " }" << std::endl;
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWriteSphere(std::ostream& out, const double3& m, double r, const double3& color)
// This function writes a sphere with center m and radius r.
{
  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteColor(out, color);
  VRMLWriteSphere(out, m, r);
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWriteCone(std::ostream& out, const double3& bottom, const double3& top, double r)
// This function writes a cone with central axis (bottom,top) and radius r.
{
  double3 a(0,1,0);
  double3 b = top-bottom;
  VRMLWriteBeginSeparator(out, 2);
  out << "  "; VRMLWriteTranslation(out, (bottom+top)/2);
  out << "  "; VRMLWriteVectorRotation(out, a, b);
  out << "    Cone {" << std::endl;
  out << "      parts ALL" << std::endl;
  out << "      bottomRadius " << r << std::endl;
  out << "      height " << (top-bottom).Length() << std::endl;
  out << "    }" << std::endl;
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWriteCone(std::ostream& out, const double3& bottom, const double3& top, double r, const double3& color)
// This function writes a cone with central axis (bottom,top) and radius r.
{
  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteColor(out, color);
  VRMLWriteCone(out, bottom, top, r);
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWritePolygon(std::ostream& out, const double3& p0, const double3& p1, const double3& p2, const double3& p3)
// This function writes a polygon containing 4 points.
{
  out << "    Coordinate3 {" << std::endl;
  out << "       point [" << std::endl;
  out << "         " << p0.x() << " " << p0.y() << " " << p0.z() << "," << std::endl;
  out << "         " << p1.x() << " " << p1.y() << " " << p1.z() << "," << std::endl;
  out << "         " << p2.x() << " " << p2.y() << " " << p2.z() << "," << std::endl;
  out << "         " << p3.x() << " " << p3.y() << " " << p3.z() << std::endl;
  out << "       ]" << std::endl;
  out << "    }" << std::endl;
  out << "    IndexedFaceSet {" << std::endl;
  out << "       coordIndex[ 0, 1, 2, 3, -1 ]" << std::endl;
  out << "    }" << std::endl;
}

void VRMLWritePolygon(std::ostream& out, const double3& p0, const double3& p1, const double3& p2, const double3& p3, const double3& color)
// This function writes a polygon containing 4 points.
{
  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteColor(out, color);
  VRMLWritePolygon(out, p0, p1, p2, p3);
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWriteLine(std::ostream& out, const double3& x0, const double3& x1)
// This function writes a line between x0 an x1.
{
  out << "    Coordinate3 {" << std::endl;
  out << "       point [" << std::endl;
  out << "         " << x0.x() << " " << x0.y() << " " << x0.z() << "," << std::endl;
  out << "         " << x1.x() << " " << x1.y() << " " << x1.z() << std::endl;
  out << "       ]" << std::endl;
  out << "    }" << std::endl;
  out << "    IndexedLineSet {" << std::endl;
  out << "       coordIndex[ 0, 1, -1 ]" << std::endl;
  out << "    }" << std::endl;
}

void VRMLWriteLine(std::ostream& out, const double3& x0, const double3& x1, const double3& color)
// This function writes a line between x0 an x1.
{
  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteColor(out, color);
  VRMLWriteLine(out, x0, x1);
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWriteArrow(std::ostream& out, const double3& x0, const double3& x1, double r)
// This function writes an arrow with consisting of a cylinder and a cone. The radius
// of the cylinder is r.
{
  double3 xm = x1 - (4*r) * (x1-x0).Unit();

  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteCylinder(out, x0, xm, r);
  VRMLWriteEndSeparator(out, 2);

  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteCone(out, xm, x1, 4*r);
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWriteArrow(std::ostream& out, const double3& x0, const double3& x1, double r, const double3& color)
// This function writes an arrow with consisting of a cylinder and a cone. The radius
// of the cylinder is r.
{
  VRMLWriteBeginSeparator(out, 2);
  VRMLWriteColor(out, color);
  VRMLWriteArrow(out, x0, x1, r);
  VRMLWriteEndSeparator(out, 2); 
}

void VRMLWritePolylineAsCylinders(std::ostream& out, const Polyline3D& poly,
                                  double r, const double3& rgb)
// This function draws a polyline as a sequence of cylinders with radius r.
{
  VRMLWriteBeginSeparator(out);
  VRMLWriteColor(out,rgb);

  for (int i=0; i<poly.Count-1; i++) {
    VRMLWriteCylinder(out, poly[i], poly[i+1], r);
  }

  VRMLWriteEndSeparator(out);
}

