// Copyright (c) 1994-2010: Wieger Wesselink <j.w.wesselink@tue.nl>
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

// Sun 06-16-1996

#include "vtensor.h"

#include <fstream>
#include <math.h>
#include <mystuff.h>
#include <bvalues.h>

#include <integr.h>
#include "pause.h"

//-----------------------------------------------------------------------//
//                          MatIndex                                     //
//-----------------------------------------------------------------------//
inline static int MatIndex(int i, int j)
{
  return 4*i + j;
}

//-----------------------------------------------------------------------//
//                          VariationalTensorBSpline                     //
//-----------------------------------------------------------------------//

QuadraticFunction* VariationalTensorBSpline::GetSimpleStretchEnergy() const
// computes a quadratic function f such that
//
//           -
//  f(P) =  |  <Fu|Fu> + <Fv|Fv> du dv
//         -
//
// First a matrix A is computed with coefficients A(row,col) = a_ijkl such that
//
//             3  3
//             -  -
// a_ijkl  =  |  |  N_ij  N_kl   +  N_ij   N_kl   du dv
//           -  -       u     u         v      v
//           2  2
{
  DoubleMatrix A(16,16);
  BSplineEvaluator.GetSimpleStretchEnergyMatrix(A);

  SharedBlockMatrix* B = new SharedBlockMatrix(NumControlPoints(), A);

  IntVector index(16);
  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++)
      B->AddBlock(GetPatchIndices(p,q));

  return new QuadraticFunction(B, 0, 0);
}

QuadraticFunction* VariationalTensorBSpline::GetSimpleBendEnergy() const
// computes a quadratic function f such that
//
//           -
//  f(P) =  |  <Fuu|Fuu> + 2 <Fuv|Fuv> + <Fvv|Fvv> du dv
//         -
//
// First a matrix A is computed with coefficients A(row,col) = a_ijkl such that
//
//             3  3
//             -  -
// a_ijkl  =  |  |  N_ij  N_kl   +  2 N_ij   N_kl   +  N_ij   N_kl   du dv
//           -  -       uu    uu          uv     uv        vv     vv
//           2  2
{
  DoubleMatrix A(16,16);
  BSplineEvaluator.GetSimpleBendEnergyMatrix(A);
      
  SharedBlockMatrix* B = new SharedBlockMatrix(NumControlPoints(), A);

  IntVector index(16);
  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++)
      B->AddBlock(GetPatchIndices(p,q));

  return new QuadraticFunction(B, 0, 0);
}

QuadraticFunctionList VariationalTensorBSpline::GetDataStretchEnergy(const VariationalTensorBSpline& G, int degree) const
// computes a list of quadratic functions f such that
//
//           -         -1
//  f(P) =  |  trace( I   I  ) sqrt(g ) du dv
//         -           G   F         G
//
// using numerical integration (valid degrees are 3,5,7 and 9)
//
// For each patch with domain rectangle [p+2,p+3] x [q+2,q+3] a matrix A is
// computed with coefficients A(row,col) = a_ijkl such that
//
//                p+3 q+3
//            1    -   -             T
// a_ijkl  =  -   |   |   grad (N_ij)  grad (N_kl)  g  du dv
//            2  -   -        G            G         G
//              p+2 q+2
//
// The basisfunction N_ij is nonzero on this rectangle for 0 <= i,j <= 3.
// Note that A is symmetric.
{
  QuadraticFunctionList Result;
  
  DoubleMatrix A(16,16);
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      // and put the result in the matrix A
std::cout << "p,q: " << p << " " << q << std::endl;
      
      A.MakeZero();
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u     = intpoints[m].Point();
        double weight = intpoints[m].Weight();
        
        FuncSample2D NN[4][4];
        BSplineEvaluator.GetNNSamples(u.x(), u.y(), NN);

        SurfSampleExt sample;
        G.GetSurfSampleExt(u + double2(p,q), sample);

        for (int i=0; i<=3; i++)
          for (int j=0; j<=3; j++)
            for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
              for (int l=0; l<=3; l++) {
                int row = MatIndex(i,j);
                int col = MatIndex(k,l);

                double2 a;
                a = sample.I_inv * double2(NN[i][j].fu, NN[i][j].fv);
                double3 grad_ij = a[0]*sample.Fu + a[1]*sample.Fv;
                a = sample.I_inv * double2(NN[k][l].fu, NN[k][l].fv);
                double3 grad_kl = a[0]*sample.Fu + a[1]*sample.Fv;
                
                A(row,col) += weight*(grad_ij,grad_kl)*sqrt(sample.g);
              }       
      }
      A.CopyUpperToLower();

      Result += new QuadraticFunction(new BlockMatrix(NumControlPoints(), GetPatchIndices(p,q), A), 0, 0);
    }
    
  return Result;
}

QuadraticFunctionList VariationalTensorBSpline::GetDataBendEnergy(const VariationalTensorBSpline& G, int degree) const
// computes a list of quadratic functions f such that
//
//           -  3           2     2
//  f(P) =  |  sum trace( (D  F_i)  ) sqrt(g ) du dv
//         -   i=1          G               G
//
// using numerical integration (valid degrees are 3,5,7 and 9)
//
// For each patch with domain rectangle [p+2,p+3] x [q+2,q+3] a matrix A is
// computed with coefficients A(row,col) = a_ijkl such that
//
//                p+3 q+3
//                 -   -           2         2
// a_ijkl  =      |   |   trace( (D  N_ij) (D  N_kl) ) sqrt(g ) du dv
//               -   -             G         G               G
//              p+2 q+2
{
  QuadraticFunctionList Result;
  
  DoubleMatrix A(16,16);
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      // and put the result in the matrix A
std::cout << "p,q: " << p << " " << q << std::endl;
      
      A.MakeZero();
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u     = intpoints[m].Point();
        double weight = intpoints[m].Weight();
        
        FuncSample2D NN[4][4];
        BSplineEvaluator.GetNNSamples(u.x(), u.y(), NN);

        SurfSampleExt sample;
        G.GetSurfSampleExt(u + double2(p,q), sample);

        for (int i=0; i<=3; i++)
          for (int j=0; j<=3; j++)
            for (int k=i; k<=3; k++)            // (k >= i) => (row <= col)
              for (int l=0; l<=3; l++) {
                int row = MatIndex(i,j);
                int col = MatIndex(k,l);

                FuncSample2D& f = NN[i][j];
                FuncSample2D& g = NN[k][l];
                double2x2 D2f = sample.I_inv*( f.Hessian() - f.fu*sample.gamma1 - f.fv*sample.gamma2 );
                double2x2 D2g = sample.I_inv*( g.Hessian() - g.fu*sample.gamma1 - g.fv*sample.gamma2 );

                A(row,col) += weight * ((D2f*D2g).Trace()) * sqrt(sample.g);
              }        
      }
      A.CopyUpperToLower();

      Result += new QuadraticFunction(new BlockMatrix(NumControlPoints(), GetPatchIndices(p,q), A), 0, 0);
    }
    
  return Result;
}

QuadraticFunction* VariationalTensorBSpline::GetPointToPointAttractor(const double2& x, const double3& P) const
// computes a quadratic function f such that
//
//                    2
//  f(P) = d(F(u,v),P)
//
//
//                  a_ijkl  =  N_ij(u,v) N_kl(u,v)
//                
//                  b_ij    =  -2 N_ij(u,v)
//                
//                  c       =  (P,P)
{
  int NC = NumControlPoints(); 
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);
  
  DoubleMatrix A(16,16);
  BSplineEvaluator.GetNNNNValues(u-p, v-q, A);
  
  DoubleVector B(16);
  BSplineEvaluator.GetNNValues(u-p, v-q, B);
  B *= (-2);

  double C = (P,P);

  
  IntVector index = GetPatchIndices(p,q);
  BlockMatrix* AA = new BlockMatrix(NC, index, A);
  QuadVector* BB  = new QuadVector(NC, index, B, P);
 
  return new QuadraticFunction(AA, BB, C);
}

QuadraticFunction* VariationalTensorBSpline::GetPointToLineAttractor(const double2& x, const Line3D& line) const
// computes a quadratic function f such that
//
//                       2
//  f(P) = d(F(u,v),line)
//
{
  int NC = NumControlPoints(); 
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);

  DoubleMatrix A(16,16);
  BSplineEvaluator.GetNNNNValues(u-p, v-q, A);
  
  DoubleVector B(16);
  BSplineEvaluator.GetNNValues(u-p, v-q, B);
  
  double3x3 Aline;
  double3 bline;
  line.GetDistanceMap(Aline, bline);

  IntVector index = GetPatchIndices(p,q);
 
  Quad2Matrix* AA = new Quad2Matrix(NC, index, A, Aline.Transpose()*Aline);
  QuadVector* BB = new QuadVector(NC, index, B, 2*(Aline.Transpose()*bline));
  double C = (bline,bline);
 
  return new QuadraticFunction(AA, BB, C);
}

QuadraticFunction* VariationalTensorBSpline::GetPointToPlaneAttractor(const double2& x, const Plane3D& plane) const
// computes a quadratic function f such that
//
//                        2
//  f(P) = d(F(u,v),plane)
//
{
  int NC = NumControlPoints(); 
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);
  
  DoubleMatrix A(16,16);
  BSplineEvaluator.GetNNNNValues(u-p, v-q, A);
  
  DoubleVector B(16);
  BSplineEvaluator.GetNNValues(u-p, v-q, B);
  
  double3 n = plane.n;
  double d = plane.d;
  double nn = (n,n);
  
  IntVector index = GetPatchIndices(p,q);
 
  Quad2Matrix* AA = new Quad2Matrix(NC, index, A, (1/nn)*DyadeMatrix(n,n));
  QuadVector* BB = new QuadVector(NC, index, B, (-2*d/nn)*n);
  double C = (d*d)/nn;
 
  return new QuadraticFunction(AA, BB, C);
}

QuadraticFunction* VariationalTensorBSpline::GetNormalToVectorAttractor(const double2& x, const double3& normal) const
// computes a quadratic function f such that
//
//                          2                      2
//  f(P) = (Fu(u,v), normal)   +  (Fv(u,v), normal) 
{
  int NC = NumControlPoints(); 
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);
  
  DoubleMatrix A(16,16);
  BSplineEvaluator.GetNuNuNvNvValues(u-p, v-q, A);
  
  IntVector index = GetPatchIndices(p,q);
 
  Quad2Matrix* AA = new Quad2Matrix(NC, index, A, DyadeMatrix(normal,normal));
 
  return new QuadraticFunction(AA, 0, 0);
}

QuadraticFunction* VariationalTensorBSpline::GetRegionToPointAttractor(const Polygon2D& gamma, const double3& P) const
// computes a quadratic function f such that
//
//             -              2
//  f(P) =    |    d(F(u,v),P)  du dv
//           -
//         gamma
//
// where gamma is a convex polygon
//
//                  a_ijkl  =  Integral_gamma N_ij(u,v) N_kl(u,v)
//                
//                  b_ij    =  Integral_gamma -2 N_ij(u,v)
//                
//                  c       =  Integral_gamma (P,P)
{
  QuadraticFunctionList* Result = new QuadraticFunctionList();
  int NC = NumControlPoints(); 
  
  List<VerticalTrapezoid2D> trapezoids = gamma.GetTrapezoids(1e-10);
  
  DoubleMatrix A(16,16);
  DoubleVector B(16);
  double C;
  
  for (int m=0; m<trapezoids.Count; m++) {
    double2 center = trapezoids[m].Center();   
    int p = USegment(center.x());
    int q = VSegment(center.y());
    trapezoids[m].Translate(double2(-p,-q));
    
    BSplineEvaluator.GetNNNNIntegrals(trapezoids[m], A);
    
    BSplineEvaluator.GetNNIntegrals(trapezoids[m], B);
    B *= (-2);   
  
    C = (P,P)*trapezoids[m].Area();
  
    IntVector index = GetPatchIndices(p,q);
    BlockMatrix* AA = new BlockMatrix(NC, index, A);
    QuadVector* BB  = new QuadVector(NC, index, B, P);
    
    *Result += (new QuadraticFunction(AA, BB, C));
  }
  
  return Result;
}

double VariationalTensorBSpline::ComputeRegionToLineAttractor(const Polygon2D& gamma, const Line3D& line) const
{
  IntPoint2DList intpoints = IntPointEvaluator.GetPolygonIntPoints(gamma, 5);
  
  double Result = 0;
  double TotalWeight = 0;

  for (int i=0; i<intpoints.Count; i++) {
    double2 u = intpoints[i].Point();
    double dist = line.Distance(GetPoint(u));
    Result += (intpoints[i].Weight() * dist * dist);
    TotalWeight += intpoints[i].Weight();
  }
  
  std::cout << "area = " << gamma.Area() << " total weight = " << TotalWeight << std::endl;
  return Result;
}

QuadraticFunction* VariationalTensorBSpline::GetRegionToLineAttractor(const Polygon2D& gamma, const Line3D& line) const
// computes a quadratic function f such that
//
//             -                 2
//  f(P) =    |    d(F(u,v),line)  du dv
//           -
//         gamma
//
// where gamma is a convex polygon
{
  QuadraticFunctionList* Result = new QuadraticFunctionList();
  int NC = NumControlPoints(); 
  
  List<VerticalTrapezoid2D> trapezoids = gamma.GetTrapezoids(1e-10);
  
  DoubleMatrix A(16,16);
  DoubleVector B(16);
  double C;
  
  for (int m=0; m<trapezoids.Count; m++) {
    double2 center = trapezoids[m].Center();   
    int p = USegment(center.x());
    int q = VSegment(center.y());
    trapezoids[m].Translate(double2(-p,-q));
    
    BSplineEvaluator.GetNNNNIntegrals(trapezoids[m], A);   
    BSplineEvaluator.GetNNIntegrals(trapezoids[m], B); 

    double3x3 Aline;
    double3 bline;
    line.GetDistanceMap(Aline, bline);
  
    IntVector index = GetPatchIndices(p,q);
    Quad2Matrix* AA = new Quad2Matrix(NC, index, A, Aline.Transpose()*Aline);
    QuadVector* BB = new QuadVector(NC, index, B, 2*(Aline.Transpose()*bline));
    C = (bline,bline)*trapezoids[m].Area();

    *Result += (new QuadraticFunction(AA, BB, C));
  }
  
  return Result;
}

QuadraticFunction* VariationalTensorBSpline::GetRegionToPlaneAttractor(const Polygon2D& gamma, const Plane3D& plane) const
// computes a quadratic function f such that
//
//             -                  2
//  f(P) =    |    d(F(u,v),plane)  du dv
//           -
//         gamma
//
// where gamma is a convex polygon
{
  QuadraticFunctionList* Result = new QuadraticFunctionList();
  int NC = NumControlPoints(); 
  
  List<VerticalTrapezoid2D> trapezoids = gamma.GetTrapezoids(1e-10);
  
  DoubleMatrix A(16,16);
  DoubleVector B(16);
  double C;

  for (int m=0; m<trapezoids.Count; m++) {
    double2 center = trapezoids[m].Center();   
    int p = USegment(center.x());
    int q = VSegment(center.y());
    trapezoids[m].Translate(double2(-p,-q));
    
    BSplineEvaluator.GetNNNNIntegrals(trapezoids[m], A);   
    BSplineEvaluator.GetNNIntegrals(trapezoids[m], B);

    double3 n = plane.n;
    double d = plane.d;
    double nn = (n,n);
    
    IntVector index = GetPatchIndices(p,q);
   
    Quad2Matrix* AA = new Quad2Matrix(NC, index, A, (1/nn)*DyadeMatrix(n,n));
    QuadVector* BB = new QuadVector(NC, index, B, (-2*d/nn)*n);
    double C = ((d*d)/nn)*trapezoids[m].Area();
   
    *Result += (new QuadraticFunction(AA, BB, C));
  }
  
  return Result;
}

QuadraticFunction* VariationalTensorBSpline::GetNormalRegionToVectorAttractor(const Polygon2D& gamma, const double3& normal) const
// computes a quadratic function f such that
//
//             -                   2                      2  
//  f(P) =    |   (Fu(u,v), normal)   +  (Fv(u,v), normal)   du dv
//           -
//         gamma
{
  QuadraticFunctionList* Result = new QuadraticFunctionList();
  int NC = NumControlPoints(); 
  
  List<VerticalTrapezoid2D> trapezoids = gamma.GetTrapezoids(1e-10);
  
  DoubleMatrix A(16,16);

  for (int m=0; m<trapezoids.Count; m++) {
    double2 center = trapezoids[m].Center();   
    int p = USegment(center.x());
    int q = VSegment(center.y());
    trapezoids[m].Translate(double2(-p,-q));
    
    BSplineEvaluator.GetNuNuNvNvIntegrals(trapezoids[m], A);

    IntVector index = GetPatchIndices(p,q);  
    Quad2Matrix* AA = new Quad2Matrix(NC, index, A, DyadeMatrix(normal,normal));
   
    *Result += (new QuadraticFunction(AA, 0, 0));
  }
  
  return Result;
}

LinearEquation* VariationalTensorBSpline::GetPointToPointConstraint(const double2& x, const double3& P) const
// returns the equation that represents the constraint
//
//                F(u,v) = P
{
  int NC = NumControlPoints(); 
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);
  double NN[4][4];
  BSplineEvaluator.GetNNValues(u-p, v-q, NN);

  DoubleVector a(16);
  int k=0;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      a(k++) = NN[i][j];
 
  return new PointConstraintEquation(NC, GetPatchIndices(p,q), a, P);
}

LinearEquation* VariationalTensorBSpline::GetPointToLineConstraint(const double2& x, const Line3D& line) const
// returns the equation that represents the constraint
//
//                F(u,v) on line
{
  int NC = NumControlPoints(); 
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);
  double NN[4][4];
  BSplineEvaluator.GetNNValues(u-p, v-q, NN);

  DoubleVector a(16);
  int k=0;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      a(k++) = NN[i][j];

  return new LineConstraintEquation(NC, GetPatchIndices(p,q), a, line);
}

LinearEquation* VariationalTensorBSpline::GetNormalToVectorConstraint(const double2& x, const double3& normal) const
// returns the equation that represents the constraint
//
//                N(u,v) parallel to normal
{
  int NC = NumControlPoints();
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);
  double NNu[4][4], NNv[4][4];
  BSplineEvaluator.GetNNuValues(u-p, v-q, NNu);
  BSplineEvaluator.GetNNvValues(u-p, v-q, NNv);
 
  DoubleVector a(16), b(16);
  int k=0;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++) {
      a(k)   = NNu[i][j];
      b(k++) = NNv[i][j];
    }
 
  return new NormalConstraintEquation(NC, GetPatchIndices(p,q), a, b, normal);
}

LinearEquation* VariationalTensorBSpline::GetPointToPlaneConstraint(const double2& x, const Plane3D& plane) const
// returns the equation that represents the constraint
//
//                F(u,v) in plane
{
  int NC = NumControlPoints(); 
  double u = x.x();
  double v = x.y();
  int p = USegment(u);
  int q = VSegment(v);
  double NN[4][4];
  BSplineEvaluator.GetNNValues(u-p, v-q, NN);

  DoubleVector a(16);
  int k=0;
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      a(k++) = NN[i][j];

  return new PlaneConstraintEquation(NC, GetPatchIndices(p,q), a, plane);
}

LinearEquation* VariationalTensorBSpline::GetHorizontalCurveConstraint(int2 u, double v0, const BSpline& curve) const
// returns the equation that represents the constraint
//
//                F(t,v0) = curve(t)          u.x() <= t <= u.y()
{
  int NC = NumControlPoints(); 
  int q = VSegment(v0);
  
  double N[4];
  BSplineEvaluator.GetNValues(v0-q,N);
  DoubleVector a(4);
  for (int k=0; k<4; k++)
    a(k) = N[k];

  CurveConstraintEquation* Result = new CurveConstraintEquation(NC, a);

  int imin = u.x() - 2; // imin = USegment(u.x()); ??
  int imax = u.y();     //
  
  for (int i = imin; i <= imax; i++) {
    int4 index;
    for (int j=0; j<=3; j++)
      index[j] = ControlPointIndex(i,j+q);
    Result->Add(index, curve.ControlPoint(i));
  }
  
  return Result;
}

LinearEquation* VariationalTensorBSpline::GetVerticalCurveConstraint(double u0, int2 v, const BSpline& curve) const
// returns the equation that represents the constraint
//
//                F(u0,t) = curve(t)          v.x() <= t <= v.y()
{
  int NC = NumControlPoints(); 
  int p = USegment(u0);
  
  double N[4];
  BSplineEvaluator.GetNValues(u0-p,N);
  DoubleVector a(4);
  for (int k=0; k<4; k++)
    a(k) = N[k];

  CurveConstraintEquation* Result = new CurveConstraintEquation(NC, a);

  int jmin = v.x() - 2; // imin = VSegment(v.x()); ??
  int jmax = v.y();     //
  
  for (int j = jmin; j <= jmax; j++) {
    int4 index;
    for (int i=0; i <= 3; i++)
      index[i] = ControlPointIndex(i+p,j);
    Result->Add(index, curve.ControlPoint(j));
  }
  
  return Result;
}

void VariationalTensorBSpline::GetBoundaryCurves(BSpline& umincurve, BSpline& umaxcurve, BSpline& vmincurve, BSpline& vmaxcurve) const
// returns the curves that represent the boundary of the surface
{
  //                  nc-1    3
  // umincurve f(v) = sum  ( sum N_i(umin) P_ij ) N_j(v)
  //                  j=0    i=0
  //
  //                  nc-1   nr-1
  // umaxcurve f(v) = sum  ( sum N_i(umax) P_ij ) N_j(v)
  //                  j=0   i=nr-4
  //
  //                  nr-1    3
  // vmincurve f(u) = sum  ( sum N_j(vmin) P_ij ) N_i(v)
  //                  i=0    j=0
  //
  //                  nr-1   nc-1
  // vmaxcurve f(u) = sum  ( sum N_j(vmax) P_ij ) N_i(v)
  //                  i=0   j=nc-4
  
  int nc = NumCols();
  int nr = NumRows();
  double N[4];

  BSplineEvaluator.GetNValues(2,N); 
  for (int j=0; j<nc; j++) {
    double3 Q(0,0,0);
    for (int i=0; i<4; i++)
      Q += N[i]*GetControlPoint(i,j);
    umincurve.InsertControlPoint(new double3(Q));  
  }

  BSplineEvaluator.GetNValues(3,N); 
  for (int j=0; j<nc; j++) {
    double3 Q(0,0,0);
    for (int i=nr-4; i<nr; i++)
      Q += N[i-nr+4]*GetControlPoint(i,j);
    umaxcurve.InsertControlPoint(new double3(Q));  
  }

  BSplineEvaluator.GetNValues(2,N); 
  for (int i=0; i<nr; i++) {
    double3 Q(0,0,0);
    for (int j=0; j<4; j++)
      Q += N[j]*GetControlPoint(i,j);
    vmincurve.InsertControlPoint(new double3(Q));  
  }

  BSplineEvaluator.GetNValues(3,N); 
  for (int i=0; i<nr; i++) {
    double3 Q(0,0,0);
    for (int j=nc-4; j<nc; j++)
      Q += N[j-nc+4]*GetControlPoint(i,j);
    vmaxcurve.InsertControlPoint(new double3(Q));  
  }
}

List<LinearEquation> VariationalTensorBSpline::GetBoundaryConstraints() const
// returns the equations that represent the boundary of the surface
//
//                F(umin,v) = ...
//                F(umax,v) = ...
//                F(u,vmin) = ...
//                F(u,vmax) = ...
{
  BSpline umincurve, umaxcurve, vmincurve, vmaxcurve;
  GetBoundaryCurves(umincurve, umaxcurve, vmincurve, vmaxcurve);

  // There is a small problem: the constraint equations of these curves
  // are linear dependent. Therefore, we have to restrict the domain
  // of the curves somewhat.

  double umin = MinUPar();
  double umax = MaxUPar();
  double vmin = MinVPar();
  double vmax = MaxVPar();

  // domain of a BSpline b is: [2, b.NumControlPoints()-1]
  int2 u,v;
  v = int2(3, umincurve.NumControlPoints()-2);
  LinearEquation* c0 = GetVerticalCurveConstraint(umin, v, umincurve);

  v = int2(3, umaxcurve.NumControlPoints()-2);
  LinearEquation* c1 = GetVerticalCurveConstraint(umax, v, umaxcurve);
  
  u = int2(2, vmincurve.NumControlPoints()-1);
  LinearEquation* c2 = GetHorizontalCurveConstraint(u, vmin, vmincurve);

  u = int2(2, vmaxcurve.NumControlPoints()-1);
  LinearEquation* c3 = GetHorizontalCurveConstraint(u, vmax, vmaxcurve);

  List<LinearEquation> Result;
  Result += c0;
  Result += c1;
  Result += c2;
  Result += c3;

  return Result;
}
    
double VariationalTensorBSpline::ComputeSimpleStretchEnergy() const
// computes the following integral
//      -
//     |   <Fu|Fu> + <Fv|Fv> du dv
//    -
{
  QuadraticFunction* f = GetSimpleStretchEnergy();
  DoubleVector X = GetControlPointVector();
  return (*f)(X);
}

double VariationalTensorBSpline::ComputeSimpleStretchEnergyNumeric(int degree) const
// computes the following integral numerically
//      -
//     |   <Fu|Fu> + <Fv|Fv> du dv
//    -
{
  double Result = 0;  
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u     = intpoints[m].Point();
        double weight = intpoints[m].Weight();       
        Result += (weight * GetSurfSample(u+double2(p,q)).SimpleArea());
      }  
    }
    
  return Result;
}

double VariationalTensorBSpline::ComputeDataStretchEnergy(const VariationalTensorBSpline& G, int degree) const
// computes the following integral
//           -         -1
//  f(P) =  |  trace( I   I  ) sqrt(g ) du dv
//         -           G   F         G
{
  double Result = 0;  
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u     = intpoints[m].Point();
        double weight = intpoints[m].Weight();
        
        SurfSample sampleF = GetSurfSample(u + double2(p,q));
        SurfSample sampleG = G.GetSurfSample(u + double2(p,q));
        double2x2 I_F = sampleF.FirstFundamentalForm();
        double2x2 I_G = sampleG.FirstFundamentalForm();
        
        Result += (weight*sqrt(I_G.Det())) * (I_G.Inverse()*I_F).Trace();
      }  
    }
    
  return Result;
}

double VariationalTensorBSpline::ComputeStretchEnergy(int degree) const
// computes the following integral
//      -
//     |   1 dA
//    -
{
  double Result = 0;  
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u     = intpoints[m].Point();
        double weight = intpoints[m].Weight();
        
        SurfSample sampleF = GetSurfSample(u + double2(p,q));
        
        double2x2 I_F = sampleF.FirstFundamentalForm();
        Result += (weight * sqrt(I_F.Det()));
      }  
    }
    
  return Result;
}

double VariationalTensorBSpline::ComputeSimpleBendEnergy() const
// computes the following integral
//      -
//     |   <Fuu|Fuu> + 2 <Fuv|Fuv> + <Fvv|Fvv> du dv
//    -
{
  QuadraticFunction* f = GetSimpleBendEnergy();
  DoubleVector X = GetControlPointVector();
  return (*f)(X);
}

double VariationalTensorBSpline::ComputeSimpleBendEnergyNumeric(int degree) const
// computes the following integral numerically
//      - 
//     |   <Fuu|Fuu> + 2 <Fuv|Fuv> + <Fvv|Fvv> du dv
//    -
{
  double Result = 0;  
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u     = intpoints[m].Point();
        double weight = intpoints[m].Weight();       
        Result += (weight * GetSurfSample(u+double2(p,q)).SimpleTotalCurvature());
      }  
    }
    
  return Result;
}

double VariationalTensorBSpline::ComputeDataBendEnergy(const VariationalTensorBSpline& G, int degree) const
// computes the following integral
//           -  3           2     2
//  f(P) =  |  sum trace( (D  F_i)  ) sqrt(g ) du dv
//         -   i=1          G               G
{
  double Result = 0;  
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u = intpoints[m].Point() + double2(p,q);
        double weight = intpoints[m].Weight();

        FuncSample2D F1, F2, F3;
        GetFuncSamples(u, F1, F2, F3);

        Result += (weight * (G.GetTraceExpr(u,F1) + G.GetTraceExpr(u,F2) + G.GetTraceExpr(u,F3)));
      }  
    }
    
  return Result;
}

double VariationalTensorBSpline::ComputeBendEnergy(int degree) const
// computes the following integral
//      -
//     |   (kappa1)^2 + (kappa2)^2  dA
//    -
{
  double Result = 0;  
  IntPoint2DList intpoints = IntPointEvaluator.GetRectangleIntPoints(Rectangle2D(2,3,2,3), degree);

  for (int p=0; p<NumUSegments(); p++)
    for (int q=0; q<NumVSegments(); q++) {
      // integrate over patch with domain [p+2,p+3] x [q+2,q+3]
      
      for (int m=0; m<intpoints.Count; m++) {
        double2 u     = intpoints[m].Point();
        double weight = intpoints[m].Weight();
        
        SurfSample sampleF = GetSurfSample(u + double2(p,q));
        
        double2x2 I_F = sampleF.FirstFundamentalForm();
        Result += (weight * sampleF.TotalCurvature() * sqrt(I_F.Det()));
      }  
    }
    
  return Result;
}

void VariationalTensorBSpline::TestSimpleStretchEnergy(double weight) const
{
  double val  = weight*ComputeSimpleStretchEnergy();
  double val3 = weight*ComputeSimpleStretchEnergyNumeric(3);
  double val5 = weight*ComputeSimpleStretchEnergyNumeric(5);
  double val7 = weight*ComputeSimpleStretchEnergyNumeric(7);
  double val9 = weight*ComputeSimpleStretchEnergyNumeric(9);

  DoubleVector X = GetControlPointVector();
  QuadraticFunction* f = GetSimpleStretchEnergy();
  (*f) *= weight;
  
  std::cout << "Simple stretch energy" << std::endl;
  std::cout << "       exact: " << val << std::endl;
  std::cout << "       exact: " << (*f)(X) << std::endl;
  std::cout << "    degree 3: " << val3 << std::endl;
  std::cout << "    degree 5: " << val5 << std::endl;
  std::cout << "    degree 7: " << val7 << std::endl;
  std::cout << "    degree 9: " << val9 << std::endl;
}

void VariationalTensorBSpline::TestDataStretchEnergy(double weight, const VariationalTensorBSpline& G) const
{
  QuadraticFunctionList f3 = GetDataStretchEnergy(G,3); f3 *= weight;
  QuadraticFunctionList f5 = GetDataStretchEnergy(G,5); f5 *= weight;
  QuadraticFunctionList f7 = GetDataStretchEnergy(G,7); f7 *= weight;
  QuadraticFunctionList f9 = GetDataStretchEnergy(G,9); f9 *= weight;

  double val3 = weight*ComputeDataStretchEnergy(G,3);
  double val5 = weight*ComputeDataStretchEnergy(G,5);
  double val7 = weight*ComputeDataStretchEnergy(G,7);
  double val9 = weight*ComputeDataStretchEnergy(G,9);

  DoubleVector X = GetControlPointVector();

  std::cout << "Data stretch energy" << std::endl;
  std::cout << "    degree 3: " << f3(X) << " " << val3 << std::endl;
  std::cout << "    degree 5: " << f5(X) << " " << val5 << std::endl;
  std::cout << "    degree 7: " << f7(X) << " " << val7 << std::endl;
  std::cout << "    degree 9: " << f9(X) << " " << val9 << std::endl;
}

void VariationalTensorBSpline::TestSimpleBendEnergy(double weight) const
{
  double val  = weight*ComputeSimpleBendEnergy();
  double val3 = weight*ComputeSimpleBendEnergyNumeric(3);
  double val5 = weight*ComputeSimpleBendEnergyNumeric(5);
  double val7 = weight*ComputeSimpleBendEnergyNumeric(7);
  double val9 = weight*ComputeSimpleBendEnergyNumeric(9);

  DoubleVector X = GetControlPointVector();
  QuadraticFunction* f = GetSimpleBendEnergy();
  (*f) *= weight;
  
  std::cout << "Simple bend energy" << std::endl;
  std::cout << "       exact: " << val << std::endl;
  std::cout << "       exact: " << (*f)(X) << std::endl;
  std::cout << "    degree 3: " << val3 << std::endl;
  std::cout << "    degree 5: " << val5 << std::endl;
  std::cout << "    degree 7: " << val7 << std::endl;
  std::cout << "    degree 9: " << val9 << std::endl;
}

void VariationalTensorBSpline::TestDataBendEnergy(double weight, const VariationalTensorBSpline& G) const
{
  QuadraticFunctionList f3 = GetDataBendEnergy(G,3); f3 *= weight;
  QuadraticFunctionList f5 = GetDataBendEnergy(G,5); f5 *= weight;
  QuadraticFunctionList f7 = GetDataBendEnergy(G,7); f7 *= weight;
  QuadraticFunctionList f9 = GetDataBendEnergy(G,9); f9 *= weight;

  double val3 = weight*ComputeDataBendEnergy(G,3);
  double val5 = weight*ComputeDataBendEnergy(G,5);
  double val7 = weight*ComputeDataBendEnergy(G,7);
  double val9 = weight*ComputeDataBendEnergy(G,9);

  DoubleVector X = GetControlPointVector();

  std::cout << "Data bend energy" << std::endl;
  std::cout << "    degree 3: " << f3(X) << " " << val3 << std::endl;
  std::cout << "    degree 5: " << f5(X) << " " << val5 << std::endl;
  std::cout << "    degree 7: " << f7(X) << " " << val7 << std::endl;
  std::cout << "    degree 9: " << f9(X) << " " << val9 << std::endl;
}

void VariationalTensorBSpline::TestPointToPointAttractor(double weight, const double2& u, const double3& P) const
{
  QuadraticFunction* f = GetPointToPointAttractor(u, P);
  (*f) *= weight;
  DoubleVector X = GetControlPointVector();
  
  SurfSample sample = GetSurfSample(u); 
  double dist = (sample.F - P).Length();

  std::cout << "Point to point attractor" << std::endl;
  std::cout << "    " << (*f)(X) << " " << weight*dist*dist << std::endl;
}

void VariationalTensorBSpline::TestPointToLineAttractor(double weight, const double2& u, const Line3D& line) const
{
  QuadraticFunction* f = GetPointToLineAttractor(u, line);
  (*f) *= weight;
  DoubleVector X = GetControlPointVector();
  
  SurfSample sample = GetSurfSample(u);
  double dist = line.Distance(sample.F);

  std::cout << "Point to line attractor" << std::endl;
  std::cout << "    " << (*f)(X) << " " << weight*dist*dist << std::endl;
}

void VariationalTensorBSpline::TestPointToPlaneAttractor(double weight, const double2& u, const Plane3D& plane) const
{
  QuadraticFunction* f = GetPointToPlaneAttractor(u, plane);
  (*f) *= weight;
  DoubleVector X = GetControlPointVector();
  
  SurfSample sample = GetSurfSample(u);
  double dist = plane.Distance(sample.F);

  std::cout << "Point to plane attractor" << std::endl;
  std::cout << "    " << (*f)(X) << " " << weight*dist*dist << std::endl;
}

void VariationalTensorBSpline::TestRegionToPointAttractor(const Polygon2D& gamma1, const double3& P) const
{
  Polygon2D gamma(gamma1);
  double2 u = gamma.Center();
  
  DoubleVector X = GetControlPointVector();

  QuadraticFunction* f = GetPointToPointAttractor(u,P);
  std::cout << "f(X) = " << (*f)(X) << std::endl;
  
  for (int i=0; i<10; i++) {
    QuadraticFunctionList g = GetRegionToPointAttractor(gamma, P);
    std::cout << "g(X) = " << g(X)/gamma.Area() << std::endl;
    gamma.Scale(gamma.Center(), 0.5);
  }
}

void VariationalTensorBSpline::TestRegionToLineAttractor(const Polygon2D& gamma1, const Line3D& line) const
{
  Polygon2D gamma(gamma1);
  double2 u = gamma.Center();
  
  DoubleVector X = GetControlPointVector();

  QuadraticFunction* f = GetPointToLineAttractor(u,line);
  std::cout << "f(X) = " << (*f)(X) << std::endl;
  
  for (int i=0; i<10; i++) {
    QuadraticFunctionList g = GetRegionToLineAttractor(gamma, line);
    std::cout << "g(X) = " << g(X)/gamma.Area() << std::endl;
    gamma.Scale(gamma.Center(), 0.5);
  }
}

void VariationalTensorBSpline::TestRegionToPlaneAttractor(const Polygon2D& gamma1, const Plane3D& plane) const
{
  Polygon2D gamma(gamma1);
  double2 u = gamma.Center();
  
  DoubleVector X = GetControlPointVector();

  QuadraticFunction* f = GetPointToPlaneAttractor(u,plane);
  std::cout << "f(X) = " << (*f)(X) << std::endl;
  
  for (int i=0; i<10; i++) {
    QuadraticFunctionList g = GetRegionToPlaneAttractor(gamma, plane);
    std::cout << "g(X) = " << g(X)/gamma.Area() << std::endl;
    gamma.Scale(gamma.Center(), 0.5);
  }
}

void VariationalTensorBSpline::TestPointToPointConstraint(const double2& u, const double3& P) const
{
  LinearEquation* f = GetPointToPointConstraint(u, P);
  DoubleVector X = GetControlPointVector();
  
  SurfSample sample = GetSurfSample(u);  
  double dist = (sample.F - P).Length();

  std::cout << "Point to point constraint" << std::endl;
  f->Check(X);
  std::cout << "    d(F(u,v),P) = " << dist << std::endl;
}

void VariationalTensorBSpline::TestPointToLineConstraint(const double2& u, const Line3D& line) const
{
  LinearEquation* f = GetPointToLineConstraint(u, line);
  DoubleVector X = GetControlPointVector();
  
  SurfSample sample = GetSurfSample(u);  
  double dist = line.Distance(sample.F);

  std::cout << "Point to line constraint" << std::endl;
  f->Check(X);
  std::cout << "    d(F(u,v),line) = " << dist << std::endl;
}

void VariationalTensorBSpline::TestPointToPlaneConstraint(const double2& u, const Plane3D& plane) const
{
  LinearEquation* f = GetPointToPlaneConstraint(u, plane);
  DoubleVector X = GetControlPointVector();
  
  SurfSample sample = GetSurfSample(u);  
  double dist = fabs((plane.n,sample.F) - plane.d);

  std::cout << "Point to plane constraint" << std::endl;
  f->Check(X);
  std::cout << "    |(n,F(u,v)) - d| = " << dist << std::endl;
}

void VariationalTensorBSpline::TestNormalToVectorConstraint(const double2& u, const double3& normal) const
{
  LinearEquation* f = GetNormalToVectorConstraint(u, normal);
  DoubleVector X = GetControlPointVector();

  double3 N = GetNormal(u); 
 
  std::cout << "Normal to vector constraint" << std::endl;
  f->Check(X);
  std::cout << "    |(N(u) ^ normal| = " << (N^normal).Length() << std::endl;
}
 
void VariationalTensorBSpline::TestHorizontalCurveConstraint(int2 u, double v0, const BSpline& curve) const
{
  LinearEquation* f = GetHorizontalCurveConstraint(u, v0, curve);
  DoubleVector X = GetControlPointVector();
  
  std::cout << "Curve constraint" << std::endl;
  f->Check(X);
}

void VariationalTensorBSpline::TestVerticalCurveConstraint(double u0, int2 v, const BSpline& curve) const
{
  LinearEquation* f = GetVerticalCurveConstraint(u0, v, curve);
  DoubleVector X = GetControlPointVector();
  
  std::cout << "Curve constraint" << std::endl;
  f->Check(X);
}

void VariationalTensorBSpline::TestBoundaryConstraints() const
{
  List<LinearEquation> f = GetBoundaryConstraints();
  DoubleVector X = GetControlPointVector();

  std::cout << "Boundary constraints" << std::endl;
  //f.Check(X);
}

void VariationalTensorBSpline::WriteInfo() const
// writes info of the surface to a file
{
  std::cout << "VariationalTensor product B-spline surface" << std::endl;
  std::cout << "  parameter domain         : [" << MinUPar() << "," << MaxUPar() << "] x ["
       <<                                     MinVPar() << "," << MaxVPar() << "]" << std::endl;
  std::cout << "  number of patches        : " << NumPatches() << std::endl;
  std::cout << "  number of control points : " << NumControlPoints() << std::endl;
  std::cout << "  simple stretch energy    : " << ComputeSimpleStretchEnergy() << std::endl;
  std::cout << "  simple bend energy       : " << ComputeSimpleBendEnergy() << std::endl;
  std::cout << "  exact stretch energy     : " << ComputeStretchEnergy(5) << std::endl;
  std::cout << "  exact bend energy        : " << ComputeBendEnergy(5) << std::endl;
}

